//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Omni-Family
//  File: NumericVectObj.h
//  Classe: TNumericVectObj
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 06/14/2000
//  Version: 0.0
//
//  Description:
//	  Object to be indexed by vector and metric space.
//    Numeric vectors version.
//==============================================================================

#ifndef NumericVectObjH
#define NumericVectObjH

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <typeinfo>

//------------------------------------------------------------------------------
// Class TNumericVectObj definitions
//------------------------------------------------------------------------------
template <class vectorType>
class TNumericVectObj {
  public:
    TNumericVectObj(); // null constructor
    TNumericVectObj(int dim); // main constructor
    TNumericVectObj(vectorType *vet, int dim); // constructs and initialize
    TNumericVectObj(int objSize, char *buff); // mapping constructor
         // It is faster to be used when it is not desired to
         // create a new object. The idea is to share the data space and use the
         // internal methods. It is used to avoid expensive buffer copies.
         // The user must take care about allocating and deleting the object
         // data space. ObjSize can be obtained with another object created by
         // the main constructor. ObjSize is not the number of objects in the
         // vector, but the total size of the data space. It is used instead of
         // dimension because some data (to be indexed) doesnt have dimension
         // and the indexing method wont be able to give it.
    TNumericVectObj(const TNumericVectObj<vectorType>&cObj); // copy constructor
    TNumericVectObj(char *fName, int dim); // SampleSet constructor
                                    // (to create a query object of type biased)
    TNumericVectObj(int dim, vectorType *minRg, vectorType *maxRg);
                    // randomObject constructor (to create a query object of
                    // type random)
    ~TNumericVectObj();

    // to be used with the null constructor. The obj is used as an example to
    // get definitions (the dimension in this case).
    void allocateAs(TNumericVectObj<vectorType> *obj);
    void setDim(int dim);
    void setSize(int oSize);
    vectorType& operator [](int i);
    const vectorType& operator [](int i) const;
    void setReadObjectFunction(int (*rdFunc)(char *, int, char *));
    void readObj(char *cmdLine, int condition, char *aux);
    void randomize(); // get biased
    void printVector();
    void readFromTxtFile(FILE *fdIn);
    void writeToTxtFile(FILE *fdOut);
    TNumericVectObj<vectorType>& operator=(
                                 const TNumericVectObj<vectorType> &cObj);
    void toObject(char *buff); // copies the buffer to the object. The buffer
          // MUST have the same size of the object buffer data space
    void toBuffer(char *buff); // copies the object buffer space to the buffer.
          // The user must allocate and delete the buffer
    void mapBuffer(char *buff); // map method - must be used with map
                                // constructor. The objSize must be initialized
                                // with the map constructor. Use NULL as the
                                // buffer parameter if you need.
    char* bufferArea(); // returns a pointer to a buffer area
            // containing the object. The user can NEVER delete this area
    int objectSize() { return(dataSize); };

//    typedef float distanceType;
    typedef vectorType distanceType;
    distanceType distance(TNumericVectObj<vectorType> *obj);
    // check if obj is inside the circle (*this + radius). Returns true if it
    // is (and its distance), false otherwise (and 0 as the distance).
    char isObjectIncluded(TNumericVectObj<vectorType> *obj,
                          distanceType radius, distanceType *dist);

    enum metricType {metricLInf, metricL1, metricL2, metricL3};
//    void setMetric(metricType newMetric) {metricInUse = newMetric;}

  private:
    enum tObjectClass {
      ocAllocated,
      ocMaped,
      ocSample,
      ocRandom,
      ocNull
    }; // indicates whether the constructor must delete data or not
    tObjectClass myClass;

    enum dataType {intData, floatData, doubleData};
    dataType myType;

    int dimension;

    vectorType *data;
    int dataSize;

    vectorType **sampleSet; // objects to perform biased queries
    int sampleSetSize; // indicates how many objects are in vRand
    int currentSample; // indicates how many objects are in vRand
    vectorType *minRange;
    vectorType *maxRange;

    metricType metricInUse;

    int (*readFunc)(char *, int, char *);
    void initAll(); // initialize the object with all NULL
};

//------------------------------------------------------------------------------
// Null constructor
//------------------------------------------------------------------------------
template <class vectorType>
TNumericVectObj<vectorType>::TNumericVectObj() {
  initAll();
}

//------------------------------------------------------------------------------
// Main constructor
//------------------------------------------------------------------------------
template <class vectorType>
TNumericVectObj<vectorType>::TNumericVectObj(int dim) {
  initAll();

  myClass = ocAllocated;
  dimension = dim;
  dataSize = dimension * sizeof(vectorType);
  data = new vectorType[dimension];

  memset(data, '\0', dataSize);  // initialize with 0
}

//------------------------------------------------------------------------------
// Constructs and initialize
//------------------------------------------------------------------------------
template <class vectorType>
TNumericVectObj<vectorType>::TNumericVectObj(vectorType *vet, int dim) {
  initAll();

  myClass = ocAllocated;
  dimension = dim;
  dataSize = dimension * sizeof(vectorType);
  data = new vectorType[dimension];

  memcpy(data, vet, dataSize);
}

//------------------------------------------------------------------------------
// Mapping Constructor
// It is faster to be used when it is not desired to
// create a new object. The idea is to share the data space and use the
// internal methods. It is used to avoid expensive buffer copies.
// The user must take care about allocating and deleting the object
// data space. ObjSize can be obtained with another object created by
// the main constructor. ObjSize is not the number of objects in the
// vector, but the total size of the data space. It is used instead of
// dimension because some data (to be indexed) doesnt have dimension
// and the indexing method wont be able to give it.
//------------------------------------------------------------------------------
template <class vectorType>
TNumericVectObj<vectorType>::TNumericVectObj(int objSize, char *buff) {
  initAll();

  myClass = ocMaped;
  dataSize = objSize;
  dimension = dataSize / sizeof(vectorType);
  data = (vectorType *)buff;
}

//------------------------------------------------------------------------------
// Copy constructor
//------------------------------------------------------------------------------
template <class vectorType>
TNumericVectObj<vectorType>::TNumericVectObj(
                             const TNumericVectObj<vectorType>&cObj) {
  if (cObj.myClass != ocAllocated) {
    // to avoid instability
    printf("TNumericVectObj::TNumericVectObj copy constructor error!\n"
           " The copy constructor can be used only with allocated objects.\n");
    exit(1);
  }

  initAll();

  myClass = ocAllocated;
  dimension = cObj.dimension;
  dataSize = cObj.dataSize;
  data = new vectorType[dimension];
  readFunc = cObj.readFunc;
  metricInUse = cObj.metricInUse;

  memcpy(data, cObj.data, dataSize);
}

//------------------------------------------------------------------------------
// SampleSet constructor (to create a query object of type biased).
//------------------------------------------------------------------------------
template <class vectorType>
TNumericVectObj<vectorType>::TNumericVectObj(char *fName, int dim) {
  FILE *fdIn;
  char line[256];
  char format[5];
  int i, j;

  initAll();

  myClass = ocSample;
  dimension = dim;
  dataSize = dimension * sizeof(vectorType);

  j = strlen(fName);
  for(i = 0; (i < j) && (fName[i] != '.'); i++);
  if (i == j) {
    fName[i++]='.';
    fName[i++]='s';
    fName[i++]='t';
    fName[i++]='c';
    fName[i]=0;
  }

  if ((fdIn = fopen(fName, "rt")) == NULL) {
    printf("Error! Can't open file '%s'.\n", fName);
    exit(1);
  }

  fgets(line, 256, fdIn);
  sscanf(line, "%d", &sampleSetSize);

  sampleSet = new vectorType*[sampleSetSize];
  for (i = 0; i < sampleSetSize; i++)
    sampleSet[i] = new vectorType[dimension];

  switch (myType) {
    case intData:
      sprintf(format, "%s", "%d ");
    break;
    case floatData:
      sprintf(format, "%s", "%f ");
    break;
    case doubleData:
      sprintf(format, "%s", "%lf ");
    break;
  }

  for(i = 0; i < sampleSetSize; i++) {
    data = sampleSet[i];
    for (j = 0; j < dimension; j++) {
      if (!fscanf(fdIn, format, &(data[j]))) {
        printf("TNumericVectObj::setRandBiased error!"
               " Can't read input file.\n");
        exit(1);
      }
    }
  }

  currentSample = 0;

  fclose(fdIn);
}

//------------------------------------------------------------------------------
// RandomObject constructor (to create a query object of type random)
//------------------------------------------------------------------------------
template <class vectorType>
TNumericVectObj<vectorType>::TNumericVectObj(int dim, vectorType *minRg,
                                             vectorType *maxRg) {
  myClass = ocRandom;
  dimension = dim;
  dataSize = dimension * sizeof(vectorType);
  data = new vectorType[dimension];
  minRange = new vectorType[dimension];
  maxRange = new vectorType[dimension];

  memcpy(minRange, minRg, dataSize);
  memcpy(maxRange, maxRg, dataSize);
}

//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
template <class vectorType>
TNumericVectObj<vectorType>::~TNumericVectObj() {
  switch (myClass) {
    case ocAllocated:
      if (data) {
        delete[] data;
        data = NULL;
      }
    break;

    case ocSample:
      for(currentSample = 0; currentSample < sampleSetSize; currentSample++) {
        delete[] sampleSet[currentSample];
        sampleSet[currentSample] = NULL;
      }
      delete[] sampleSet;
      sampleSet = NULL;
    break;

    case ocRandom:
      if (data) {
        delete[] data;
        data = NULL;
      }

      delete minRange;
      delete maxRange;
    break;
  }
}

//------------------------------------------------------------------------------
// Allocate vector
// To be used with the null constructor. The obj parameter is used as an
// example to get definitions (the dimension in this case). Obj will NOT be
// copied to the new one.
//------------------------------------------------------------------------------
template <class vectorType> void
TNumericVectObj<vectorType>::allocateAs(TNumericVectObj<vectorType> *obj) {
  if ((myClass != ocAllocated) && (myClass != ocNull)) {
    // to avoid instability
    printf("TNumericVectObj::allocateAs error!\n This method "
           "can only be used with allocated and null class objects.\n");
    exit(1);
  }

  if (data) {
    delete[] data;
  }

  myClass = ocAllocated;
  dimension = obj->dimension;
  dataSize = dimension * sizeof(vectorType);
  data = new vectorType[dimension];
  readFunc = obj->readFunc;
  metricInUse = obj->metricInUse;

  memset(data, '\0', dataSize);  // initialize with 0
}

//------------------------------------------------------------------------------
// Change the dimension to the new one
//------------------------------------------------------------------------------
template <class vectorType> void
TNumericVectObj<vectorType>::setDim(int dim) {
  if ((myClass != ocAllocated) && (myClass != ocNull)) {
    printf("TNumericVectObj::setDim error! This method "
           "can only be used with allocated and null class objects.\n");
    exit(1);
  }

  if (data) {
    delete[] data;
  }

  myClass = ocAllocated;
  dimension = dim;
  dataSize = dimension * sizeof(vectorType);
  data = new vectorType[dimension];

  memset(data, '\0', dataSize);  // initialize with 0
}

//------------------------------------------------------------------------------
// Change the size (and dimension) to the new one
//------------------------------------------------------------------------------
template <class vectorType> void
TNumericVectObj<vectorType>::setSize(int oSize) {
  if ((myClass != ocAllocated) && (myClass != ocNull)) {
    printf("TNumericVectObj::setDim error! This method "
           "can only be used with allocated and null class objects.\n");
    exit(1);
  }

  if (data) {
    delete[] data;
  }

  myClass = ocAllocated;
  dataSize = oSize;
  dimension = oSize / sizeof(vectorType);
  data = new vectorType[dimension];

  memset(data, '\0', dataSize);  // initialize with 0
}

//------------------------------------------------------------------------------
// Returns the value in position i
//------------------------------------------------------------------------------
template <class vectorType> vectorType&
TNumericVectObj<vectorType>::operator [](int i) {
  if (!data) {
    printf("TNumericVectObj::operator []! Dimension not defined.\n");
    exit(1);
  }

  if ((i < 0) || (i >= dimension)) {
    printf("TNumericVectObj::operator [] error! Access violation.\n");
    exit(1);
  }

  return data[i];
}

//------------------------------------------------------------------------------
// Returns the value in position i
//------------------------------------------------------------------------------
template <class vectorType> const vectorType&
TNumericVectObj<vectorType>::operator [](int i) const {
  if (!data) {
    printf("&TNumericVectObj::operator []! Dimension not defined.\n");
    exit(1);
  }

  if ((i < 0) || (i >= dimension)) {
    printf("TNumericVectObj::operator [] error! Access violation.\n");
    exit(1);
  }

  return data[i];
}

//------------------------------------------------------------------------------
// Sets the function to be used to read objects
//------------------------------------------------------------------------------
template <class vectorType> void
TNumericVectObj<vectorType>::setReadObjectFunction(int (*rdFunc)(char *,
                                                                 int, char *)) {
  if ((myClass != ocAllocated) && (myClass != ocNull)) {
    printf("TNumericVectObj::setReadObjectFunction error! This method "
           "can only be used with allocated and null class objects.\n");
    exit(1);
  }

  readFunc = rdFunc;
}

//------------------------------------------------------------------------------
// Reads an object with the function specified in setReadFunction
//------------------------------------------------------------------------------
template <class vectorType> void
TNumericVectObj<vectorType>::readObj(char *cmdLine, int condition, char *aux) {
  int i;
  vectorType *p;

  if (!readFunc) {
    printf("TNumericVectObj::readObj error! Read function not defined.\n");
    exit(1);
  }

  if (!data) {
    printf("TNumericVectObj::readObj error! Dimension not defined.\n");
    exit(1);
  }

  p = data;
  for(i = 0; i < dimension; i++, p++) {
    switch (myType) {
      case intData:
        (*readFunc)(cmdLine, condition, aux);
        *p = atoi(cmdLine);
      break;
      case floatData:
        (*readFunc)(cmdLine, condition, aux);
        *p = (float) atof(cmdLine);
      break;
      case doubleData: ;
        (*readFunc)(cmdLine, condition, aux);
        *p = atof(cmdLine);
      break;
    }
  }
}

//------------------------------------------------------------------------------
// Generate random values to the object, or get the next biased one
//------------------------------------------------------------------------------
template <class vectorType> void
TNumericVectObj<vectorType>::randomize() {
  vectorType *p, *min, *max;
  int i;

  switch (myClass) {
    case ocSample:
      data = sampleSet[currentSample];
      currentSample = (currentSample + 1) % sampleSetSize;
    break;

    case ocRandom:
      memset(data, '\0', dataSize);  // initialize with 0

      p = data;
      min = minRange;
      max = maxRange;
      for (i = 0; i < dimension; i++, p++, min++, max++)
        *p = (vectorType) ((((float) rand()) / (RAND_MAX + 1)) *
                           (*max - *min) + *min);
    break;

    default:
      printf("TNumericVectObj::randomize error! This is not a query object."
             " Operation not allowed.\n");
      exit(1);
  }
}

//------------------------------------------------------------------------------
// Prints the vector
//------------------------------------------------------------------------------
template <class vectorType> void
TNumericVectObj<vectorType>::printVector() {
  char format[5];
  int i;

  switch (myType) {
    case intData:
      sprintf(format, "%s", "%d ");
    break;
    case floatData:
      sprintf(format, "%s", "%f ");
    break;
    case doubleData:
      sprintf(format, "%s", "%lf ");
    break;
  }

  printf("%d\n", dimension);
  for(i = 0; i < dimension; i++)
    printf(format, data[i]);

  printf("\n");
}

//------------------------------------------------------------------------------
// Reads the vector from a text file
//------------------------------------------------------------------------------
template <class vectorType> void
TNumericVectObj<vectorType>::readFromTxtFile(FILE *fdIn) {
  char format[5];
  int i;

  if (!data) {
    printf("TNumericVectObj::readFromTxtFile error! Object not allocated.\n");
    exit(1);
  }

  switch (myType) {
    case intData:
      sprintf(format, "%s", "%d ");
    break;
    case floatData:
      sprintf(format, "%s", "%f ");
    break;
    case doubleData:
      sprintf(format, "%s", "%lf ");
    break;
  }

  for (i = 0; i < dimension; i++) {
    if (!fscanf(fdIn, format, &(data[i]))) {
      printf("TNumericVectObj::readFromTxtFile error!"
             "Can't read input file.\n");
      exit(1);
    }
  }
}

//------------------------------------------------------------------------------
// Writes the vector to a text file
//------------------------------------------------------------------------------
template <class vectorType> void
TNumericVectObj<vectorType>::writeToTxtFile(FILE *fdOut) {
  char format[5];
  int i;

  if (!data) {
    printf("TNumericVectObj::writeToTxtFile error! Object not allocated.\n");
    exit(1);
  }

  switch (myType) {
    case intData:
      sprintf(format, "%s", "%d ");
    break;
    case floatData:
      sprintf(format, "%s", "%f ");
    break;
    case doubleData:
      sprintf(format, "%s", "%lf ");
    break;
  }

  for (i = 0; i < dimension; i++) {
    if (!fprintf(fdOut, format, data[i])) {
      printf("TNumericVectObj::writeToTxtFile error! Can't write file.\n");
      exit(1);
    }
  }
  fprintf(fdOut, "\n");
}

//------------------------------------------------------------------------------
// Assignment operator
//------------------------------------------------------------------------------
template <class vectorType> TNumericVectObj<vectorType>&
TNumericVectObj<vectorType>::operator =(
                             const TNumericVectObj<vectorType> &cObj) {
  if ((myClass != ocAllocated) && (myClass != ocNull)) {
    printf("TNumericVectObj::operator = error! This method "
           "can only be used with allocated and null class objects.\n");
    exit(1);
  }

  if (this == &cObj)
    return *this;

  if (dimension != cObj.dimension) {
    if (data)
      delete[] data;

    dimension = cObj.dimension;
    dataSize = cObj.dataSize;
    data = new vectorType[dimension];
    myClass = ocAllocated;
  }

  if (dimension)
    memcpy(data, cObj.data, dataSize);

  return *this;
}

//------------------------------------------------------------------------------
// Copies the buffer to object's data space. Buffer must be of the same as the
// object data size (not dimension).
//------------------------------------------------------------------------------
template <class vectorType> void
TNumericVectObj<vectorType>::toObject(char *buff) {
  if ((myClass != ocAllocated) && (myClass != ocMaped)) {
    printf("TNumericVectObj::operator = error! This method "
           "can only be used with allocated and maped class objects.\n");
    exit(1);
  }

  memcpy(data, buff, dataSize);
}

//------------------------------------------------------------------------------
// Copies the object's data space to buffer. Buffer must be of the same as the
// object data size (not dimension).
//------------------------------------------------------------------------------
template <class vectorType> void
TNumericVectObj<vectorType>::toBuffer(char *buff) {
  memcpy(buff, data, dataSize);
}

//------------------------------------------------------------------------------
// Map method - same as map constructor
//------------------------------------------------------------------------------
template <class vectorType> void
TNumericVectObj<vectorType>::mapBuffer(char *buff) {
  if (myClass != ocMaped) {
    printf("TNumericVectObj::operator = error! This method "
           "can only be used with maped class objects.\n");
    exit(1);
  }

  data = (vectorType *)buff;
}

//------------------------------------------------------------------------------
// Returns a pointer to a buffer area containing the object. The user can NEVER
// delete this area. This method is TOO DANGEROUS. Do not modify the buffer.
// Use it only to make buffer copies.
//------------------------------------------------------------------------------
template <class vectorType> char*
TNumericVectObj<vectorType>::bufferArea() {
  if (myClass == ocNull) {
    printf("TNumericVectObj::operator = error! This method "
           "can not be used with null class objects.\n");
    exit(1);
  }

  return ((char *) data);
}

//------------------------------------------------------------------------------
// Calculates the distance between 2 objects
//------------------------------------------------------------------------------
template <class vectorType> TNumericVectObj<vectorType>::distanceType
TNumericVectObj<vectorType>::distance(TNumericVectObj<vectorType> *obj) {
  distanceType result, difference;
  int i;
  vectorType *a, *b;

  if (dimension != obj->dimension) {
    printf("TNumericVectObj::distance error!"
           " Objects are of diferent dimensions.\n");
    exit(1);
  }

  result = 0;
  for (i = 0, a = data, b = obj->data; i < dimension; i++, a++, b++) {
    difference = (distanceType) (*a - *b);
    result += difference * difference;
  }

  return ((distanceType)sqrt(result));
}

//------------------------------------------------------------------------------
// Check if obj is inside the circle (*this + radius). Returns true if it
// is (and its distance), false otherwise (and 0 as the distance).
//------------------------------------------------------------------------------
template <class vectorType> char
TNumericVectObj<vectorType>::isObjectIncluded(TNumericVectObj<vectorType> *obj,
                      TNumericVectObj<vectorType>::distanceType radius,
                      TNumericVectObj<vectorType>::distanceType *dist) {
  char result = true;

  // it must be changed to suport other metrics
  *dist = distance(obj);
  if (*dist > radius)
    result = false;

  return (result);
}

//------------------------------------------------------------------------------
// Initializes the object with all NULL
//------------------------------------------------------------------------------
template <class vectorType> void
TNumericVectObj<vectorType>::initAll() {
  myClass = ocNull;

  dimension = 0;
  dataSize = 0;
  data = NULL;

  readFunc = NULL;

  sampleSet = NULL;
  sampleSetSize = 0;
  currentSample = 0;
  minRange = NULL;
  maxRange = NULL;

  metricInUse = metricL2;

  if(typeid(int) == typeid(vectorType))
    myType = intData;
  else if(typeid(float) == typeid(vectorType))
    myType = floatData;
  else
    myType = doubleData;
}
//---------------------------------------------------------------------------
#endif
