//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Omni-Family
//  File: TStringObj.cpp
//  Classe: TStringObj
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 06/14/2000
//  Version: 0.0
//
//  Description:
//	  Object to be indexed by vector and metric space.
//    String version.
//==============================================================================


#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "StringObj.h"

//------------------------------------------------------------------------------
// Null constructor
//------------------------------------------------------------------------------
TStringObj::TStringObj() {
  initAll();
}

//------------------------------------------------------------------------------
// Main constructor
//------------------------------------------------------------------------------
TStringObj::TStringObj(int sSize) {
  initAll();

  myClass = ocAllocated;
  mySize = sSize;
  data = new char[mySize];

  memset(data, '\0', mySize);  // initialize with null
}

//------------------------------------------------------------------------------
// Constructs and initialize
//------------------------------------------------------------------------------
TStringObj::TStringObj(char *string, int sSize) {
  int i;

  initAll();

  myClass = ocAllocated;
  mySize = sSize;
  data = new char[mySize];

//  memcpy(data, string, mySize);
  for( i = 0; i < mySize; i++)
    data[i] = (char) tolower(string[i]);
}

//------------------------------------------------------------------------------
// Mapping Constructor
// It is faster to be used when it is not desired to
// create a new object. The idea is to share the data space and use the
// internal methods. It is used to avoid expensive buffer copies.
// The user must take care about allocating and deleting the object
// data space. ObjSize can be obtained with another object created by
// the main constructor.
//------------------------------------------------------------------------------
TStringObj::TStringObj(int sSize, char *buff) {
  initAll();

  myClass = ocMaped;
  mySize = sSize;
  data = (char *)buff;
}

//------------------------------------------------------------------------------
// Copy constructor
//------------------------------------------------------------------------------
TStringObj::TStringObj(const TStringObj &cObj) {
  int i;

  if (cObj.myClass != ocAllocated) {
    // to avoid instability
    printf("TStringObj::TStringObj copy constructor error!\n"
           " The copy constructor can be used only with allocated objects.\n");
    exit(1);
  }

  initAll();

  myClass = ocAllocated;
  mySize = cObj.mySize;
  data = new char[mySize];
  readFunc = cObj.readFunc;

//  memcpy(data, cObj.data, mySize);
  for(i = 0; i < mySize; i++)
    data[i] = (char) tolower(cObj.data[i]);
}

//------------------------------------------------------------------------------
// SampleSet constructor (to create a query object of type biased).
// nop is used only to distinguish the constructors
//------------------------------------------------------------------------------
TStringObj::TStringObj(char *fName, int sSize, int nop) {
  FILE *fdIn;
  char line[256];
  int i, j;

  initAll();

  myClass = ocSample;
  mySize = sSize;

  nop++; // just to avoid warnings

  j = strlen(fName);
  for(i = 0; (i < j) && (fName[i] != '.'); i++);
  if (i == j) {
    fName[i++]='.';
    fName[i++]='s';
    fName[i++]='t';
    fName[i++]='c';
    fName[i]=0;
  }

  if ((fdIn = fopen(fName, "rt")) == NULL) {
    printf("TStringObj::TStringObj sampleSet constructor error!\n"
           "Can't open file '%s'.\n", fName);
    exit(1);
  }

  fgets(line, 256, fdIn);
  sscanf(line, "%d", &sampleSetSize);

  sampleSet = new char*[sampleSetSize];
  for (i = 0; i < sampleSetSize; i++)
    sampleSet[i] = new char[mySize];

  for(i = 0; i < sampleSetSize; i++) {
    data = sampleSet[i];
    if (!fscanf(fdIn, "%s ", data)) {
      printf("TStringObj::TStringObj sampleSet constructor error!"
             "Can't read input file '%s'.\n", fName);
      exit(1);
    }
    for(j = 0; j < mySize; j++)
      data[j] = (char) tolower(data[j]);
  }

  currentSample = 0;

  fclose(fdIn);
}

//------------------------------------------------------------------------------
// RandomObject constructor (to create a query object of type random)
//------------------------------------------------------------------------------
TStringObj::TStringObj(int sSize, int value) {
  myClass = ocRandom;
  mySize = sSize;
  data = new char[mySize];
  maxValue = value;
}

//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
TStringObj::~TStringObj() {
  switch (myClass) {
    case ocAllocated:
      if (data) {
        delete[] data;
        data = NULL;
      }
    break;

    case ocSample:
      for(currentSample = 0; currentSample < sampleSetSize; currentSample++) {
        delete[] sampleSet[currentSample];
        sampleSet[currentSample] = NULL;
      }
      delete[] sampleSet;
      sampleSet = NULL;
    break;

    case ocRandom:
      if (data) {
        delete[] data;
        data = NULL;
      }
    break;
  }
}

//------------------------------------------------------------------------------
// Allocate vector
// To be used with the null constructor. The obj parameter is used as an
// example, to get definitions (the mySize in this case). Obj will NOT be
// copied to the new one.
//------------------------------------------------------------------------------
void
TStringObj::allocateAs(TStringObj *obj) {
  if ((myClass != ocAllocated) && (myClass != ocNull)) {
    // to avoid instability
    printf("TStringObj::allocateAs error!\n This method "
           "can only be used with allocated and null class objects.\n");
    exit(1);
  }

  if (data) {
    delete[] data;
  }

  myClass = ocAllocated;
  mySize = obj->mySize;
  data = new char[mySize];
  readFunc = obj->readFunc;

  memset(data, '\0', mySize);  // initialize with 0
}

//------------------------------------------------------------------------------
// Change the mySize to the new one
//------------------------------------------------------------------------------
void
TStringObj::setSize(int sSize) {
  if ((myClass != ocAllocated) && (myClass != ocNull)) {
    printf("TStringObj::setSize error! This method "
           "can only be used with allocated and null class objects.\n");
    exit(1);
  }

  if (data) {
    delete[] data;
  }

  myClass = ocAllocated;
  mySize = sSize;
  data = new char[mySize];

  memset(data, '\0', mySize);  // initialize with 0
}

//------------------------------------------------------------------------------
// Returns the value in position i
//------------------------------------------------------------------------------
char&
TStringObj::operator [](int i) {
  if (!data) {
    printf("TStringObj::operator [] error! mySize not defined.\n");
    exit(1);
  }

  if ((i < 0) || (i >= mySize)) {
    printf("TStringObj::operator [] error! Access violation.\n");
    exit(1);
  }

  return data[i];
}

//------------------------------------------------------------------------------
// Returns the value in position i
//------------------------------------------------------------------------------
const char&
TStringObj::operator [](int i) const {
  if (!data) {
    printf("&TStringObj::operator [] error! mySize not defined.\n");
    exit(1);
  }

  if ((i < 0) || (i >= mySize)) {
    printf("TStringObj::operator [] error! Access violation.\n");
    exit(1);
  }

  return data[i];
}

//------------------------------------------------------------------------------
// Sets the function to be used to read objects
//------------------------------------------------------------------------------
void
TStringObj::setReadObjectFunction(int (*rdFunc)(char *, int, char *)) {
  if ((myClass != ocAllocated) && (myClass != ocNull)) {
    printf("TStringObj::setReadObjectFunction error! This method "
           "can only be used with allocated and null class objects.\n");
    exit(1);
  }

  readFunc = rdFunc;
}

//------------------------------------------------------------------------------
// Reads an object with the function specified in setReadFunction
//------------------------------------------------------------------------------
void
TStringObj::readObj(char *cmdLine, int condition, char *aux) {
  int i;

  if (!readFunc) {
    printf("TStringObj::readObj error! Read function not defined.\n");
    exit(1);
  }

  if (!data) {
    printf("TStringObj::readObj error! mySize not defined.\n");
    exit(1);
  }

  (*readFunc)(cmdLine, condition, aux);
//  memcpy(data, cmdLine, mySize);
  for( i = 0; i < mySize; i++)
    data[i] = (char) tolower(cmdLine[i]);
}

//------------------------------------------------------------------------------
// Generate random values to the object, or get the next biased one
//------------------------------------------------------------------------------
void
TStringObj::randomize() {
  char *p;
  int i, n;

  switch (myClass) {
    case ocSample:
      data = sampleSet[currentSample];
      currentSample = (currentSample + 1) % sampleSetSize;
    break;

    case ocRandom:
      memset(data, '\0', mySize);  // initialize with 0

      n = (int)((((float) rand()) / (RAND_MAX + 1)) * maxValue);
      for (i = 0; i < n; i++, p++)
           *p = (char)((((float) rand()) / (RAND_MAX + 1)) * 26 + 'a');
    break;

    default:
      printf("TStringObj::randomize error! This is not a query object."
             " Operation not allowed.\n");
      exit(1);
  }
}

//------------------------------------------------------------------------------
// Prints the string
//------------------------------------------------------------------------------
void
TStringObj::printString() {
  if (!data) {
    printf("TStringObj::printString error! Object not allocated.\n");
    exit(1);
  }

  printf("%s\n", data);
}

//------------------------------------------------------------------------------
// Reads the string from a text file
//------------------------------------------------------------------------------
void
TStringObj::readFromTxtFile(FILE *fdIn) {
  int i;

  if (!data) {
    printf("TStringObj::readFromTxtFile error! Object not allocated.\n");
    exit(1);
  }

  if (!fscanf(fdIn, "%s ", data)) {
    printf("TStringObj::readFromTxtFile error! Can't read input file.\n");
    exit(1);
  }
  for(i = 0; i < mySize; i++)
    data[i] = (char) tolower(data[i]);
}

//------------------------------------------------------------------------------
// Writes the string to a text file
//------------------------------------------------------------------------------
void
TStringObj::writeToTxtFile(FILE *fdOut) {
  if (!data) {
    printf("TStringObj::writeToTxtFile error! Object not allocated.\n");
    exit(1);
  }

  fprintf(fdOut, "%s\n", data);
}

//------------------------------------------------------------------------------
// Assignment operator
//------------------------------------------------------------------------------
TStringObj&
TStringObj::operator =(const TStringObj &cObj) {
  int i;

  if ((myClass != ocAllocated) && (myClass != ocNull)) {
    printf("TStringObj::operator = error! This method "
           "can only be used with allocated and null class objects.\n");
    exit(1);
  }

  if (this == &cObj)
    return *this;

  if (mySize != cObj.mySize) {
    if (data)
      delete[] data;

    mySize = cObj.mySize;
    data = new char[mySize];
    myClass = ocAllocated;
  }


  if (mySize) {
//    memcpy(data, cObj.data, mySize);
    for( i = 0; i < mySize; i++)
      data[i] = (char) tolower(cObj.data[i]);
  }

  return *this;
}

//------------------------------------------------------------------------------
// Copies the buffer to object's data space. Buffer must be of the same as the
// object data size.
//------------------------------------------------------------------------------
void
TStringObj::toObject(char *buff) {
  int i;

  if ((myClass != ocAllocated) && (myClass != ocMaped)) {
    printf("TStringObj::toObject error! This method "
           "can only be used with allocated and maped class objects.\n");
    exit(1);
  }

//  memcpy(data, buff, mySize);
  for( i = 0; i < mySize; i++)
    data[i] = (char) tolower(buff[i]);
}

//------------------------------------------------------------------------------
// Copies the object's data space to buffer. Buffer must be of the same as the
// object data size.
//------------------------------------------------------------------------------
void
TStringObj::toBuffer(char *buff) {
  memcpy(buff, data, mySize);
}

//------------------------------------------------------------------------------
// Map method - same as map constructor
//------------------------------------------------------------------------------
void
TStringObj::mapBuffer(char *buff) {
  if (myClass != ocMaped) {
    printf("TNumericVectObj::mapBuffer error! This method "
           "can only be used with maped class objects.\n");
    exit(1);
  }

  data = buff;
}

//------------------------------------------------------------------------------
// Returns a pointer to a buffer area containing the object. The user can NEVER
// delete this area. This method is TOO DANGEROUS. Do not modify the buffer.
// Use it only to make buffer copies.
//------------------------------------------------------------------------------
char*
TStringObj::bufferArea() {
  if (myClass == ocNull) {
    printf("TStringObj::bufferArea error! This method "
           "can not be used with null class objects.\n");
    exit(1);
  }

  return (data);
}

//------------------------------------------------------------------------------
// Calculates the distance between 2 objects
//------------------------------------------------------------------------------
TStringObj::distanceType
TStringObj::distance(TStringObj *obj) {
  return wordDiff(data, obj->data);
}

//------------------------------------------------------------------------------
// Check if obj is inside the circle (*this + radius). Returns true if it
// is (and its distance), false otherwise (and 0 as the distance).
//------------------------------------------------------------------------------
char
TStringObj::isObjectIncluded(TStringObj *obj, TStringObj::distanceType radius,
                                               TStringObj::distanceType *dist) {
  char result = true;

  // it must be changed to suport other metrics
  *dist = distance(obj);
  if (*dist > radius)
    result = false;

  return (result);
}

//------------------------------------------------------------------------------
// Initializes the object with all NULL
//------------------------------------------------------------------------------
void
TStringObj::initAll() {
  myClass = ocNull;

  mySize = 0;
  data = NULL;

  readFunc = NULL;

  sampleSet = NULL;
  sampleSetSize = 0;
  currentSample = 0;
  maxValue = 0;
}

//------------------------------------------------------------------------------
// Calculates the minimum between 3 int
//------------------------------------------------------------------------------
int
TStringObj::min3(int a, int b, int c) {
  if ((a <= b) && (a <= c))
    return(a);
  else if (b <= c)
    return(b);
  else
    return(c);
}

//------------------------------------------------------------------------------
// Original LEdit metric
//------------------------------------------------------------------------------
TStringObj::distanceType
TStringObj::wordDiff(char *s1, char *s2) {
  int lens, lenl;
  int i, j, temp;
  int diff1[80][80];
  char *w1, *w2;

  /* worddiffcalcount++; */
  lens = strlen(s1);
  lenl = strlen(s2);
  w1 = s1;
  w2 = s2;

  if (lens > lenl) {
    temp = lens;
    lens = lenl;
    lenl = temp;
    w1 = s2;
    w2 = s1;
  }

  if ((lens >= 50) || (lenl >= 50)) {
    printf("Word too long\n");
    exit(1);
  }

  for (i=0; i <= lens;  i++)
    diff1[i][0] = i;
  for (i=1; i <= lenl;  i++)
    diff1[0][i] = i;

  for (i = 1; i <= lens; i++) {
    for (j = 1; j <= i; j++)
      diff1[i-j+1][ j]  = min3(diff1[i-j][ j-1] + (w1[i-j] != w2[j-1]),
                               diff1[i-j][ j] + 1,
                               diff1[i-j+1][ j-1] + 1);
  }

  for (i = 1; i <= lenl - lens - 1; i++) {
    for (j = 1; j <= lens; j++)
      diff1[lens-j+1][ i + j]  =
                     min3(diff1[lens-j][ i+j-1] + (w1[lens-j] != w2[i + j-1]),
                          diff1[lens-j][ i+j] + 1,
                          diff1[lens-j+1][ i+j-1] + 1);
      }

  for (i = 1; i <= lens; i++) {
    for (j = 0; j <= lens - i; j++)
      diff1[lens-j][ lenl- lens + j + i]  =
                     min3(diff1[lens-j - 1][ lenl - lens + i+j-1] +
	                              (w1[lens-j - 1] != w2[lenl - lens + i + j-1]),
                          diff1[lens-j - 1][ lenl - lens + i+j] + 1,
                          diff1[lens-j][ lenl - lens + i+j-1] + 1);
  }
  return((distanceType)diff1[lens][lenl]);
}

