//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Buffer manager for spatial/metric indexing methods
//  File: TBuffer.cpp
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 06/14/2000
//  Version: 0.01
//
//  Description:
//	  Object to manage a buffer file. The options include open, close and
//  create files and read, write (change) and create buffers.
//
//    See header file for more details...
//
//  Modifications:
//    Author:
//    Date:
//    Description:
//==============================================================================

#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdio.h>

//---------------------------------------------------------------------------
#pragma hdrstop

#include "TBuffer.h"
//------------------------------------------------------------------------------
// Null constructor
//------------------------------------------------------------------------------
TBuffer::TBuffer() {
  zeroAll(); // initalize all variables with zero or NULL
}

//------------------------------------------------------------------------------
// Open an existing file. The file fName MUST exists
//------------------------------------------------------------------------------
TBuffer::TBuffer(char *fName) {
  zeroAll(); // initalize all variables with zero or NULL

  openFile(fName);
}

//------------------------------------------------------------------------------
// create and open a new file (ovewrites the old one)
//------------------------------------------------------------------------------
TBuffer::TBuffer(const char *fName, int pageSize, int bufferSize,
                   int userHeaderSize, int cacheNPages) {
  zeroAll(); // initalize all variables with zero or NULL

  createFile(fName, pageSize, bufferSize, userHeaderSize, cacheNPages);
}

//------------------------------------------------------------------------------
//Destructor
//------------------------------------------------------------------------------
TBuffer::~TBuffer() {
  if (fd >= 0) // The file still opened - probably, the user's program has a
          // logic error, thats why the program will be aborted.
    printErrorMsgNAbort("~TBuffer error! The file still opened.");
  // if (fd >= 0)
}

//------------------------------------------------------------------------------
// Open an existing file. The file fName MUST exists. This object CAN'T have
// another file opened.
//------------------------------------------------------------------------------
void
TBuffer::openFile(char *fName) {
  if (fd >= 0) // another file is opened - probably, the user's program has a
          // logic error, thats why the program will be aborted.
    printErrorMsgNAbort("openFile error! There is another file opened.");
  // if (fd >= 0)

  strcpy(fileName, fName);

  fd = open(fileName, O_RDWR | O_BINARY);
  if (fd < 0)
    printErrorMsgNAbort("TBuffer error! Unable to open file.");
  // if (fd < 0)

  readHeader();
  initCache();

  firstBufferDiskAddress = TBuffHEADERSIZE + fileHeader.userHeaderSize;

  pagesPerBuffer = fileHeader.bufferSize / fileHeader.pageSize;

  if ((pagesPerBuffer * fileHeader.pageSize) < fileHeader.bufferSize)
    pagesPerBuffer++; // it is the same as ceil(), but doesn't requires casting
  // if ((pagesPerBuffer * fileHeader.pageSize) < fileHeader.bufferSize)
}

//------------------------------------------------------------------------------
// Close the actual file - must have one opened
// IMPORTANT: the user is responsible to write his own header
// The user MUST read the counter before closing the file (to guaranty
// consistence). If te user is interested in the cache performance and wants
// to consider all disk writes, it must call flushThis() before colecting
// the counters.
//------------------------------------------------------------------------------
void
TBuffer::closeThis() {
  if (fd < 0) // there is no file to be closed - probably, the user's program
          // has a logic error, thats why the program will be aborted.
    printErrorMsgNAbort("closeThis error! There is no file opened.");
  // if (fd < 0)

  // the order of the folowing commands must be maintained
  delete cache;
  writeHeader();
  close(fd);
//  fd = -1;

  zeroAll(); // initalize all variables with zero or NULL
}

//------------------------------------------------------------------------------
// Create and open a new file (overwrites the old one). This object CAN'T
// have another file opened.
// Required values:
//   pageSize >= 64
//   bufferSize >= 64
//   userHeaderSize >= 0
//   cacheNPages >= 5 (default = 100)
//------------------------------------------------------------------------------
void
TBuffer::createFile(const char *fName, int pageSize, int bufferSize,
                                    int userHeaderSize, int cacheNPages) {
  if (fd >= 0) // another file is opened
    printErrorMsgNAbort("createFile error! There is another file opened.");
  // if (fd >= 0)

  strcpy(fileName, fName);
  fd = open(fileName, O_CREAT | O_TRUNC | O_RDWR | O_BINARY,
            S_IREAD | S_IWRITE);
  if (fd < 0)
    printErrorMsgNAbort("createFile error! Unable to create file.");
  // if (fd < 0)

  if ((pageSize < 64) || (bufferSize < 64) || (userHeaderSize < 0) ||
      (cacheNPages < 5))
    printErrorMsgNAbort("createFile error! Invalid parameters.");
  // if (fd < 0)

  fileHeader.bufferSize = bufferSize;
  fileHeader.pageSize = pageSize;
  fileHeader.cacheTotalPages = cacheNPages;
  fileHeader.totalBuffersInUse = 0;
  fileHeader.disposeList = 0; // the disposeList is buid with disk address, not
                              // bufferId. Since the header is located at
                              // position 0 of the file, there is no problem
                              // on using it as invalid address
  fileHeader.totalBuffersInDisposeList = 0;
  fileHeader.userHeaderSize = userHeaderSize;

  writeHeader();

  firstBufferDiskAddress = TBuffHEADERSIZE + fileHeader.userHeaderSize;

  initCache();

  pagesPerBuffer = fileHeader.bufferSize / fileHeader.pageSize;

  if ((pagesPerBuffer * fileHeader.pageSize) < fileHeader.bufferSize)
    pagesPerBuffer++; // it is the same as ceil(), but doesn't requires casting
  // if ((pagesPerBuffer * fileHeader.pageSize) < fileHeader.bufferSize)
}

//------------------------------------------------------------------------------
// Close the actual file and reopen overwriting all data. There must be an
// opened file. The user header and all data will be destroyed. The user must
// save his header before calling this method (if he needs to save it).
// Only the original characteristics (PageSize, buferSize, userHeaderSize and
// cacheTotalPages) are maintained.
//------------------------------------------------------------------------------
void
TBuffer::dropThis() {
  tFileHeader auxHeader;
  char fName[TBuffMAXFNAMELENGTH];

  if (fd < 0) // there is no file to be droped
    printErrorMsgNAbort("dropThis error! There is no file opened.");
  // if (fd < 0)

  auxHeader = fileHeader;
  strcpy(fName, fileName);

  closeThis();

  createFile(fName, auxHeader.pageSize, auxHeader.bufferSize,
             auxHeader.userHeaderSize, auxHeader.cacheTotalPages);
}

//------------------------------------------------------------------------------
// Returns only the size occupied by the user. It includes the user header and
// buffers in use. Tt DOES NOT counts the system header and disposed buffers.
//------------------------------------------------------------------------------
long
TBuffer::getUserFileSize() {
  return (getFileSize() - (TBuffHEADERSIZE +
              (fileHeader.totalBuffersInDisposeList * fileHeader.bufferSize)));
}
//------------------------------------------------------------------------------
// Read user header
//------------------------------------------------------------------------------
void
TBuffer::readUserHeader(void *buffer) {
  if (fd < 0) // there is no file opened
    printErrorMsgNAbort("readUserHeader error! There is no file opened.");
  // if (fd < 0)

  lseek(fd, TBuffHEADERSIZE, SEEK_SET);
  if (read(fd, buffer, fileHeader.userHeaderSize) != fileHeader.userHeaderSize)
    printErrorMsgNAbort("readUserHeader error! Error reading user's header.");
  // if (read(fd, buffer, fileHeader.userHeaderSize)....

  #ifdef TBuffSTATISTICS
    counters.headerRead = (counters.headerRead + 1) % TBuffCOUNTERSMAXVALUE;
  #endif
}

//------------------------------------------------------------------------------
// Write user header
//------------------------------------------------------------------------------
void
TBuffer::writeUserHeader(void *buffer) {
  if (fd < 0) // there is no file opened
    printErrorMsgNAbort("writeUserHeader error! There is no file opened.");
  // if (fd < 0)

  lseek(fd, TBuffHEADERSIZE, SEEK_SET);
  if (write(fd, buffer, fileHeader.userHeaderSize) !=
                                                     fileHeader.userHeaderSize)
    printErrorMsgNAbort("writeUserHeader error! Error writing user's header.");
  // if (write(fd, buffer, fileHeader.userHeaderSize)....

  #ifdef TBuffSTATISTICS
    counters.headerWrite = (counters.headerWrite + 1) % TBuffCOUNTERSMAXVALUE;
  #endif
}

//------------------------------------------------------------------------------
// Reads a buffer from cache
//------------------------------------------------------------------------------
void
TBuffer::readBuffer(long bufferId, void *buffer) {
  void *auxBuffer;

  if (fd < 0) // there is no file opened
    printErrorMsgNAbort("readBuffer error! There is no file opened.");
  // if (fd < 0)

  cache->readBuffer(bufferId, &auxBuffer);

  memcpy(buffer, auxBuffer, fileHeader.bufferSize);

  #ifdef TBuffSTATISTICS
    counters.bufferRead = (counters.bufferRead + 1) % TBuffCOUNTERSMAXVALUE;
  #endif
}

//------------------------------------------------------------------------------
// Writes the buffer into cache
//------------------------------------------------------------------------------
void
TBuffer::writeBuffer(long bufferId, void *buffer) {
  void *auxBuffer;

  if (fd < 0) // there is no file opened
    printErrorMsgNAbort("writeBuffer error! There is no file opened.");
  // if (fd < 0)

  cache->readBuffer(bufferId, &auxBuffer); // brings the buffer to cache
  memcpy(auxBuffer, buffer, fileHeader.bufferSize); // writes the buffer

  cache->writeBuffer(); // set write flag on cache
  #ifdef TBuffSTATISTICS
    counters.bufferWrite = (counters.bufferWrite + 1) % TBuffCOUNTERSMAXVALUE;
  #endif
}

//------------------------------------------------------------------------------
// Creates a new buffer and returns the id
//------------------------------------------------------------------------------
long
TBuffer::insertNewBuffer(void *buffer) {
  long bufferId, *pAuxId;
  void *auxBuffer;

  if (fd < 0) // there is no file opened
    printErrorMsgNAbort("newBuffer error! There is no file opened.");
  // if (fd < 0)

  if (fileHeader.disposeList) { // test if disposed list is not empty
    bufferId = toBufferId(fileHeader.disposeList);

    cache->readBuffer(bufferId, &auxBuffer); // bring the new buffer to cache

    pAuxId = (long *) auxBuffer;
    fileHeader.disposeList = *pAuxId; // remove buffer from dipose list
    fileHeader.totalBuffersInUse++;
    fileHeader.totalBuffersInDisposeList--;

    memcpy(auxBuffer, buffer, fileHeader.bufferSize); // write the buffer
    cache->writeBuffer(); // set write flag on cache
  } // if (fileHeader.disposeList)
  else {
    bufferId = getTotalBuffersIncludingDisposed();
    fileHeader.totalBuffersInUse++; // it MUST be here, otherwise bufferId will
                                    // be invalid to the method diskWriteBuff

    diskWriteBuff(bufferId, buffer); // write the new buffer and expand file

    cache->newBuffer(bufferId, &auxBuffer); // adjust cache and bring new
                                            // (blank) buffer
  } // else -> if (fileHeader.disposeList)

  writeHeader();

  #ifdef TBuffSTATISTICS
    counters.bufferWrite = (counters.bufferWrite + 1) % TBuffCOUNTERSMAXVALUE;
    counters.newBuffer = (counters.newBuffer + 1) % TBuffCOUNTERSMAXVALUE;
  #endif

  return bufferId;
}

//------------------------------------------------------------------------------
// Insert the buffer in the dispose list
//------------------------------------------------------------------------------
void
TBuffer::freeBuffer(long bufferId) {
  long *pAuxId;
  void *auxBuffer;

  if (fd < 0) // there is no file opened
    printErrorMsgNAbort("freeBuffer error! There is no file opened.");
  // if (fd < 0)

  cache->readBuffer(bufferId, &auxBuffer); // bring buffer to cache

  // insert buffer in dispose list
  pAuxId = (long *) auxBuffer;
  *pAuxId = fileHeader.disposeList;
  fileHeader.disposeList = toFileCursor(bufferId);

  fileHeader.totalBuffersInDisposeList++;
  fileHeader.totalBuffersInUse--;
  writeHeader();

  cache->writeBuffer(); // set write flag on cache

  #ifdef TBuffSTATISTICS
    counters.bufferWrite = (counters.bufferWrite + 1) % TBuffCOUNTERSMAXVALUE;
  #endif
}

//------------------------------------------------------------------------------
// Reinstantiate the cache with newCachePages number of pages and makes it
// default for future use
//------------------------------------------------------------------------------
void
TBuffer::resizeCache(int newCacheNPages) {
  if (fd < 0) // there is no file opened
    printErrorMsgNAbort("resizeCache error! There is no file opened.");
  // if (fd < 0)

  delete cache;
  fileHeader.cacheTotalPages = newCacheNPages;
  writeHeader();
  initCache();
}

//------------------------------------------------------------------------------
// Read system header
//------------------------------------------------------------------------------
void
TBuffer::readHeader() {
  lseek(fd, 0, SEEK_SET);
  if (read(fd, &fileHeader, TBuffHEADERSIZE) != TBuffHEADERSIZE)
    printErrorMsgNAbort("readHeader error! Error reading system header.");
  // if (read(fd, &fileHeader, HEADERSIZE) != HEADERSIZE)
}

//------------------------------------------------------------------------------
// Write system header
//------------------------------------------------------------------------------
void
TBuffer::writeHeader() {
  lseek(fd, 0, SEEK_SET);
  if (write(fd, &fileHeader, TBuffHEADERSIZE) != TBuffHEADERSIZE)
    printErrorMsgNAbort("writeHeader error! Error writing system header.");
  // if (write(fd, &fileHeader, HEADERSIZE) != HEADERSIZE)
}

//------------------------------------------------------------------------------
// starts the cache manager
//------------------------------------------------------------------------------
void
TBuffer::initCache() {
  cache = new TNSCache<class TBuffer>(this, &TBuffer::diskReadBuff,
            &TBuffer::diskWriteBuff, getTotalBuffersIncludingDisposed(),
            FROT, fileHeader.cacheTotalPages, fileHeader.bufferSize);
  // the cache manages user buffers and not file pages
}

//------------------------------------------------------------------------------
// Used by the cache to read a buffer from disk
//------------------------------------------------------------------------------
void
TBuffer::diskReadBuff(long bufferId, void *buffer) {
  int i;
  char *bufferCursor;

  if (bufferId >= getTotalBuffersIncludingDisposed())
    printErrorMsgNAbort("diskReadBuff error! Invalid bufferId.");
  // if (bufferId >= getTotalBuffersIncludingDisposed())

  lseek(fd, toFileCursor(bufferId), SEEK_SET);
  bufferCursor = (char *) buffer;
  for (i = 1; i <= pagesPerBuffer; i++) {
    if (i == pagesPerBuffer) // reads the last part of buffer. Since it can be
                             // less than pageSize, the amount of bytes to be
                             // read must be calculated. To verify if it is the
                             // last part, instead of making a subtraction on
                             // every loop (i == pagesPerBuffer - 1), the for
                             // counter (i) must start with 1
      read(fd, bufferCursor, fileHeader.bufferSize -
                             ((i - 1) * fileHeader.pageSize));
    // if (i == pagesPerBuffer)
    else
      read(fd, bufferCursor, fileHeader.pageSize);
    // else
    bufferCursor += fileHeader.pageSize;
  } // for

  #ifdef TBuffSTATISTICS
    counters.diskRead = (counters.diskRead + 1) % TBuffCOUNTERSMAXVALUE;
  #endif
}

//------------------------------------------------------------------------------
// Used by the cache to write a buffer to disk
//------------------------------------------------------------------------------
void
TBuffer::diskWriteBuff(long bufferId, void *buffer) {
  int i;
  char *bufferCursor;

  if (bufferId >= getTotalBuffersIncludingDisposed())
    printErrorMsgNAbort("diskWriteBuff error! Invalid bufferId.");
  // if (bufferId >= getTotalBuffersIncludingDisposed())

  lseek(fd, toFileCursor(bufferId), SEEK_SET);
  bufferCursor = (char *) buffer;
  for (i = 1; i <= pagesPerBuffer; i++) {
    if (i == pagesPerBuffer) // writes the last part of buffer. Since it can be
                             // less than pageSize, the amount of bytes to be
                             // read must be calculated. To verify if it is the
                             // last part, instead of making a subtraction on
                             // every loop (i == pagesPerBuffer - 1), the for
                             // counter (i) must start with 1
      write(fd, bufferCursor, fileHeader.bufferSize -
                             ((i - 1) * fileHeader.pageSize));
    // if (i == pagesPerBuffer)
    else
      write(fd, bufferCursor, fileHeader.pageSize);
    // else
    bufferCursor += fileHeader.pageSize;
  } // for

  #ifdef TBuffSTATISTICS
    counters.diskWrite = (counters.diskWrite + 1) % TBuffCOUNTERSMAXVALUE;
  #endif
}

//------------------------------------------------------------------------------
// Print an error message and aborts the program
//------------------------------------------------------------------------------
void
TBuffer::printErrorMsgNAbort(char *errorMsg) {
  if (fd >= 0) // test if the file still opened
    close(fd);

  printf("TBuffer::%s\n", errorMsg);
  exit(1);
}

//------------------------------------------------------------------------------
// initalize all variables with zero or NULL
//------------------------------------------------------------------------------
void
TBuffer::zeroAll(){
  strcpy(fileHeader.fileVersion, TBuffVERSION);
  fileHeader.bufferSize =
  fileHeader.pageSize =
  fileHeader.totalBuffersInUse =
  fileHeader.disposeList =
  fileHeader.totalBuffersInDisposeList =
  fileHeader.userHeaderSize = 0;
  fileHeader.cacheTotalPages = 100;

  *fileName = '\0';

  fd = -1; // fd is integer. Zero and NULL doesn't work. -1 is the invalid value

  firstBufferDiskAddress = TBuffHEADERSIZE;

  cache = NULL;

  pagesPerBuffer = 0;

  #ifdef TBuffSTATISTICS
    resetStatistics();
  #endif
}

