//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Buffer manager for spatial/metric indexing methods
//  File: TBuffer.h
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 06/14/2000
//  Version: 0.01
//
//  Description:
//	  Object to manage a buffer file. The options include open, close and
//  create files and read, write (change) and create buffers. The disk page
//  does not need to be the same size of the buffer. If bufferSize > pageSize,
//  more than one page are used to store the buffer. If bufferSize <= pageSize,
//  exactly one page is used (even if a page can store more than one buffer).
//  Only bufferSize bytes are written and read by the user. So, the user doesn't
//  need to be worried about the pageSize. If the bufferSize < pageSize, the
//  user won't need to allocate pageSize bytes to interacts with this object.
//
//    The user uses the entire page. It means that no one internal (control)
//  data is written in the page. Only when a page is disposed an internal data
//  is written (a pointer to the next disposed page or null). Since absolutely
//  not one single test is performed to verify if the page is a disposed one
//  (because it is too expensive) it's up to the user to avoid this kind of
//  problem. If the user wants to make some logical mark to indicates that
//  it is a disposed page, it can use any position after the 20th byte
//  (actually, it should be after the 4th, but I decided to reserve some more
//  just in case...). It must be observed that, if the user writes on a
//  disposed page, the data have great chance to be lost.
//
//    About header... Since 1) the disk access counter consider every buffer
//  access and 2) the idea is to use this object to aid on developing and
//  evaluating the performance of indexing methods, it would make no sense on
//  counting the disk accesses to the index header. So, two methods were
//  implemented (readHeader and writeHeader) which can be used to access the
//  user header without counting as a disk access (the header access has
//  it own counters - see below, counters definition, for more details). It is
//  not necessary to have the same size for the header and the buffer. The user
//  must inform this size on creation time. The size CAN'T be changed. So, it is
//  a good idea to define a size bigger than what is actually in use. Since
//  there is only one header per file, there will be no header id. Only this
//  object knows where the header is located and the user can access it only
//  through the two commented methods.
//
//    The intention with this object is only to help on building indexing
//  methods for spatial and metric data, where the data to be stored can be so
//  big that it will not fit in only one disk page. Yet, the idea is to use this
//  as an academic tool for testing and tuning indexing methods, and not as a
//  commercial one. So, it is not RELIABLE! Don't expect too much.
//
//    To aid the disk access, a cache manager is used. Its page replacement
//  policy is the Natural Selection Page Replacement Policy (a mixing between
//  the LRU and LFU replacement policies).
//
//  Modifications:
//    Author:
//    Date:
//    Description:
//==============================================================================

#ifndef __TBufferH
#define __TBufferH

#include <stdlib.h>

// ---- C++ Builder ----
#if defined(__BORLANDC__)
  #include <io.h>
  #include <values.h>
#endif

// ---- GNU GCC ----
#ifdef __GNUC__
  #include <unistd.h>
  #include <values.h>
  #ifndef O_BINARY
    #define O_BINARY 0
    // Unix treats binary and text files by the same way. So, it doesn't have
    // the O_BINARY open mode.
    // To be used with Unix, O_BINARY is defined as 0 because it doesn't change
    // the value off '|' operator.
  #endif
#endif

// ---- Visual C++ ----
#ifdef _MSC_VER
  #include <io.h>
  #include <limits.h>
  #define MAXLONG LONG_MAX
#endif

//---- Constant definitions ----
#define TBuffMAXFNAMELENGTH 256
#define TBuffSTATISTICS
#define TBuffHEADERRESERVED 100
#define TBuffCOUNTERSMAXVALUE (MAXLONG - 100)

#include "TNSCache.h"
template <class TpObjCntin> class TNSCache;

//------------------------------------------------------------------------------
// Class TBuffer definitions
//------------------------------------------------------------------------------
class TBuffer {
  public:
    //---------------------------------------------
    // Constructors and destructor methods
    //---------------------------------------------
    TBuffer(); // null constructor - just initializes variables
    TBuffer(char *fName); // open an existing file. The file fName MUST exists

    // create and open a new file (ovewrites the old one)
    TBuffer(const char *fName, int pageSize, int bufferSize, int userHeaderSize,
                                                         int cacheNPages = 100);
    ~TBuffer();

    //---------------------------------------------
    // File methods
    //---------------------------------------------
    // Open an existing file. The file fName MUST exists. This object CAN'T
    // have another file opened.
    void openFile(char *fName);
    void closeThis(); // close the actual file - must have one opened

    // create and open a new file (overwrites the old one). This object CAN'T
    // have another file opened.
    void createFile(const char *fName, int pageSize, int bufferSize,
                                     int userHeaderSize, int cacheNPages = 100);
    char isOpened() { // returns 1 if file is opened, 0 otherwise
      return (fd >= 0);
    }

    // Close the actual file and reopen overwriting all data. There must be an
    // opened file
    void dropThis();

   // Returns the version of the buffer manager
    float getFileVersion() {
      return ((float)atof(&fileHeader.fileVersion[14]));
    }

    // Returns the size of the file including headers and records in dispose
    // list (if it is not empty)
    long getFileSize() {
      return lseek(fd, 0, SEEK_END);
    }

    // Returns only the size occupied by the user. It includes the user header
    // and buffers in use. It DOES NOT counts the system header and disposed
    // buffers.
    long getUserFileSize();

    //---------------------------------------------
    // User's header methods. The disk is directly
    // accessed - no cache operation
    //---------------------------------------------
    void readUserHeader(void *buffer); // read user header
    void writeUserHeader(void *buffer); // write user header

    //---------------------------------------------
    // Buffer methods
    //---------------------------------------------
    void readBuffer(long bufferId, void *buffer); // reads a buffer from cache
    // writes the buffer into cache
    void writeBuffer(long bufferId, void *buffer);
    long insertNewBuffer(void *buffer); // create a new buffer and return the id
    void freeBuffer(long bufferId); // insert the buffer in the dispose list

    //---------------------------------------------
    // Cache methods
    //---------------------------------------------
    void flushThis() { // flush the cache
  //    cache->flushCache();
    }

    int cacheTotalPages() { // return the number of pages in cache
      return fileHeader.cacheTotalPages;
    }

    // reinstantiate the cache with newCachePages number of pages and makes it
    // default for future use
    void resizeCache(int newCacheNPages);

    //---------------------------------------------
    // General methods
    //---------------------------------------------
    int getBufferSize() { // return the buffer size in use
      return fileHeader.bufferSize;
    }

    int getPageSize() { // return the page size in use
      return fileHeader.pageSize;
    }

    // Return the number of buffers in use it ignores the disposed ones
    long getTotalBuffersInUse() {
      return fileHeader.totalBuffersInUse;
    }

    // Return the total number of buffers, including disposed
    long getTotalBuffersIncludingDisposed() {
      return (fileHeader.totalBuffersInUse +
              fileHeader.totalBuffersInDisposeList);
    }

    // Return the total number of pages in use it ignores the disposed ones
    long getTotalPagesInUse() {
      return (getTotalBuffersInUse() * pagesPerBuffer);
    }

    // Return the total number of pages, including disposed
    long getTotalPagesIncludingDisposed() {
      return (getTotalBuffersIncludingDisposed() * pagesPerBuffer);
    }

    //---------------------------------------------
    // Statistics Related - When activated
    //---------------------------------------------
    #ifdef TBuffSTATISTICS
      // Counters for performance meter
      struct performanceCounters {
        long headerRead;  // read user's header counter
        long headerWrite; // write user's header counter
        long bufferRead;  // read buffer counter
        long bufferWrite; // write buffer counter - new buffers are counted as
                          // buffer write too
        long newBuffer;   // new buffer counter
        long relativeTotalDiskAccess; // considering that the number of pages
             // per buffer may be greater than 1, this counter contains the
             // relative number of disk accesses considering this number. So,
             // it will be (bufferRead + bufferWrite) * pagesPerBuffer.
             // This is the counter to be used by indexing methods evaluate
             // their performance. It considers ALL accesses, even if the page
             // was found in the cache.
        long relativeTotalDiskRead; // same as above, but considering only read
                                    // access
        long relativeTotalDiskWrite; // same as above, but considering only
                                    // write access
        // the next counters are incremented only when a cache fault occurs
        // it values reflects the actual number of accessed pages. If
        // bufferSize > pageSize, it will count the number of pages accessed to
        // retrieve an entire buffer. To measure the cache efficiency, these
        // counters must be compared with the relativeTotalDiskAccess counter.
        long diskRead;  // read disk counter
        long diskWrite; // write disk counter - new pages are counted as disk
                        // write too
      };
      performanceCounters counters;

      performanceCounters getStatistics() {
        counters.relativeTotalDiskRead = counters.bufferRead * pagesPerBuffer;
        counters.relativeTotalDiskWrite = counters.bufferWrite * pagesPerBuffer;
        counters.relativeTotalDiskAccess = counters.relativeTotalDiskRead +
                                           counters.relativeTotalDiskWrite;

        return counters;
      }
      void resetStatistics() {
        counters.headerRead =
        counters.headerWrite =
        counters.bufferRead =
        counters.bufferWrite =
        counters.newBuffer =
        counters.relativeTotalDiskAccess =
        counters.relativeTotalDiskRead =
        counters.relativeTotalDiskWrite =
        counters.diskRead =
        counters.diskWrite = 0;
      }
    #endif

  private:
    //---------------------------------------------
    // Constants and types
    //---------------------------------------------
    #define FROT 4  // 16 records per block - cache attribute
    #define TBuffVERSION "BFMUGA06/2000-0.01\0" // file version - used to check
                               // if the file was built in the same version
                               // the first 6 characters indicates that it is
                               // a file created by TBuffer object. Than, comes
                               // the month/year-version number of the object.

    struct tFileHeader {
      char fileVersion[20];  // version of the TBuffer object used to create it
      int bufferSize; // size of each buffer - defined in creation time
      int pageSize; // size of each page - defined in creation time
      int cacheTotalPages; // number of pages in cache
      long totalBuffersInUse; // total number of valid buffers
      long disposeList; // lists of the disposed records
      long totalBuffersInDisposeList; // total number of disposed buffers
      long userHeaderSize; // size of the user's header
      char reserved[TBuffHEADERRESERVED]; // reservs space for future use
    };
    #define TBuffHEADERSIZE sizeof(tFileHeader)

    //---------------------------------------------
    // Variables
    //---------------------------------------------
    char fileName[TBuffMAXFNAMELENGTH];  // name of the file in use
    int  fd; // file descriptor
    tFileHeader fileHeader;
    long firstBufferDiskAddress; // position on thisk of the first buffer. It is
                                 // used to calculate other buffers addresses
    TNSCache<TBuffer> *cache;  // cache pointer
    int pagesPerBuffer; // ceil(bufferSize/PageSize)

    //---------------------------------------------
    // Header methods
    //---------------------------------------------
    void readHeader(); // read system header
    void writeHeader(); // write system header

    //---------------------------------------------
    // Address convertion methods
    //---------------------------------------------
    long toBufferId(long fileCursor) { // from disk address to buffer id
      return ((fileCursor - firstBufferDiskAddress)/(pagesPerBuffer *
                                        fileHeader.pageSize));
    }

    long toFileCursor(long bufferId) { // from buffer id to disk address
      return (firstBufferDiskAddress + (pagesPerBuffer * bufferId *
                                        fileHeader.pageSize));
    }

    //---------------------------------------------
    // Disk <-> cache methods
    //---------------------------------------------
    void initCache(); // starts the cache manager
    // used by the cache to read a buffer from disk
    void diskReadBuff(long bufferId, void *buffer);
    // used by the cache to write a buffer to disk
    void diskWriteBuff(long bufferId, void *buffer);

    //---------------------------------------------
    // General methods
    //---------------------------------------------
    void printErrorMsgNAbort(char *errorMsg); // as the name says
    void zeroAll(); // initalize all variables with zero or NULL
};
#endif

/*
//==============================================================================
// ShortCuts
//==============================================================================
//---------------------------------------------
// Constructors and destructor methods
//---------------------------------------------
TBuffer();
TBuffer(fName);
TBuffer(fName, pageSize, bufferSize, userHeaderSize, cacheNPages);
~TBuffer();

//---------------------------------------------
// File methods
//---------------------------------------------
openFile(fName);
closeThis();
createFile(fName, pageSize, bufferSize, userHeaderSize, cacheNPages);
isOpened()
dropThis();
getFileVersion()
getFileSize()
getUserFileSize();

//---------------------------------------------
// User's header methods. The disk is directly
// accessed - no cache operation
//---------------------------------------------
readUserHeader(buffer);
writeUserHeader(buffer);

//---------------------------------------------
// Buffer methods
//---------------------------------------------
readBuffer(bufferId, buffer);
writeBuffer(bufferId, buffer);
insertNewBuffer(buffer);
freeBuffer(bufferId);

//---------------------------------------------
// Cache methods
//---------------------------------------------
flushThis();
cacheTotalPages();
resizeCache(newCacheNPages);

//---------------------------------------------
// General methods
//---------------------------------------------
getBufferSize();
getPageSize();
getTotalBuffersInUse();
getTotalBuffersIncludingDisposed();
getTotalPagesInUse();
getTotalPagesIncludingDisposed();

//---------------------------------------------
// Statistics Related - When activated
//---------------------------------------------
#ifdef TBuffSTATISTICS
counters.headerRead;
counters.headerWrite;
counters.bufferRead;
counters.bufferWrite;
counters.newBuffer;
counters.relativeTotalDiskAccess;
counters.relativeTotalDiskRead;
counters.relativeTotalDiskWrite;
counters.diskRead;
counters.diskWrite;

getStatistics();
resetStatistics();
#endif

private:
//---------------------------------------------
// Constants and types
//---------------------------------------------
TBuffMAXFNAMELENGTH
TBuffSTATISTICS
TBuffHEADERRESERVED
TBuffCOUNTERSMAXVALUE
FROT
TBuffVERSION
TBuffHEADERSIZE

//---------------------------------------------
// Variables
//---------------------------------------------
fileName;
fd;

fileHeader.fileVersion
fileHeader.bufferSize
fileHeader.pageSize
fileHeader.cacheTotalPages
fileHeader.totalBuffersInUse
fileHeader.disposeList
fileHeader.totalBuffersInDisposeList
fileHeader.userHeaderSize
fileHeader.reserved

firstBufferDiskAddress;

cache->readBuffer(bufferId, &auxBuffer);
cache->writeBuffer();
cache->newBuffer(bufferId, &auxBuffer);
cache->flushCache();

pagesPerBuffer;

//---------------------------------------------
// Header methods
//---------------------------------------------
readHeader();
writeHeader();

//---------------------------------------------
// Address convertion methods
//---------------------------------------------
toBufferId(fileCursor);
toFileCursor(bufferId);

//---------------------------------------------
// Disk <-> cache methods
//---------------------------------------------
initCache();
diskReadBuff(bufferId, buffer);
diskWriteBuff(bufferId, buffer);

//---------------------------------------------
// General methods
//---------------------------------------------
printErrorMsgNAbort(errorMsg);
zeroAll();

*/
