//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Omni-Family
//  File: TExtractFoci.h
//  Classes: TExtractFoci
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 05/12/2001
//  Version: 0.00
//
//  Description:
//  Extracts a foci base from a dataset.
//
//  Modifications:
//    Author:
//    Date:
//    Description:
//==============================================================================

#ifndef __TExtractFociH
#define __TExtractFociH

#include <stdlib.h>

// ---- C++ Builder && GNU GCC ----
#if defined(__BORLANDC__) || defined(__GNUC__)
  #include <values.h>
#endif

// ---- Visual C++ ----
#if defined(_MSC_VER)
  #include <limits.h>
  #define MAXINT INT_MAX
#endif

#define __TExtractFociH__LINESIZE 1024

//------------------------------------------------------------------------------
// Class TExtractFoci definitions
//------------------------------------------------------------------------------
template <class TObjType>
class TExtractFoci {
  public:
    TExtractFoci(); // null constructor
    ~TExtractFoci();

    typedef typename TObjType::distanceType TDistType;
    // extracts the foci base from a given dataset
    void extractFoci(char *dataSetFName, int oSize, int numFoci);
    // sequentially reads the foci base from a given file
    void readFromFile(char *fociFName, int oSize, int numFoci);
    // save the foci base into a file
    void saveToFile(char *fociFName);
    // returns a pointer to the foci base. The user CANNOT destroy this.
    TObjType *getFociBase();
    // reset. Prepare for a new extraction.
    void reset();

  private:
    TObjType *fociBase;
    TObjType *dataSet;
    int nFoci;
    int objSize;
    long dataSetSize;

    void allocFociBase(TObjType **foci);
    void readDataSet(char *dataSetFName);
    // execute the HF algorithm to find the foci base
    void execHF();
    void execHFRemaining(TDistType edge, int nFound);

    void printErrorMsgNAbort(char *errorMsg); // as the name says
    void zeroAll(); // initalize all variables with zero or NULL
};

//------------------------------------------------------------------------------
// Null constructor
//------------------------------------------------------------------------------
template <class TObjType>
TExtractFoci<TObjType>::TExtractFoci() {
  zeroAll();
}

//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
template <class TObjType>
TExtractFoci<TObjType>::~TExtractFoci() {
  reset();
}

//------------------------------------------------------------------------------
// Extracts the foci base from a given dataset
//------------------------------------------------------------------------------
template <class TObjType>
void TExtractFoci<TObjType>::extractFoci(char *dataSetFName, int oSize,
                                        int numFoci) {
  if (fociBase)
    printErrorMsgNAbort("extractFoci error! There is another base created. "
                    "Before creating a new one, please execute 'reset' first.");

  if (numFoci < 1)
    printErrorMsgNAbort("extractFoci error! Invalid number of foci.\n");
  if (oSize < 1)
    printErrorMsgNAbort("extractFoci error! Invalid object size.\n");

  nFoci = numFoci;
  objSize = oSize;
  allocFociBase(&fociBase);
  readDataSet(dataSetFName);
  execHF();

  delete[] dataSet;
  dataSet = NULL;
}

//------------------------------------------------------------------------------
// Sequentially reads the foci base from a given file
//------------------------------------------------------------------------------
template <class TObjType>
void TExtractFoci<TObjType>::readFromFile(char *fociFName, int oSize,
                                         int numFoci) {
  int i;
  FILE *fdIn;

  if (fociBase)
    printErrorMsgNAbort("readFromFile error! There is another base created. "
               "Before creating a new one, please execute 'reset()' function.");

  if ((fdIn = fopen(fociFName, "rt")) == NULL) {
    char aux[100];

    sprintf(aux, "readFromFile error! Can't open file '%s'.\n", fociFName);
    printErrorMsgNAbort(aux);
  }

  fscanf(fdIn, "%d", &nFoci);

  if ((numFoci > nFoci) || (numFoci < 1))
    printErrorMsgNAbort("readFromFile error! Invalid number of foci.\n");
  if (oSize < 1)
    printErrorMsgNAbort("readFromFile error! Invalid object size.\n");

  nFoci = numFoci;
  objSize = oSize;
  allocFociBase(&fociBase);

  for(i = 0; i < nFoci; i++)
    fociBase[i].readFromTxtFile(fdIn);

  fclose(fdIn);
}

//------------------------------------------------------------------------------
// Save the foci base into a file
//------------------------------------------------------------------------------
template <class TObjType>
void TExtractFoci<TObjType>::saveToFile(char *fociFName) {
  int i;
  FILE *fdOut;

  if (!fociBase)
    printErrorMsgNAbort("saveToFile error! There is no foci base to be saved.");

  if ((fdOut = fopen(fociFName, "wt")) == NULL) {
    char aux[100];

    sprintf(aux, "saveToFile error! Can't open file '%s'.\n", fociFName);
    printErrorMsgNAbort(aux);
  }

  fprintf(fdOut, "%d\n", nFoci);

  for(i = 0; i < nFoci; i++)
    fociBase[i].writeToTxtFile(fdOut);

  fclose(fdOut);
}

//------------------------------------------------------------------------------
// returns a pointer to the foci base. The user CANNOT destroy this.
//------------------------------------------------------------------------------
template <class TObjType>
TObjType *TExtractFoci<TObjType>::getFociBase() {
  if (!fociBase)
    printErrorMsgNAbort("getFociBase error! There is no foci base created.");

  return (fociBase);
}

//------------------------------------------------------------------------------
// Reset. Prepare for a new extraction.
//------------------------------------------------------------------------------
template <class TObjType>
void TExtractFoci<TObjType>::reset() {
  if (fociBase)
    delete[] fociBase;

  if (dataSet)
    delete[] dataSet;

  zeroAll();
}

//------------------------------------------------------------------------------
// Allocates the foci base
//------------------------------------------------------------------------------
template <class TObjType>
void TExtractFoci<TObjType>::allocFociBase(TObjType **foci) {
  int i;

  *foci = new TObjType[nFoci];
  for (i = 0; i < nFoci; i++)
    (*foci)[i].setSize(objSize);
}

//------------------------------------------------------------------------------
// Load the sample dataset to main memory
//------------------------------------------------------------------------------
template <class TObjType>
void TExtractFoci<TObjType>::readDataSet(char *dataSetFName) {
  int i;
  FILE *fdIn;

  if ((fdIn = fopen(dataSetFName, "rt")) == NULL) {
    char aux[100];

    sprintf(aux, "readDataSet error! Can't open file '%s'.\n", dataSetFName);
    printErrorMsgNAbort(aux);
  }

  fscanf(fdIn, "%d", &dataSetSize);

  if ((dataSetSize < 1) || (dataSetSize < nFoci))
    printErrorMsgNAbort("readDataSet error! Invalid number of objects in"
                        "dataset.\n");

  dataSet = new TObjType[dataSetSize];
  for (i = 0; i < dataSetSize; i++)
    dataSet[i].setSize(objSize);

  for(i = 0; i < dataSetSize; i++)
    dataSet[i].readFromTxtFile(fdIn);

  fclose(fdIn);
}

//------------------------------------------------------------------------------
// execute the HF algorithm to find the foci base
//------------------------------------------------------------------------------
template <class TObjType> void
TExtractFoci<TObjType>::execHF() {
  TDistType edge, error, bestError;
  long i, nFound, bestCandidate;

  nFound = 0;

  // set the first object as the base object to find the first focus
  bestCandidate = 0;
  error = 0;
  bestError = 0;
  // traverse dataset and get first foci
  for(i = 1; i < dataSetSize; i++) {
    // compare first object with actual and keep if it is furthest than the
    // actual furthest
    error = dataSet[0].distance(&dataSet[i]);
    if (error > bestError) {
      bestCandidate = i;
      bestError = error;
    }
  }

  // insert first focus and removes it from dataSet
  fociBase[nFound] = dataSet[bestCandidate];
  dataSet[bestCandidate] = dataSet[dataSetSize - 1];
  dataSetSize--;
  nFound++;

  if (nFound < nFoci) {
    // search for the second focus
    bestCandidate = 0;
    error = 0;
    bestError = 0;
    for(i = 0; i < dataSetSize; i++) {
      // compares with actual best and keep if it is better
      error = fociBase[0].distance(&dataSet[i]);
      if (error > bestError) {
        bestCandidate = i;
        bestError = error;
      }
    }

    // insert second focus and removes it from dataSet
    fociBase[nFound] = dataSet[bestCandidate];
    dataSet[bestCandidate] = dataSet[dataSetSize - 1];
    dataSetSize--;
    nFound++;

    // find the remaining focus
    edge = fociBase[0].distance(&fociBase[1]);
    execHFRemaining(edge, nFound);
  }
}

//------------------------------------------------------------------------------
// Works with choosePivotsSampled to find the remaining foci
//------------------------------------------------------------------------------
template <class TObjType> void
TExtractFoci<TObjType>::execHFRemaining(TDistType edge, int nFound) {
  TDistType error, bestError, delta;
  long i, j, bestCandidate;

  // find the remaining foci
  while (nFound < nFoci) {
    bestError = (TDistType) MAXINT;
    bestCandidate = 0;
    for(i = 0; i < dataSetSize; i++) {
      error = 0;
      // calculates the error
      for(j = 0; j < nFound; j++) {
        delta = edge - fociBase[j].distance(&dataSet[i]);
        if (delta < 0)
          delta *= -1;
        error += delta;
      }

      // compares with actual best and keep if it is better
      if (error < bestError) {
        bestCandidate = i;
        bestError = error;
      }
    }

    // insert focus and removes it from dataSet
    fociBase[nFound] = dataSet[bestCandidate];
    dataSet[bestCandidate] = dataSet[dataSetSize - 1];
    dataSetSize--;
    nFound++;
  }
}

//------------------------------------------------------------------------------
// Print an error message and aborts the program
//------------------------------------------------------------------------------
template <class TObjType> void
TExtractFoci<TObjType>::printErrorMsgNAbort(char *errorMsg) {
  reset();

  printf("TOmniTree::%s\n", errorMsg);
  exit(1);
}

//------------------------------------------------------------------------------
// initalize all variables with zero or NULL
//------------------------------------------------------------------------------
template <class TObjType> void
TExtractFoci<TObjType>::zeroAll() {
  fociBase = NULL;
  dataSet = NULL;
  nFoci = 0;
  objSize = 0;
  dataSetSize = 0;
}
//---------------------------------------------------------------------------
#endif
