//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Spatial/metric indexing methods
//  File: TLList.cpp
//  Classes: TLList
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 06/24/2000
//  Version: .01
//
//  Description:
//  Dinamic linked list.
//
//  Modifications:
//    Author:
//    Date:
//    Description:
//==============================================================================

#ifndef __TLListH
#define __TLListH

//---------------------------------------------------------------------------
#include <stdio.h>

//------------------------------------------------------------------------------
// Class TLList definitions
//------------------------------------------------------------------------------
template <class TObjType, class keyType>
class TLList {
  public:
    //---------------------------------------------
    // Constructor methods
    //---------------------------------------------
    TLList(char buildOrdered = false);

    //---------------------------------------------
    // Entry methods
    //---------------------------------------------
    void insertEntry(TObjType *obj, long objId) { // FILO insertion
      if (!isOrdered)
        printErrorMsgNAbort("insertEntry error! It was suposed"
                            " to be an ordered insertion.");

      insertEntry(obj, objId, (keyType) 0);
    }

    // ordered insertion
    void insertEntry(TObjType *obj, long objId, keyType key);

    // all the following methods returns false if list is empty.
    // Obj must be allocated before calling these methods.
    // I decide to do by this way to save some attribution (copies).
    // If the method returns the entry, it'll have to make a first copy to
    // delete the node and a second one through the return.

    // delete the first entry and returns to user
    char getEntry(TObjType *obj, long *objId) {
      keyType aux;

      return (getEntry(obj, objId, &aux));
    }

    // same but with ordered list
    char getEntry(TObjType *obj, long *objId, keyType *key);

    // delete the positionth entry and returns to user. 0 means first position.
    char getEntry(int position, TObjType *obj, long *objId) {
      keyType aux;

      return (getEntry(position, obj, objId, &aux));
    }

    // with ordered list
    char getEntry(int position, TObjType *obj, long *objId, keyType *key);

    //---------------------------------------------
    // List Methods
    //---------------------------------------------
    void destroyList();

    //---------------------------------------------
    // General methods
    //---------------------------------------------
    char isEmpty() {
      return ((char)(!listSize));
    }

    long getSize() {
      return (listSize);
    }

  private:
    //---------------------------------------------
    // Types
    //---------------------------------------------
    struct listEntry {
      TObjType obj;
      long objId;
      keyType key;
      listEntry *next;
    };

    //---------------------------------------------
    // Variables
    //---------------------------------------------
    listEntry *header;
    listEntry *p; // work pointer
    char isOrdered; // value true if ordered, false otherwise
    long listSize;

    //---------------------------------------------
    // General methods
    //---------------------------------------------
    void zeroAll();
};

//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
template <class TObjType, class keyType>
TLList<TObjType, keyType>::TLList(char buildOrdered) {
  zeroAll();
  isOrdered = buildOrdered;
}

//------------------------------------------------------------------------------
// Ordered insert a new entry
//------------------------------------------------------------------------------
template <class TObjType, class keyType> void
TLList<TObjType, keyType>::insertEntry(TObjType *obj, long objId, keyType key) {
  listEntry **father, *newEntry;
  char stop;

  newEntry = new listEntry;
  newEntry->obj = *obj;
  newEntry->objId = objId;
  newEntry->key = key;

  p = header;
  father = &header;

  stop = false;
  while (!stop && p) {
    if (key <= p->key)
      stop = true;
    else {
      father = &(p->next);
      p = p->next;
    }
  }
  *father = newEntry;
  newEntry->next = p;

  listSize++;
}

//------------------------------------------------------------------------------
// Delete the first entry and returns the object to user
//------------------------------------------------------------------------------
template <class TObjType, class keyType> char
TLList<TObjType, keyType>::getEntry(TObjType *obj, long *objId, keyType *key) {
  char result = false;

  p = header;
  if (p) {
    result = true;

    *obj = p->obj;
    *objId = p->objId;
    *key = p->key;

    header = header->next;
    delete p;

    listSize--;
  }

  return (result);
}

//------------------------------------------------------------------------------
// delete the positionth entry and returns to user. 0 means first position.
//------------------------------------------------------------------------------
template <class TObjType, class keyType> char
TLList<TObjType, keyType>::getEntry(int position, TObjType *obj, long *objId,
                                                                 keyType *key) {
  listEntry **father;
  char result;
  int i;

  p = header;
  father = &header;
  i = 0;
  result = false;

  while (!result && p) {
    if (i == position) {
      result = true;

      *obj = p->obj;
      *objId = p->objId;
      *key = p->key;

      *father = p->next;
      delete p;

      listSize--;
    }
    else {
      father = &(p->next);
      p = p->next;
    }
  }

  return (result);
}

//------------------------------------------------------------------------------
// Destroy the list
//------------------------------------------------------------------------------
template <class TObjType, class keyType> void
TLList<TObjType, keyType>::destroyList() {
  p = header;

  while (header) {
    header = header->next;
    delete p;
    p = header;
  }

  zeroAll();
}

//------------------------------------------------------------------------------
// initalize all variables with zero or NULL
//------------------------------------------------------------------------------
template <class TObjType, class keyType> void
TLList<TObjType, keyType>::zeroAll(){
  header = NULL;
  p = NULL;

  isOrdered = false;
  listSize = 0;
}
#endif
