//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Omni-RTree for spatial/metric data
//  File: SeqScan.cpp
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 05/09/2001
//  Version: 1.0 - Cached Version
//
//  Description:
//  Performs range and nearest neghbors queries by indexing the omni coordinates
//  with a R-Tree.
//
//  Modifications:
//    Author:
//    Date:
//    Description:
//==============================================================================
//---------------------------------------------------------------------------
#include <stdlib.h>
#include <time.h>
#include <sys/stat.h>
#include <errno.h>
#include <math.h>

// ---- C++ Builder ----
#ifdef __BORLANDC__
  #include <dir.h>
#endif

// ---- GNU GCC ----
#ifdef __GNUC__
  #include <unistd.h>
#endif

// ---- Visual C++ ----
#ifdef _MSC_VER
  #include <direct.h>
#endif

#ifdef CLOCKS_PER_SEC
  #define CLOCK_RATIO CLOCKS_PER_SEC
#elif
  #define CLOCK_RATIO CLK_TCK
#endif

//---------------------------------------------------------------------------
#pragma hdrstop

//---------------------------------------------------------------------------
#include "CommandProc.h"
#include "TOmniTree.h"
#include "TLList.h"
#include "NumericVectObj.h"
#include "StringObj.h"
#include "TExtractFoci.h"

//---------------------------------------------------------------------------
enum TObjTypeInUse {numObj, stringObj};
TObjTypeInUse objTypeInUse = numObj;

typedef float numTypeInUse;
//typedef double numTypeInUse;

int dimension = 0;
int nFoci = 0;

enum randMethod {rmBiased, rmUnBiased};
randMethod CurrentRandMethod = rmUnBiased;

char randFName[100];
char idxName[100] = "OmniRTree.dat\0";

/*--------- Global Set Parameters ------------------------------------------*/
int NQueries = 500; // Number of queries to generate statistics about queries
char outDirName[100];
bool isOutDirSet = false;

int objBuffSize = 4096;
int mbOrBuffSize = 4096;

//-Internal Methods ----------------------------------------------------------
//   Computes the Average and Standard Deviation of a set of points
//    X - x-points (in)
//    n - number of points (in)
//    sd - Standard Deviation (out)
// result - Average.(out)
double Average(float *X, int n, float *sd) {
  int i;
  double avg=0., diffs=0.;
    for (i=0; i<n; i++) avg += (float)X[i];
    avg /= n;
    for (i=0; i<n; i++) diffs+=(avg-(float)X[i])*(avg-(float)X[i]);
    *sd = sqrt(diffs/(n-1));
    return (avg);
}

//---------------------------------------------------------------------------
// Execute initMinMax
TNumericVectObj<numTypeInUse> *initMinMax() {
  TNumericVectObj<numTypeInUse> *qObj;
  numTypeInUse *min, *max;
  int i;

  min = new numTypeInUse[dimension];
  max = new numTypeInUse[dimension];

  for (i = 0; i < dimension; i++) {
    min[i] = (numTypeInUse) 0.0;
    max[i] = (numTypeInUse) 1.0;
  }

  qObj = new TNumericVectObj<numTypeInUse>(dimension, min, max);
  return(qObj);
}

//---------------------------------------------------------------------------
// Execute WritePairs
void WritePairs(float *x, float *y, int vSize, char *fName) {
  FILE *fd;

  fd = fopen(fName, "w");

  for(;vSize; vSize--, x++, y++)
    fprintf(fd, "%f\t%f\n", *x, *y);

  fclose(fd);
}

//---------------------------------------------------------------------------
// Execute CatOutPath command
void OutPath(char *fName, char *path) {
  if (isOutDirSet)
    // ---- GNU GCC ----
    #ifdef __GNUC__
      sprintf(path, "%s/%s\0", outDirName, fName);
    #endif
    // ---- C++ Builder || Visual C++ ----
    #if defined(__BORLANDC__) || defined(_MSC_VER)
      sprintf(path, "%s\\%s\0", outDirName, fName);
    #endif
  else
    sprintf(path, "%s\0", fName);
}

//---------------------------------------------------------------------------
// Execute Exec_MInsert command
template <class TOmniTreeObj>
void Exec_MInsert(TOmniTreeObj *insObj) {
  int i, // counter
      totalIns, // total number of records to be inserted
      objInserted; // number of inserted records
  float *objDiskAccess, // vector with number of disk accesses
        objAvgDiskAccess, // average number of disk accesses
        objStdDevDiskAccess, // standard deviation of number of disk accesses
        *mbOrDiskAccess, // vector with number of disk accesses
        mbOrAvgDiskAccess, // average number of disk accesses
        mbOrStdDevDiskAccess, // standard deviation of number of disk accesses
        *objNumDistCal, // vector with number of distance calculations
        objAvgDistCal, // average number of distance calculations
        objStdDevDistCal, // standard deviation of number of distance calculations
        *mbOrNumDistCal, // vector with number of distance calculations
        mbOrAvgDistCal, // average number of distance calculations
        mbOrStdDevDistCal; // standard deviation of number of distance calculations

  clock_t totalTime; // CPU time
  char cmdLine[100]; // command line
  FILE *fdOut; // file descriptor to output results
  char path[100]; // output file name and path

  TOmniTree<TOmniTreeObj> omniIndex;

  typename TOmniTree<TOmniTreeObj>::performanceCounters performCounters;

  // reading parameters
  // using ReadCommand instead of ReadWord to make it possible to read the
  // database from another file
  ReadCommand(cmdLine, false, "minsert> #Objects: ");
  totalIns = atoi(cmdLine);

  OutPath(idxName, path);
  omniIndex.openIndex(path);

  // initializing variables
  objDiskAccess = new float[totalIns];
  objNumDistCal = new float[totalIns];

  mbOrDiskAccess = new float[totalIns];
  mbOrNumDistCal = new float[totalIns];
  for(i = 0; i < totalIns; i ++) {
    objDiskAccess[i] =
    objNumDistCal[i] =
    mbOrDiskAccess[i] =
    mbOrNumDistCal[i] = 0;
  }

  objInserted = 0;

  OutPath("OmniMInsert.txt", path);
  fdOut = fopen(path, "a");
  if (!(fdOut)) {
    printf("MInsert Error! Unable to open output file 'OmniMInsert.txt'.");
    exit(1);
  }

  printf("\n========================================\n");
  printf("\tInsert\n");
  printf("\tTotal #Objects = %d\n", totalIns);
  printf("========================================\n");

  totalTime = clock();
  for(i = 0; i < totalIns; i++) {
    omniIndex.resetStatistics();

    insObj->readObj(cmdLine, false, "features: ");

    omniIndex.insert(insObj, objInserted);

    objInserted++;

    performCounters = omniIndex.getStatistics();
    objDiskAccess[i] = (float) performCounters.objTotalDiskAccess;
    objNumDistCal[i] = (float) performCounters.objDistanceCalculation;
    mbOrDiskAccess[i] = (float) performCounters.mbOrTotalDiskAccess;
    mbOrNumDistCal[i] = (float) performCounters.mbOrDistanceCalculation;
  }
  totalTime = clock() - totalTime; // stop time counter

  // Calculate the disk access average and
  // Standard Deviation for <uniforme and biased> data
  objAvgDiskAccess = Average(objDiskAccess, i, &objStdDevDiskAccess);
  objAvgDistCal = Average(objNumDistCal, i, &objStdDevDistCal);
  mbOrAvgDiskAccess = Average(mbOrDiskAccess, i, &mbOrStdDevDiskAccess);
  mbOrAvgDistCal = Average(mbOrNumDistCal, i, &mbOrStdDevDistCal);

  // Output the results of each interval
  fprintf(fdOut,
      //  "nFoci\t#Obj\tRTreeHeigh\t#Pag\tTime\tObjDisk\tStdDev\tmbOrDisk\t"
      //  "StdDev\tObjDistCalc\tStdDev\tmbOrDistCalc\tStdDev\t"
          "%d\t%ld\t%ld\t%ld\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\n",
          omniIndex.getNFoci(), objInserted, omniIndex.getRTreeHeight(),
          omniIndex.getTotalNodes(), (float)(totalTime/CLOCK_RATIO),
          objAvgDiskAccess, objStdDevDiskAccess, mbOrAvgDiskAccess,
          mbOrStdDevDiskAccess, objAvgDistCal, objStdDevDistCal, mbOrAvgDistCal,
          mbOrStdDevDistCal);

  printf("Total %d objects. Insertion time : %6.2f seconds.\n", objInserted,
         (float)(totalTime/CLOCK_RATIO));
  printf("Obj Disk Accesses: Average-> %.2f StdDev-> %.2f\n",
         objAvgDiskAccess, objStdDevDiskAccess);
  printf("Obj Distance Calculations: Average-> %.2f StdDev-> %.2f\n",
         objAvgDistCal, objStdDevDistCal);
  printf("mbOr Disk Accesses: Average-> %.2f StdDev-> %.2f\n",
         mbOrAvgDiskAccess, mbOrStdDevDiskAccess);
  printf("mbOr Distance Calculations: Average-> %.2f StdDev-> %.2f\n",
         mbOrAvgDistCal, mbOrStdDevDistCal);
  printf("Number of nodes: %d\n", omniIndex.getTotalNodes());
  printf("RTree heigh: %ld.\n", omniIndex.getRTreeHeight());


  totalTime = clock();
  printf("\nSynchronizing Object File...\n");
  omniIndex.synchronizeObjFile();
  totalTime = clock() - totalTime; // stop time counter
  printf("Synchronized. Time : %6.2f seconds.\n", (float)(totalTime/CLOCK_RATIO));

  omniIndex.closeIndex();

  delete[] objDiskAccess;
  delete[] objNumDistCal;
  delete[] mbOrDiskAccess;
  delete[] mbOrNumDistCal;

  fclose(fdOut);
}

//---------------------------------------------------------------------------
// Execute Exec_MRQ command
template <class TOmniTreeObj>
void Exec_MRQ(TOmniTreeObj *qObj) {
  int i; // counter
  float *objDiskAccess, // vector with number of disk accesses
        objAvgDiskAccess, // average number of disk accesses
        objStdDevDiskAccess, // standard deviation of number of disk accesses
        *mbOrDiskAccess, // vector with number of disk accesses
        mbOrAvgDiskAccess, // average number of disk accesses
        mbOrStdDevDiskAccess, // standard deviation of number of disk accesses
        *objNumDistCal, // vector with number of distance calculations
        objAvgDistCal, // average number of distance calculations
        objStdDevDistCal, // standard deviation of number of distance calculations
        *mbOrNumDistCal, // vector with number of distance calculations
        mbOrAvgDistCal, // average number of distance calculations
        mbOrStdDevDistCal, // standard deviation of number of distance calculations
        *found, // number of objects that qualified
        avgFound, // average of found
        stdDevFound, // standard deviation of found
        *searchTime, // vector with the time for each search
        avgSearchTime, // vector with the time for each search
        stdDevSearchTime; // vector with the time for each search

  clock_t totalTime; // CPU time
  char cmdLine[100]; // command line
  typename TOmniTreeObj::distanceType radius; // radius to be queried
  FILE *fdOut; // file descriptor to output results
  char path[100]; // output file name and path

  TOmniTree<TOmniTreeObj> omniIndex;

  typename TOmniTree<TOmniTreeObj>::performanceCounters performCounters;

  // ---- C++ Builder || GNU GCC ----
  #if defined(__BORLANDC__) || defined(__GNUC__)
  TLList<TOmniTreeObj, typename TOmniTreeObj::distanceType> qResponseList(true);
  #endif
  // ---- Visual C++ ----
  #ifdef _MSC_VER
  TLList<TOmniTreeObj, TOmniTreeObj::distanceType> qResponseList(true);
  #endif

  OutPath(idxName, path);
  omniIndex.openIndex(path);

  // reading parameters
  // using ReadCommand instead of ReadWord to make it possible to read the
  // database from another file
  ReadCommand(cmdLine, false, "mrq> radius: ");
  radius = (typename TOmniTreeObj::distanceType) atof(cmdLine);

  // initializing variables
  objDiskAccess = new float[NQueries];
  objNumDistCal = new float[NQueries];

  mbOrDiskAccess = new float[NQueries];
  mbOrNumDistCal = new float[NQueries];

  searchTime = new float[NQueries];
  found = new float[NQueries];

  OutPath("OmniMRQ.txt", path);
  fdOut = fopen(path, "a");
  if (!(fdOut)) {
    printf("MRQ Error! Unable to open output file 'OmniMRQ.txt'.");
    exit(1);
  }

  printf("\n========================================\n");
  printf("\tRange Query\n");
  printf("\tTotal Queries = %d\n", NQueries);
  printf("\tRadius = %f\n", (float) radius);
  printf("========================================\n");

  totalTime = clock();
  for(i = 0; i < NQueries; i++) {
    omniIndex.resetStatistics();
    searchTime[i] = clock();

    qObj->randomize();

    found[i] = omniIndex.rQuery(qObj, radius);
    // found[i] = omniIndex.rQuery(qObj, radius, &qResponseList);
    // qResponseList.destroyList();
    // printf("%.0f\t", found[i]);

    performCounters = omniIndex.getStatistics();
    objDiskAccess[i] = (float) performCounters.objTotalDiskAccess;
    objNumDistCal[i] = (float) performCounters.objDistanceCalculation;
    mbOrDiskAccess[i] = (float) performCounters.mbOrTotalDiskAccess;
    mbOrNumDistCal[i] = (float) performCounters.mbOrDistanceCalculation;
    searchTime[i] = clock() - searchTime[i];
  }
  totalTime = clock() - totalTime; // stop time counter

  // Calculate the disk access average and
  // Standard Deviation for <uniforme and biased> data
  objAvgDiskAccess = Average(objDiskAccess, i, &objStdDevDiskAccess);
  objAvgDistCal = Average(objNumDistCal, i, &objStdDevDistCal);
  mbOrAvgDiskAccess = Average(mbOrDiskAccess, i, &mbOrStdDevDiskAccess);
  mbOrAvgDistCal = Average(mbOrNumDistCal, i, &mbOrStdDevDistCal);
  avgFound = Average(found, i, &stdDevFound);
  avgSearchTime = Average(searchTime, i, &stdDevSearchTime);

  // Output the results of each interval
  fprintf(fdOut,
      //  "nFoci\t#Obj\tRadius\tTime\tAvg\tObjDisk\tStdDev\tmbOrDisk\tStdDev\t"
      //  "ObjDistCalc\tStdDev\tmbOrDistCalc\tStdDev\tobjFound\tStdDev"
          "%d\t%ld\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\n",
          omniIndex.getNFoci(), omniIndex.getTotalObjects(), (float) radius,
          (float)(totalTime/CLOCK_RATIO), (float)(avgSearchTime/CLOCK_RATIO),
          objAvgDiskAccess, objStdDevDiskAccess, mbOrAvgDiskAccess,
          mbOrStdDevDiskAccess, objAvgDistCal, objStdDevDistCal,
          mbOrAvgDistCal, mbOrStdDevDistCal, avgFound, stdDevFound);

  printf("Total time to %d queries of radius %f: %.2f seconds.\n", NQueries,
           (float) radius, (float)(totalTime/CLOCK_RATIO));
  printf("Qualified objects: Average-> %.2f StdDev-> %.2f\n",
                                     avgFound, stdDevFound);
  printf("Obj Disk Accesses: Average-> %.2f StdDev-> %.2f\n",
                                     objAvgDiskAccess, objStdDevDiskAccess);
  printf("Obj Distance Calculations: Average-> %.2f StdDev-> %.2f\n",
                                     objAvgDistCal, objStdDevDistCal);
  printf("mbOr Disk Accesses: Average-> %.2f StdDev-> %.2f\n",
                                     mbOrAvgDiskAccess, mbOrStdDevDiskAccess);
  printf("mbOr Distance Calculations: Average-> %.2f StdDev-> %.2f\n",
                                     mbOrAvgDistCal, mbOrStdDevDistCal);

  omniIndex.closeIndex();

  delete[] objDiskAccess;
  delete[] objNumDistCal;
  delete[] mbOrDiskAccess;
  delete[] mbOrNumDistCal;
  delete[] searchTime;
  delete[] found;

  fclose(fdOut);
}

//---------------------------------------------------------------------------
// Execute Exec_MNN command
template <class TOmniTreeObj>
void Exec_MNN(TOmniTreeObj *qObj) {
  int i; // counter
  float *objDiskAccess, // vector with number of disk accesses
        objAvgDiskAccess, // average number of disk accesses
        objStdDevDiskAccess, // standard deviation of number of disk accesses
        *mbOrDiskAccess, // vector with number of disk accesses
        mbOrAvgDiskAccess, // average number of disk accesses
        mbOrStdDevDiskAccess, // standard deviation of number of disk accesses
        *objNumDistCal, // vector with number of distance calculations
        objAvgDistCal, // average number of distance calculations
        objStdDevDistCal, // standard deviation of number of distance calculations
        *mbOrNumDistCal, // vector with number of distance calculations
        mbOrAvgDistCal, // average number of distance calculations
        mbOrStdDevDistCal, // standard deviation of number of distance calculations
        *found, // number of objects that qualified
        avgFound, // average of found
        stdDevFound, // standard deviation of found
        *searchTime, // vector with the time for each search
        avgSearchTime, // vector with the time for each search
        stdDevSearchTime; // vector with the time for each search

  clock_t totalTime; // CPU time
  char cmdLine[100]; // command line
  int kNN; // number of beighbors to be queried
  FILE *fdOut; // file descriptor to output results
  char path[100]; // output file name and path

  TOmniTree<TOmniTreeObj> omniIndex;

  typename TOmniTree<TOmniTreeObj>::performanceCounters performCounters;

  // ---- C++ Builder || GNU GCC ----
  #if defined(__BORLANDC__) || defined(__GNUC__)
  TLList<TOmniTreeObj, typename TOmniTreeObj::distanceType> qResponseList(true);
  #endif
  // ---- Visual C++ ----
  #ifdef _MSC_VER
  TLList<TOmniTreeObj, TOmniTreeObj::distanceType> qResponseList(true);
  #endif

  OutPath(idxName, path);
  omniIndex.openIndex(path);

  // reading parameters
  // using ReadCommand instead of ReadWord to make it possible to read the
  // database from another file
  ReadCommand(cmdLine, false, "mnn> k-NN: ");
  kNN = atoi(cmdLine);

  // initializing variables
  objDiskAccess = new float[NQueries];
  objNumDistCal = new float[NQueries];

  mbOrDiskAccess = new float[NQueries];
  mbOrNumDistCal = new float[NQueries];

  searchTime = new float[NQueries];
  found = new float[NQueries];

  OutPath("OmniMNN.txt", path);
  fdOut = fopen(path, "a");
  if (!(fdOut)) {
    printf("MNN Error! Unable to open output file 'OmniMNN.txt'.");
    exit(1);
  }

  printf("\n========================================\n");
  printf("\tk-Nearest Neighbors Query\n");
  printf("\tTotal Queries = %d\n", NQueries);
  printf("\tk = %d\n", kNN);
  printf("========================================\n");

  totalTime = clock();
  for(i = 0; i < NQueries; i++) {
    omniIndex.resetStatistics();
    searchTime[i] = clock();

    qObj->randomize();

    qResponseList = omniIndex.kNNQuery(qObj, kNN);
    found[i] = qResponseList.getSize();
    qResponseList.destroyList();
    // printf("%ld\t", found[j]);

    performCounters = omniIndex.getStatistics();
    objDiskAccess[i] = (float) performCounters.objTotalDiskAccess;
    objNumDistCal[i] = (float) performCounters.objDistanceCalculation;
    mbOrDiskAccess[i] = (float) performCounters.mbOrTotalDiskAccess;
    mbOrNumDistCal[i] = (float) performCounters.mbOrDistanceCalculation;
    searchTime[i] = clock() - searchTime[i];
  }
  totalTime = clock() - totalTime; // stop time counter

  // Calculate the disk access average and
  // Standard Deviation for <uniforme and biased> data
  objAvgDiskAccess = Average(objDiskAccess, i, &objStdDevDiskAccess);
  objAvgDistCal = Average(objNumDistCal, i, &objStdDevDistCal);
  mbOrAvgDiskAccess = Average(mbOrDiskAccess, i, &mbOrStdDevDiskAccess);
  mbOrAvgDistCal = Average(mbOrNumDistCal, i, &mbOrStdDevDistCal);
  avgFound = Average(found, i, &stdDevFound);
  avgSearchTime = Average(searchTime, i, &stdDevSearchTime);

  // Output the results of each interval
  fprintf(fdOut,
      //  "nFoci\t#Obj\tTime\tObjDisk\tStdDev\tmbOrDisk\tStdDev\t"
      //  "ObjDistCalc\tStdDev\tmbOrDistCalc\tStdDev\tobjFound\tStdDev"
          "%d\t%ld\t%d\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\n",
          omniIndex.getNFoci(), omniIndex.getTotalObjects(), kNN,
          (float)(totalTime/CLOCK_RATIO), (float)(avgSearchTime/CLOCK_RATIO),
          objAvgDiskAccess, objStdDevDiskAccess, mbOrAvgDiskAccess,
          mbOrStdDevDiskAccess, objAvgDistCal, objStdDevDistCal,
          mbOrAvgDistCal, mbOrStdDevDistCal, avgFound, stdDevFound);

  printf("Total time to %d queries with k = %d: %.2f seconds.\n", NQueries,
           kNN, (float)(totalTime/CLOCK_RATIO));
  printf("Qualified objects: Average-> %.2f StdDev-> %.2f\n",
                                     avgFound, stdDevFound);
  printf("Obj Disk Accesses: Average-> %.2f StdDev-> %.2f\n",
                                     objAvgDiskAccess, objStdDevDiskAccess);
  printf("Obj Distance Calculations: Average-> %.2f StdDev-> %.2f\n",
                                     objAvgDistCal, objStdDevDistCal);
  printf("mbOr Disk Accesses: Average-> %.2f StdDev-> %.2f\n",
                                     mbOrAvgDiskAccess, mbOrStdDevDiskAccess);
  printf("mbOr Distance Calculations: Average-> %.2f StdDev-> %.2f\n",
                                     mbOrAvgDistCal, mbOrStdDevDistCal);

  omniIndex.closeIndex();

  delete[] objDiskAccess;
  delete[] objNumDistCal;
  delete[] mbOrDiskAccess;
  delete[] mbOrNumDistCal;
  delete[] searchTime;
  delete[] found;

  fclose(fdOut);
}

//---------------------------------------------------------------------------
// Execute RQStep command
template <class TOmniTreeObj>
void Exec_RQStep (TOmniTreeObj *qObj, TOmniTreeObj *insObj) {
  int i, j, // counters
      totalIns, // total number of records to be inserted
      steps, // Number of statistical data to be collected
      quotient, // number of records to be inserted per interval
      rest, // number of intervals with different number of records
      objInserted; // number of inserted records
  float *objDiskAccess, // vector with number of disk accesses
        objAvgDiskAccess, // average number of disk accesses
        objStdDevDiskAccess, // standard deviation of number of disk accesses
        *mbOrDiskAccess, // vector with number of disk accesses
        mbOrAvgDiskAccess, // average number of disk accesses
        mbOrStdDevDiskAccess, // standard deviation of number of disk accesses
        *objNumDistCal, // vector with number of distance calculations
        objAvgDistCal, // average number of distance calculations
        objStdDevDistCal, // standard deviation of number of distance calculations
        *mbOrNumDistCal, // vector with number of distance calculations
        mbOrAvgDistCal, // average number of distance calculations
        mbOrStdDevDistCal, // standard deviation of number of distance calculations
        *found, // number of objects that qualified
        avgFound, // average of found
        stdDevFound, // standard deviation of found
        *searchTime, // vector with the time for each search
        avgSearchTime, // vector with the time for each search
        stdDevSearchTime; // vector with the time for each search

  clock_t totalTime; // CPU time
  typename TOmniTreeObj::distanceType radius; // radius to be queried
  char cmdLine[100]; // command line
  FILE *fdOut; // file descriptor to output results
  char path[100]; // output file name and path

  TOmniTree<TOmniTreeObj> omniIndex;
  typename TOmniTree<TOmniTreeObj>::performanceCounters performCounters;

  // ---- C++ Builder || GNU GCC ----
  #if defined(__BORLANDC__) || defined(__GNUC__)
  TLList<TOmniTreeObj, typename TOmniTreeObj::distanceType> qResponseList(true);
  #endif
  // ---- Visual C++ ----
  #ifdef _MSC_VER
  TLList<TOmniTreeObj, TOmniTreeObj::distanceType> qResponseList(true);
  #endif

  // reading parameters
  ReadWord(cmdLine, false, "rqstep> #Steps, Radius, #Objects: ");
  steps = atoi(cmdLine);
  ReadWord(cmdLine, false, "Radius, #Objects: ");
  radius = (typename TOmniTreeObj::distanceType) atof(cmdLine);
  // using ReadCommand instead of ReadWord to make it possible to read the
  // database from another file
  ReadCommand(cmdLine, false, "#Objects: ");
  totalIns = atoi(cmdLine);

  OutPath(idxName, path);
  omniIndex.openIndex(path);

  // initializing variables
  quotient = totalIns / steps; // size of each interval
  rest = totalIns % steps; // number of intervals with 1 more record
                           // necessary when totalInc isn't a multiple of steps

  objInserted = 0;

  objDiskAccess = new float[
                      (NQueries > (quotient + 1) ? NQueries : (quotient + 1))];
  objNumDistCal = new float[
                      (NQueries > (quotient + 1) ? NQueries : (quotient + 1))];

  mbOrDiskAccess = new float[
                      (NQueries > (quotient + 1) ? NQueries : (quotient + 1))];
  mbOrNumDistCal = new float[
                      (NQueries > (quotient + 1) ? NQueries : (quotient + 1))];
  searchTime = new float[
                      (NQueries > (quotient + 1) ? NQueries : (quotient + 1))];

  found = new float[(NQueries > (quotient + 1) ? NQueries : (quotient + 1))];

  OutPath("OmniRQS.txt", path);
  fdOut = fopen(path, "a");
  if (!(fdOut)) {
    printf("RQStep Error! Unable to open output file 'OmniRQ.txt'.");
    exit(1);
  }

  fprintf(fdOut, "Omni Scan - Range Query X Data Set Size - Radius = \t%f\n",
          (float) radius);
  fprintf(fdOut, "\t\t\t\tInsertion\t\t\t\t\t\t\t\tSelection\n");
  fprintf(fdOut, "\tTotal\t#Obj\t#\tTime\tObjDisk\t\tmbOrDisk\t\t"
                        "ObjDistCalc\t\tmbOrDistCalc\t\tTime\t\tObjDisk\t"
                        "\tmbOrDisk\t\tObjDistCalc\t\tmbOrDistCalc\t\t"
                        "Found\n");
  fprintf(fdOut, "Step\tObjs\tStep\tPgs\tIns\tAvg\tStdDev\tAvg\tStdDev\t"
                        "Avg\tStdDev\tAvg\tStdDev\tTotal\tAvg\tAvg\t"
                        "StdDev\tAvg\tStdDev\tAvg\tStdDev\tAvg\tStdDev\t"
                        "Avg\tStdDev\n");

  printf("\n========================================\n");
  printf("\tInsert & Range Query\n");
  printf("\tTotal #Objects = %d\n", totalIns);
  printf("========================================\n");

  // intervals loop
  for(i = 0; i < steps; i++) {
    printf ("\n------------- Interval[%2d] -------------\n", i);

    // insertions loop
    // initializing variables
    printf ("Inserting %5d objects.\n", quotient + (i < rest? 1: 0));
    totalTime = clock();
    for(j = 0; j < (quotient + (i < rest? 1: 0)); j++) {
      omniIndex.resetStatistics();

      insObj->readObj(cmdLine, false, "features: ");

      omniIndex.insert(insObj, objInserted);

      objInserted++;

      performCounters = omniIndex.getStatistics();
      objDiskAccess[j] = (float) performCounters.objTotalDiskAccess;
      objNumDistCal[j] = (float) performCounters.objDistanceCalculation;
      mbOrDiskAccess[j] = (float) performCounters.mbOrTotalDiskAccess;
      mbOrNumDistCal[j] = (float) performCounters.mbOrDistanceCalculation;
    }
    totalTime = clock() - totalTime; // stop time counter
    omniIndex.synchronizeObjFile();

    // Calculate the disk access average and
    // Standard Deviation for <uniforme and biased> data
    objAvgDiskAccess = Average(objDiskAccess, j, &objStdDevDiskAccess);
    objAvgDistCal = Average(objNumDistCal, j, &objStdDevDistCal);
    mbOrAvgDiskAccess = Average(mbOrDiskAccess, j, &mbOrStdDevDiskAccess);
    mbOrAvgDistCal = Average(mbOrNumDistCal, j, &mbOrStdDevDistCal);

    // Output the results of each interval
    fprintf(fdOut,
        //  "\tTotal\t#Obj\t#\tTime\tObjDisk\t\tmbOrDisk\t\t"
        //  "ObjDistCalc\t\tmbOrDistCalc\t\t"
            "%d\t%ld\t%ld\t%ld\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t",
            i, objInserted, j, omniIndex.getTotalNodes(),
            (float)(totalTime/CLOCK_RATIO), objAvgDiskAccess, objStdDevDiskAccess,
            mbOrAvgDiskAccess, mbOrStdDevDiskAccess, objAvgDistCal,
            objStdDevDistCal, mbOrAvgDistCal, mbOrStdDevDistCal);

    printf("Total %d objects. Interval time : %6.2f seconds.\n", objInserted,
             (float)(totalTime/CLOCK_RATIO));
    printf("Obj Disk Accesses: Average-> %.2f StdDev-> %.2f\n",
                                       objAvgDiskAccess, objStdDevDiskAccess);
    printf("Obj Distance Calculations: Average-> %.2f StdDev-> %.2f\n",
                                       objAvgDistCal, objStdDevDistCal);
    printf("mbOr Disk Accesses: Average-> %.2f StdDev-> %.2f\n",
                                       mbOrAvgDiskAccess, mbOrStdDevDiskAccess);
    printf("mbOr Distance Calculations: Average-> %.2f StdDev-> %.2f\n",
                                       mbOrAvgDistCal, mbOrStdDevDistCal);
    printf("Number of nodes: %d\n", omniIndex.getTotalNodes());

    // queries loop
    // initializing variables
    printf ("\nRange Query: %d queries with radius = %.3f\n", NQueries,
            (float) radius);
    totalTime = clock();
    for(j = 0; j < NQueries; j++) { // average the DAaccess
      omniIndex.resetStatistics();
      searchTime[j] = clock();

      qObj->randomize();

      found[i] = omniIndex.rQuery(qObj, radius);
      // found[i] = omniIndex.rQuery(qObj, radius, &qResponseList);
      // qResponseList.destroyList();
      // printf("%ld\t", found[j]);

      performCounters = omniIndex.getStatistics();
      objDiskAccess[j] = (float) performCounters.objTotalDiskAccess;
      objNumDistCal[j] = (float) performCounters.objDistanceCalculation;
      mbOrDiskAccess[j] = (float) performCounters.mbOrTotalDiskAccess;
      mbOrNumDistCal[j] = (float) performCounters.mbOrDistanceCalculation;
      searchTime[j] = clock() - searchTime[j];
    }
    totalTime = clock() - totalTime; // stop time counter

    // Calculate the disk access average and
    // Standard Deviation for <uniforme and biased> data
    objAvgDiskAccess = Average(objDiskAccess, j, &objStdDevDiskAccess);
    objAvgDistCal = Average(objNumDistCal, j, &objStdDevDistCal);
    mbOrAvgDiskAccess = Average(mbOrDiskAccess, j, &mbOrStdDevDiskAccess);
    mbOrAvgDistCal = Average(mbOrNumDistCal, j, &mbOrStdDevDistCal);
    avgFound = Average(found, j, &stdDevFound);
    avgSearchTime = Average(searchTime, j, &stdDevSearchTime);

    // Output the results of each interval
    fprintf(fdOut,
        //  "Time\t\tObjDisk\t\tmbOrDisk\t\tObjDistCalc\t\t"
        //  "mbOrDistCalc\t\tFound\n"
            "%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\n",
            (float)(totalTime/CLOCK_RATIO), (float)(avgSearchTime/CLOCK_RATIO),
            objAvgDiskAccess, objStdDevDiskAccess, mbOrAvgDiskAccess,
            mbOrStdDevDiskAccess, objAvgDistCal, objStdDevDistCal,
            mbOrAvgDistCal, mbOrStdDevDistCal, avgFound, stdDevFound);

    printf("Total time to %d queries of radius %f: %.2f seconds.\n", NQueries,
             (float) radius, (float)(totalTime/CLOCK_RATIO));
    printf("Qualified objects: Average-> %.2f StdDev-> %.2f\n",
                                       avgFound, stdDevFound);
    printf("Obj Disk Accesses: Average-> %.2f StdDev-> %.2f\n",
                                       objAvgDiskAccess, objStdDevDiskAccess);
    printf("Obj Distance Calculations: Average-> %.2f StdDev-> %.2f\n",
                                       objAvgDistCal, objStdDevDistCal);
    printf("mbOr Disk Accesses: Average-> %.2f StdDev-> %.2f\n",
                                       mbOrAvgDiskAccess, mbOrStdDevDiskAccess);
    printf("mbOr Distance Calculations: Average-> %.2f StdDev-> %.2f\n",
                                       mbOrAvgDistCal, mbOrStdDevDistCal);
  }
  // jumps one line for other results to be appended
  fprintf(fdOut, "\n");

  omniIndex.closeIndex();

  delete[] objDiskAccess;
  delete[] objNumDistCal;
  delete[] mbOrDiskAccess;
  delete[] mbOrNumDistCal;
  delete[] searchTime;
  delete[] found;

  fclose(fdOut);
}

//---------------------------------------------------------------------------
// Execute Exec_NNQStep command
template <class TOmniTreeObj>
void Exec_NNQStep (TOmniTreeObj *qObj, TOmniTreeObj *insObj) {
  int i, j, // counters
      totalIns, // total number of records to be inserted
      steps, // Number of statistical data to be collected
      quotient, // number of records to be inserted per interval
      rest, // number of intervals with different number of records
      objInserted, // number of inserted records
      kNN;  // number of neighbors to be found
  float *objDiskAccess, // vector with number of disk accesses
        objAvgDiskAccess, // average number of disk accesses
        objStdDevDiskAccess, // standard deviation of number of disk accesses
        *mbOrDiskAccess, // vector with number of disk accesses
        mbOrAvgDiskAccess, // average number of disk accesses
        mbOrStdDevDiskAccess, // standard deviation of number of disk accesses
        *objNumDistCal, // vector with number of distance calculations
        objAvgDistCal, // average number of distance calculations
        objStdDevDistCal, // standard deviation of number of distance calculations
        *mbOrNumDistCal, // vector with number of distance calculations
        mbOrAvgDistCal, // average number of distance calculations
        mbOrStdDevDistCal, // standard deviation of number of distance calculations
        *found, // number of objects that qualified
        avgFound, // average of found
        stdDevFound, // standard deviation of found
        *searchTime, // vector with the time for each search
        avgSearchTime, // vector with the time for each search
        stdDevSearchTime; // vector with the time for each search

  clock_t totalTime; // CPU time
  char cmdLine[100]; // command line
  FILE *fdOut; // file descriptor to output results
  char path[100]; // output file name and path

  TOmniTree<TOmniTreeObj> omniIndex;
  typename TOmniTree<TOmniTreeObj>::performanceCounters performCounters;
  // ---- C++ Builder || GNU GCC ----
  #if defined(__BORLANDC__) || defined(__GNUC__)
  TLList<TOmniTreeObj, typename TOmniTreeObj::distanceType> qResponseList(true);
  #endif
  // ---- Visual C++ ----
  #ifdef _MSC_VER
  TLList<TOmniTreeObj, TOmniTreeObj::distanceType> qResponseList(true);
  #endif

  // reading parameters
  ReadWord(cmdLine, false, "nnqstep> #Steps, k-NN, #Objects: ");
  steps = atoi(cmdLine);
  ReadWord(cmdLine, false, "k-NN, #Objects: ");
  kNN = atoi(cmdLine);
  // using ReadCommand instead of ReadWord to make it possible to read the
  // database from another file
  ReadCommand(cmdLine, false, "#Objects: ");
  totalIns = atoi(cmdLine);

  OutPath(idxName, path);
  omniIndex.openIndex(path);

  // initializing variables
  quotient = totalIns / steps; // size of each interval
  rest = totalIns % steps; // number of intervals with 1 more record
                           // necessary when totalInc isn't a multiple of steps

  objInserted = 0;

  objDiskAccess = new float[
                      (NQueries > (quotient + 1) ? NQueries : (quotient + 1))];
  objNumDistCal = new float[
                      (NQueries > (quotient + 1) ? NQueries : (quotient + 1))];

  mbOrDiskAccess = new float[
                      (NQueries > (quotient + 1) ? NQueries : (quotient + 1))];
  mbOrNumDistCal = new float[
                      (NQueries > (quotient + 1) ? NQueries : (quotient + 1))];

  searchTime = new float[
                      (NQueries > (quotient + 1) ? NQueries : (quotient + 1))];

  found = new float[(NQueries > (quotient + 1) ? NQueries : (quotient + 1))];

  OutPath("OmniNNQS.txt", path);
  fdOut = fopen(path, "a");

  if (!(fdOut)) {
    printf("NNQStep Error! Unable to open output file 'OmniNN.txt'.");
    exit(1);
  }

  fprintf(fdOut, "Omni Scan - Nearest Neighbor Query X Data Set Size - kNN = \t"
          "%d\n", kNN);
  fprintf(fdOut, "\t\t\t\tInsertion\t\t\t\t\t\t\t\tSelection\n");
  fprintf(fdOut, "\tTotal\t#Obj\t#\tTime\tObjDisk\t\tmbOrDisk\t\t"
                        "ObjDistCalc\t\tmbOrDistCalc\t\tTime\t\tObjDisk\t"
                        "\tmbOrDisk\t\tObjDistCalc\t\tmbOrDistCalc\t\t"
                        "Found\n");
  fprintf(fdOut, "Step\tObjs\tStep\tPgs\tIns\tAvg\tStdDev\tAvg\tStdDev\t"
                        "Avg\tStdDev\tAvg\tStdDev\tTotal\tAvg\tAvg\t"
                        "StdDev\tAvg\tStdDev\tAvg\tStdDev\tAvg\tStdDev\t"
                        "Avg\tStdDev\n");

  printf("\n========================================\n");
  printf("\tInsert & Nearest Neighbor Query\n");
  printf("\tTotal #Objects = %d\n", totalIns);
  printf("========================================\n");

  // intervals loop
  for(i = 0; i < steps; i++) {
    printf ("\n------------- Interval[%2d] -------------\n", i);

    // insertions loop
    // initializing variables
    printf ("Inserting %5d objects.\n", quotient + (i < rest? 1: 0));
    totalTime = clock();
    for(j = 0; j < (quotient + (i < rest? 1: 0)); j++) {
      omniIndex.resetStatistics();

      insObj->readObj(cmdLine, false, "features: ");

      omniIndex.insert(insObj, objInserted);

      objInserted++;

      performCounters = omniIndex.getStatistics();
      objDiskAccess[j] = (float) performCounters.objTotalDiskAccess;
      objNumDistCal[j] = (float) performCounters.objDistanceCalculation;
      mbOrDiskAccess[j] = (float) performCounters.mbOrTotalDiskAccess;
      mbOrNumDistCal[j] = (float) performCounters.mbOrDistanceCalculation;
    }
    totalTime = clock() - totalTime; // stop time counter
    omniIndex.synchronizeObjFile();

    // Calculate the disk access average and
    // Standard Deviation for <uniforme and biased> data
    objAvgDiskAccess = Average(objDiskAccess, j, &objStdDevDiskAccess);
    objAvgDistCal = Average(objNumDistCal, j, &objStdDevDistCal);
    mbOrAvgDiskAccess = Average(mbOrDiskAccess, j, &mbOrStdDevDiskAccess);
    mbOrAvgDistCal = Average(mbOrNumDistCal, j, &mbOrStdDevDistCal);

    // Output the results of each interval
    fprintf(fdOut,
        //  "\tTotal\t#Obj\t#\tTime\tObjDisk\t\tmbOrDisk\t\t"
        //  "ObjDistCalc\t\tmbOrDistCalc\t\t"
            "%d\t%ld\t%ld\t%ld\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t",
            i, objInserted, j, omniIndex.getTotalNodes(),
            (float)(totalTime/CLOCK_RATIO), objAvgDiskAccess, objStdDevDiskAccess,
            mbOrAvgDiskAccess, mbOrStdDevDiskAccess, objAvgDistCal,
            objStdDevDistCal, mbOrAvgDistCal, mbOrStdDevDistCal);

    printf("Total %d objects. Interval time : %6.2f seconds.\n", objInserted,
             (float)(totalTime/CLOCK_RATIO));
    printf("Obj Disk Accesses: Average-> %.2f StdDev-> %.2f\n",
                                       objAvgDiskAccess, objStdDevDiskAccess);
    printf("Obj Distance Calculations: Average-> %.2f StdDev-> %.2f\n",
                                       objAvgDistCal, objStdDevDistCal);
    printf("mbOr Disk Accesses: Average-> %.2f StdDev-> %.2f\n",
                                       mbOrAvgDiskAccess, mbOrStdDevDiskAccess);
    printf("mbOr Distance Calculations: Average-> %.2f StdDev-> %.2f\n",
                                       mbOrAvgDistCal, mbOrStdDevDistCal);
    printf("Number of nodes: %d\n", omniIndex.getTotalNodes());

    // kNN queries loop
    // initializing variables
    printf ("\nk-Nearest Neighbors Query: %d queries with k = %d\n", NQueries,
                                                                           kNN);
    totalTime = clock();
    for(j = 0; j < NQueries; j++) { // average the DAaccess
      omniIndex.resetStatistics();
      searchTime[j] = clock();

      qObj->randomize();

      qResponseList = omniIndex.kNNQuery(qObj, kNN);
      found[i] = qResponseList.getSize();
      qResponseList.destroyList();
      // printf("%ld\t", found[j]);

      performCounters = omniIndex.getStatistics();
      objDiskAccess[j] = (float) performCounters.objTotalDiskAccess;
      objNumDistCal[j] = (float) performCounters.objDistanceCalculation;
      mbOrDiskAccess[j] = (float) performCounters.mbOrTotalDiskAccess;
      mbOrNumDistCal[j] = (float) performCounters.mbOrDistanceCalculation;
      searchTime[j] = clock() - searchTime[j];
    }
    totalTime = clock() - totalTime; // stop time counter

    // Calculate the disk access average and
    // Standard Deviation for <uniforme and biased> data
    objAvgDiskAccess = Average(objDiskAccess, j, &objStdDevDiskAccess);
    objAvgDistCal = Average(objNumDistCal, j, &objStdDevDistCal);
    mbOrAvgDiskAccess = Average(mbOrDiskAccess, j, &mbOrStdDevDiskAccess);
    mbOrAvgDistCal = Average(mbOrNumDistCal, j, &mbOrStdDevDistCal);
    avgFound = Average(found, j, &stdDevFound);
    avgSearchTime = Average(searchTime, j, &stdDevSearchTime);

    // Output the results of each interval
    fprintf(fdOut,
        //  "Time\t\tObjDisk\t\tmbOrDisk\t\tObjDistCalc\t\t"
        //  "mbOrDistCalc\t\tFound\n"
            "%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\n",
            (float)(totalTime/CLOCK_RATIO), (float)(avgSearchTime/CLOCK_RATIO),
            objAvgDiskAccess, objStdDevDiskAccess, mbOrAvgDiskAccess,
            mbOrStdDevDiskAccess, objAvgDistCal, objStdDevDistCal,
            mbOrAvgDistCal, mbOrStdDevDistCal, avgFound, stdDevFound);

    printf("Total time to nn-query %d objects: %.2f seconds.\n", NQueries,
             (float)(totalTime/CLOCK_RATIO));
    printf("Qualified objects: Average-> %.2f StdDev-> %.2f\n",
                                       avgFound, stdDevFound);
    printf("Obj Disk Accesses: Average-> %.2f StdDev-> %.2f\n",
                                       objAvgDiskAccess, objStdDevDiskAccess);
    printf("Obj Distance Calculations: Average-> %.2f StdDev-> %.2f\n",
                                       objAvgDistCal, objStdDevDistCal);
    printf("mbOr Disk Accesses: Average-> %.2f StdDev-> %.2f\n",
                                       mbOrAvgDiskAccess, mbOrStdDevDiskAccess);
    printf("mbOr Distance Calculations: Average-> %.2f StdDev-> %.2f\n",
                                       mbOrAvgDistCal, mbOrStdDevDistCal);
  }
  // jumps one line for other results to be appended
  fprintf(fdOut, "\n");

  omniIndex.closeIndex();

  delete[] objDiskAccess;
  delete[] objNumDistCal;
  delete[] mbOrDiskAccess;
  delete[] mbOrNumDistCal;
  delete[] searchTime;
  delete[] found;

  fclose(fdOut);
}

//============================================================================
void main(int argc, char **argv) {
  /* This is a simple programm to perform the sequential scanning method */
  char cmdLine[100];
  char path[100];

  TNumericVectObj<numTypeInUse> *qNumObj = NULL, insNumObj;
  TExtractFoci<TNumericVectObj<numTypeInUse> > numFociBase;
  TOmniTree<TNumericVectObj<numTypeInUse> > omniIndexNumVet;

  TStringObj *qStringObj = NULL, insStringObj;
  TExtractFoci<TStringObj> strFociBase;
  TOmniTree<TStringObj> omniIndexStr;

  insNumObj.setReadObjectFunction(ReadWord);
  insStringObj.setReadObjectFunction(ReadWord);

  printf ("**  Omni R-Tree: Version 1.0");
  printf ("**  Build: May-09-01\n\n");

  if (argc > 1) {
    if (redirect(argv[1]) == 0)
      printf("Openning file: %s!\n", argv[1]);
    else {
      printf("Error! Invalid parameter or file name: %s!\n", argv[1]);
      exit(1);
    }
  }

  while (true) {
    ReadCommand (cmdLine, true, "\rEnter command: ");
    if(cmdLine[0]=='?' || !strncmp(cmdLine, "help",1 )) {
      printf ("Commands:\n");
      printf ("CREAteidx ...\n");
      printf ("XTFoci ...\n");
      printf ("RDFoci ...\n");
      printf ("SVFoci ...\n");
      printf ("RSFoci ...\n");
      printf ("MINsert ...\n");
      printf ("MRQ ...\n");
      printf ("MNN ...\n");
      printf ("RQStep ...\n");
      printf ("NNQStep ...\n");
      printf ("SET ...\n");
      printf ("SHParam\n");
      printf ("READ ...\n");
      printf ("CLOSE || QUIT\n");
    }

    //-------------------------- Exec_CreateIdx ---------------------------
    else if(!strncmp(cmdLine, "createidx", 4)) {
      OutPath(idxName, path);
      if (objTypeInUse == numObj) {
        omniIndexNumVet.createIndex(path, objBuffSize, mbOrBuffSize,
                                    numFociBase.getFociBase(), nFoci);
        omniIndexNumVet.closeIndex();
      }
      else {
        omniIndexStr.createIndex(path, objBuffSize, mbOrBuffSize,
                                 strFociBase.getFociBase(), nFoci);
        omniIndexStr.closeIndex();
      }
    }

    //-------------------------- Exec_XtractFociBase ----------------------
    else if(!strncmp(cmdLine, "xtfoci", 3)) {
      ReadWord (cmdLine, false, "Sample file name: ");
      if (objTypeInUse == numObj)
        numFociBase.extractFoci(cmdLine, insNumObj.objectSize(), nFoci);
      else
        strFociBase.extractFoci(cmdLine, insStringObj.objectSize(), nFoci);
    }

    //-------------------------- Exec_ReadFociBaseFromTxtFile -------------
    else if(!strncmp(cmdLine, "rdfoci", 3)) {
      ReadWord (cmdLine, false, "Foci base txt file name to read: ");
      if (objTypeInUse == numObj)
        numFociBase.readFromFile(cmdLine, insNumObj.objectSize(), nFoci);
      else
        strFociBase.readFromFile(cmdLine, insStringObj.objectSize(), nFoci);
    }

    //-------------------------- Exec_SaveFociBaseToTxtFile ---------------
    else if(!strncmp(cmdLine, "svfoci", 3)) {
      ReadWord (cmdLine, false, "Foci base txt file name to write: ");
      if (objTypeInUse == numObj)
        numFociBase.saveToFile(cmdLine);
      else
        strFociBase.saveToFile(cmdLine);
    }

    //-------------------------- Exec_ResetFociBase -----------------------
    else if(!strncmp(cmdLine, "rsfoci", 3)) {
      numFociBase.reset();
      strFociBase.reset();
    }

    //-------------------------- Exec_MInsert ---------------------------
    else if(!strncmp(cmdLine, "minsert", 3)) {
      if (objTypeInUse == numObj)
        Exec_MInsert(&insNumObj);
      else
        Exec_MInsert(&insStringObj);
    }

    //-------------------------- Exec_MRQ ---------------------------
    else if(!strncmp(cmdLine, "mrq", 3)) {
      if (objTypeInUse == numObj) {
        if (CurrentRandMethod == rmUnBiased)
          qNumObj = initMinMax();
        else
          qNumObj = new TNumericVectObj<numTypeInUse>(randFName, dimension);
        Exec_MRQ(qNumObj);
        delete qNumObj;
        qNumObj = NULL;
      }
      else {
        if (CurrentRandMethod == rmUnBiased)
          qStringObj = new TStringObj(dimension);
        else
          qStringObj = new TStringObj(randFName, dimension, 0);
        Exec_MRQ(qStringObj);
        delete qStringObj;
        qStringObj = NULL;
      }
    }

    //-------------------------- Exec_MNN ---------------------------
    else if(!strncmp(cmdLine, "mnn", 3)) {
      if (objTypeInUse == numObj) {
        if (CurrentRandMethod == rmUnBiased)
          qNumObj = initMinMax();
        else
          qNumObj = new TNumericVectObj<numTypeInUse>(randFName, dimension);
        Exec_MNN(qNumObj);
        delete qNumObj;
        qNumObj = NULL;
      }
      else {
        if (CurrentRandMethod == rmUnBiased)
          qStringObj = new TStringObj(dimension);
        else
          qStringObj = new TStringObj(randFName, dimension, 0);
        Exec_MNN(qStringObj);
        delete qStringObj;
        qStringObj = NULL;
      }
    }

    //-------------------------- rqstep -------------------------------------
    else if(!strncmp(cmdLine, "rqstep", 3)) {
      if (objTypeInUse == numObj) {
        if (CurrentRandMethod == rmUnBiased)
          qNumObj = initMinMax();
        else
          qNumObj = new TNumericVectObj<numTypeInUse>(randFName, dimension);
        Exec_RQStep(qNumObj, &insNumObj);
        delete qNumObj;
        qNumObj = NULL;
      }
      else {
        if (CurrentRandMethod == rmUnBiased)
          qStringObj = new TStringObj(dimension);
        else
          qStringObj = new TStringObj(randFName, dimension, 0);
        Exec_RQStep(qStringObj, &insStringObj);
        delete qStringObj;
        qStringObj = NULL;
      }
    }

    //-------------------------- nnqstep ---------------------------
    else if(!strncmp(cmdLine, "nnqstep", 4)) {
      if (objTypeInUse == numObj) {
        if (CurrentRandMethod == rmUnBiased)
          qNumObj = initMinMax();
        else
          qNumObj = new TNumericVectObj<numTypeInUse>(randFName, dimension);
        Exec_NNQStep(qNumObj, &insNumObj);
        delete qNumObj;
        qNumObj = NULL;
      }
      else {
        if (CurrentRandMethod == rmUnBiased)
          qStringObj = new TStringObj(dimension);
        else
          qStringObj = new TStringObj(randFName, dimension, 0);
        Exec_NNQStep(qStringObj, &insStringObj);
        delete qStringObj;
        qStringObj = NULL;
      }
    }

    //--------------------------set--------------------------------------
    else if(!strcmp(cmdLine, "set")) {
      ReadCommand(cmdLine, true, "Set subcommand [OBJType, DIMension, \n"
         "NFoci, OBUFfersize, MBUFfersize, ABUFfersize, IDXName, OUTdir, "
         "NQueries, RAnd]:");

      //----------------------set objtype -------------------------------
      if(!strncmp(cmdLine, "objtype", 4)) {
        ReadCommand(cmdLine, true, "set objtype subcommand [NUMvect, STRing]:");
        if (!strncmp(cmdLine, "numvect", 3)) {
          objTypeInUse = numObj;
          numFociBase.reset();
        }
        else if (!strncmp(cmdLine, "string", 3)) {
          objTypeInUse = stringObj;
          strFociBase.reset();
        }
        else
          printf ("##Dropped: Set: <%s>\n", cmdLine);
        dimension = 0;
      }
      //----------------------set dimension -----------------------------
      else if(!strncmp(cmdLine, "dimension", 3)) {
        ReadWord (cmdLine, false, "Dimension:");
        dimension = atoi(cmdLine);
        insNumObj.setDim(dimension);
        insStringObj.setSize(dimension);
        numFociBase.reset();
        strFociBase.reset();
      }
      //----------------------set nfoci ---------------------------------
      else if(!strncmp(cmdLine, "nfoci", 2)) {
        ReadWord (cmdLine, false, "#Foci: ");
        nFoci = atoi(cmdLine);
        numFociBase.reset();
        strFociBase.reset();
      }
      //----------------------set obuffersize ---------------------------
      else if( !strncmp(cmdLine, "obuffersize", 4) ) {
        ReadWord (cmdLine, false, "Object file buffer size (min = 64):");
        objBuffSize = atoi(cmdLine);
        if (objBuffSize < 64)
          printf ("##Dropped: Buffer Size: <%s>\n", cmdLine);
      }
      //----------------------set mbuffersize ---------------------------
      else if( !strncmp(cmdLine, "mbuffersize", 4) ) {
        ReadWord (cmdLine, false, "mbOr file buffer size (min = 64):");
        mbOrBuffSize = atoi(cmdLine);
        if (mbOrBuffSize < 64)
          printf ("##Dropped: Buffer Size: <%s>\n", cmdLine);
      }
      //----------------------set abuffersize ---------------------------
      else if( !strncmp(cmdLine, "abuffersize", 4) ) {
        ReadWord (cmdLine, false, "Obj and mbOr file buffer size (min = 64):");
        objBuffSize = atoi(cmdLine);
        if (mbOrBuffSize < 64)
          printf ("##Dropped: Buffer Size: <%s>\n", cmdLine);
        mbOrBuffSize = objBuffSize;
      }
      //----------------------set idxname ---------------------------
      else if(!strncmp(cmdLine, "idxname", 3)) {
        ReadWord (idxName, false, "Index file name: ");
      }
      //----------------------set outdir --------------------------------
      else if (!strncmp(cmdLine, "outdir", 3)) {
        ReadWord (outDirName, false, "Output directory name:");
        isOutDirSet = true;
        // ---- GNU GCC ----
        #ifdef __GNUC__
          mkdir(outDirName, 0777);
        #endif
        // ---- C++ Builder || Visual C++ ----
        #if defined(__BORLANDC__) || defined(_MSC_VER)
          mkdir(outDirName);
        #endif
      }
      //----------------------set nqueries ------------------------------
      else if(!strncmp(cmdLine, "nqueries", 2)) {
        ReadWord (cmdLine, false, "NQueries:");
        NQueries=atoi(cmdLine);
        if (NQueries < 1)
          NQueries = 50;
      }
      //----------------------set rand ----------------------------------
      else if(!strncmp(cmdLine, "rand", 2)) {
        ReadCommand (cmdLine, true, "set rand subcommand [BIAsed, UNBIAsed]:");
        if (!strncmp(cmdLine, "biased", 3)) {
          CurrentRandMethod = rmBiased;
          ReadWord (randFName, false, "Input file name:");
        }
        else if (!strncmp(cmdLine, "unbiased", 5))
          CurrentRandMethod = rmUnBiased;
        else
          printf ("##Dropped: Set: <%s>\n", cmdLine);
      }
      //----------------------set (error) -------------------------------
      else
          printf ("##Dropped: Set: <%s>\n", cmdLine);
    }

    //--------------------------show ------------------------------------
    else if(!strncmp(cmdLine, "shparam",2)) {
      printf ("   Index file name: %s\n", idxName);
      printf ("   Object type: %s\n", (objTypeInUse == numObj ?
              "Numeric Vector" : "String"));
      printf ("   Dimension: %d\n", dimension);
      printf ("   #Foci: %d\n", nFoci);
      printf ("   Object page size: %d\n", objBuffSize);
      printf ("   mbOr page size: %d\n", mbOrBuffSize);
      printf ("   Output directory: %s\n", outDirName);
      printf ("   NQueries: %d\n", NQueries);
      printf ("   Radom type: %s\n", (CurrentRandMethod == rmBiased ? "Biased" :
              "Unbiased"));
    }

    //--------------------------quit-------------------------------------
    else if ( (!strcmp(cmdLine, "quit")) || (!strcmp(cmdLine, "close"))
                                            || (!strcmp(cmdLine, "abort")))
      // if there is no other input file openned, finish the program loop.
      break;
    else {
      printf ("##Dropped: <%s>\n", cmdLine);
    //-------------------------------------------------------------------
    }
  }  // End of while true
}// end of main


