//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Omni-RTree for spatial/metric data
//  File: TOmniCoordinates.h
//  Classes: TOmniCoordinates
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 06/24/2000
//  Version: 0.00
//
//  Description:
//  Converts an object to it's Omni coordinates.
//  Checks for overlaping between query regions and omni regions.
//
//  Modifications:
//    Author:
//    Date:
//    Description:
//==============================================================================

#ifndef __TOmniCoordinatesH
#define __TOmniCoordinatesH

#include "DR.h"

//------------------------------------------------------------------------------
// Class TOmniCoordinates definitions
//------------------------------------------------------------------------------
template <class TObjType>
class TOmniCoordinates {
  public:
    TOmniCoordinates() {
      zeroAll();
    };

    TOmniCoordinates(TObjType *foci, int numfoci, int rtId) {
      zeroAll();
      initfociots(foci, numfoci, rtId);
    }

    ~TOmniCoordinates();

    typedef typename TObjType::distanceType TDistType;
    void initFociBase(TObjType *foci, int numFoci, int rtId);
    void toOmniRegion(TObjType *obj);
    char omniIncludes(TDistType *eRegion);
    char qObjmbOrOverlaps(DRrect *rTEntry);
    double qObjmbOrDistance(DRrect *rTEntry);
    void setQRadius(TDistType radius) {
      qRadius = radius;
    }
    void addQRadius();
    char *getRegionBufferArea() {
      return ((char *) qObjRegion);
    }
    DRrect *getRectBufferArea() {
      return ((DRrect *) rTreeRect);
    }

  private:
    int nFoci;
    TObjType *fociBase;
    int rTreeId;
    // the next variables are created to aid queries by reducing the cost
    // to instantiate them every query
    TDistType *rTreeRect;
    TDistType *qObjRegion; // query object in OmniRegion format
    TDistType qRadius;

    void zeroAll() {
      nFoci = 0;
      fociBase = NULL;
      rTreeRect = NULL;
    }
};

//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
template <class TObjType>
TOmniCoordinates<TObjType>::~TOmniCoordinates() {
  if (fociBase) {
    delete[] fociBase;
    fociBase = NULL;

    delete[] qObjRegion;
    qObjRegion = NULL;

    delete[] rTreeRect;
    rTreeRect = NULL;
  }
}

//------------------------------------------------------------------------------
// Initialize the fociBase
// Since TOmniCoordinates makes a copy of the foci base vector, TOmniTree must
// delete the parameter foci.
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniCoordinates<TObjType>::initFociBase(TObjType *foci, int numFoci, int rtId) {
  int i;

  nFoci = numFoci;
  fociBase = new TObjType[nFoci];
  rTreeId = rtId;

  for(i = 0; i < nFoci; i++, foci++)
    fociBase[i] = *foci;

  // the both next variables are created to aid queries by reducing the cost
  // to instantiate them every query
  qObjRegion = new TDistType[numFoci]; // query object in OmniRegion
                                                // format
  rTreeRect = (TDistType *)DRrectNew(rTreeId);
}

//------------------------------------------------------------------------------
// Calculates the OmniRegion that involves the TObjType
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniCoordinates<TObjType>::toOmniRegion(TObjType *obj) {
  int i;
  TDistType *region, *pRectL, *pRectH;

  pRectL = rTreeRect;
  pRectH = pRectL + nFoci;
  region = qObjRegion;
  for(i = 0; i < nFoci; i++, region++, pRectL++, pRectH++) {
    *region = fociBase[i].distance(obj);
    *pRectL = *region;
    *pRectH = *region;
  }
}

//------------------------------------------------------------------------------
// Check if a OmniRegion plus a radius includes an objectRegion.
// Returns true or false. The method convertNKeepToQuery MUST be called prior to
// the first call to this method, while performing a query.
//------------------------------------------------------------------------------
template <class TObjType> char
TOmniCoordinates<TObjType>::omniIncludes(TDistType *omniRegion) {
  int i;
  char result;
  TDistType *p;

  result = true;
  for(i = 0, p = qObjRegion; (i < nFoci) && result; i++, omniRegion++, p++) {
    result = ((*omniRegion >= (*p - qRadius)) &&
              (*omniRegion <= (*p + qRadius)));
  }
  return (result);
}

//------------------------------------------------------------------------------
// returns the distance from rectangle "r" to query point "rectp"
//------------------------------------------------------------------------------
template <class TObjType> char
TOmniCoordinates<TObjType>::qObjmbOrOverlaps(DRrect *rTEntry) {
  int i;
  char result;
  TDistType *p, *pEntryL, *pEntryH;

  pEntryL = (TDistType *) rTEntry;
  pEntryH = pEntryL + nFoci;
  p = qObjRegion;

  drDistCalc++;
  result = true;
  for(i = 0; (i < nFoci) && result; i++, p++, pEntryL++, pEntryH++) {
    result = !((*pEntryL > (*p + qRadius)) ||
               (*pEntryH < (*p - qRadius)));
  }
  return (result);
}

//------------------------------------------------------------------------------
// returns the distance from rectangle "r" to query point "rectp"
//------------------------------------------------------------------------------
template <class TObjType> double
TOmniCoordinates<TObjType>::qObjmbOrDistance(DRrect *rTEntry) {
  int i;
  double difference;
  double result;
  TDistType *p, *pEntryL, *pEntryH;

  pEntryL = (TDistType *) rTEntry;
  pEntryH = pEntryL + nFoci;
  p = qObjRegion;

  drDistCalc++;
  result = 0;
  for(i = 0; i < nFoci; i++, p++, pEntryL++, pEntryH++) {
    difference = (double)((*p) - ((*pEntryH - *pEntryL) * 0.5));
    result += difference * difference;
  }

  return (sqrt(result));
}

//------------------------------------------------------------------------------
// Add the radius to the query region stored into the rTreeRect
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniCoordinates<TObjType>::addQRadius() {
  int i;
  TDistType *pRectL, *pRectH;

  pRectL = rTreeRect;
  pRectH = pRectL + nFoci;
  for(i = 0; i < nFoci; i++, pRectL++, pRectH++) {
    *pRectL -= qRadius;
    *pRectH += qRadius;
  }
}
//---------------------------------------------------------------------------
#endif


