//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Omni-RTree for spatial/metric data
//  File: TOmniTree.cpp
//  Class: TOmniTree
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 06/24/2000
//  Version: 1.00 - Cached Version
//
//  Description:
//  Sequential scan method.
//  Most of the checkings (if the index is opened, if the
//  index is initiated...) are unnecessary because the object TBuffer does it.
//  So, I decided to save some "if"s...
//
//  Modifications:
//    Author:
//    Date:
//    Description:
//==============================================================================

#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <typeinfo>

//---------------------------------------------------------------------------
#pragma hdrstop

#include "DRdef.h"
#include "TOmniTree.h"

// DR file table
extern DRftab_entry DRftab[DR_FTAB_SIZE];

// statistic variables as global
extern long drDiskAccess;
extern long drDistCalc;

//==============================================================================
// Class TOmniTree definitions
//==============================================================================
//------------------------------------------------------------------------------
// Null constructor
//------------------------------------------------------------------------------
template <class TObjType>
TOmniTree<TObjType>::TOmniTree() {
  zeroAll(); // initalize all variables with zero or NULL

  // initialize DR
  DR_Init();
}

//------------------------------------------------------------------------------
// Open an existing index. The index idxName MUST exists
//------------------------------------------------------------------------------
template <class TObjType>
TOmniTree<TObjType>::TOmniTree(char *idxName) {
  zeroAll(); // initalize all variables with zero or NULL

  // initialize DR
  DR_Init();

  openIndex(idxName);
}

//------------------------------------------------------------------------------
// Create and open a new index (ovewrites the old one)
// Since TOmniTree makes a copy of the focus vector, the user must delete
// the parameter foci. Actually, TOmniCoordinates is the one who performs the
// copy.
//------------------------------------------------------------------------------
template <class TObjType>
TOmniTree<TObjType>::TOmniTree(const char *idxName, int objNodeSize,
                                int mbOrNodeSize, TObjType *foci, int numFoci) {
  zeroAll(); // initalize all variables with zero or NULL

  // initialize DR
  DR_Init();

  createIndex(idxName, objNodeSize, mbOrNodeSize, foci, numFoci);
}

//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
template <class TObjType>
TOmniTree<TObjType>::~TOmniTree() {
  // If the index is opened, objBuff and mbOrBuff will print an error message
}

//------------------------------------------------------------------------------
// Open an existing index. The index idxName MUST exists. This object CAN'T have
// another index opened.
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::openIndex(char *idxName) {
  char objTypeNameAux[TOmniTreeOBJTYPENAME];
  int i, idxNameLen;

  strcpy(indexName, idxName);
  idxNameLen = strlen(indexName);
  for(i = idxNameLen - 1; ((indexName[i] != '.') && (i > 0)); i--);

  if (i > 0)
    if (indexName[i - 1] != '.')
      idxNameLen = i;

  i = idxNameLen;
  indexName[i++] = '.';
  indexName[i++] = 'o';
  indexName[i++] = 'x';
  indexName[i++] = 'o';
  indexName[i] = '\0';
  memset(objTypeNameAux, '\0', TOmniTreeOBJTYPENAME);
  strcpy(objTypeNameAux, objHeader.objTypeName);
  objBuffer.openFile(indexName);
  objReadHeader();
  if (strncmp(objTypeNameAux, objHeader.objTypeName, TOmniTreeOBJTYPENAME))
    printErrorMsgNAbort("openIndex error! Objects are of different type.");
  objNode.allocNode(objHeader.nodeSize, objHeader.objSize);

  i = idxNameLen;
  indexName[i++] = '.';
  indexName[i++] = 'o';
  indexName[i++] = 'x';
  indexName[i++] = 'r';
  indexName[i] = '\0';
  // open index
  if ((rTreeId = DR_OpenIndex(indexName)) < 0)
    printErrorMsgNAbort("createIndex error! Unable to open new index.");

  i = idxNameLen;
  indexName[i++] = '.';
  indexName[i++] = 'o';
  indexName[i++] = 'x';
  indexName[i++] = 'h';
  indexName[i] = '\0';

  memset(objTypeNameAux, '\0', TOmniTreeOBJTYPENAME);
  strcpy(objTypeNameAux, omniHeader.mbOrTypeName);

  omniHeaderFile = open(indexName, O_RDWR | O_BINARY);
  if (omniHeaderFile < 0)
    printErrorMsgNAbort("openIndex error! Unable to open header file.");

  readOmniHeader();

  close(omniHeaderFile);

  if (strncmp(objTypeNameAux, omniHeader.mbOrTypeName, TOmniTreeOBJTYPENAME))
    printErrorMsgNAbort("openIndex error! MbOrs are of different type.");
}

//------------------------------------------------------------------------------
// close the actual index - must have one opened
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::closeIndex() {
  if (objIsNodeInUseChanged) {
    objBuffer.writeBuffer(objIdNodeInUse, objNode.getNodeArea());
    objIsNodeInUseChanged = false;
  }
  objWriteHeader();
  objBuffer.closeThis();
  objNode.deleteNode();

  DR_CloseIndex(rTreeId);

  zeroAll(); // initalize all variables with zero or NULL
}

//------------------------------------------------------------------------------
// Create and open a new index (overwrites the old one). This object CAN'T
// have another index opened.
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::createIndex(const char *idxName, int objNodeSize,
                                 int mbOrNodeSize, TObjType *foci, int numFoci) {
  int i, idxNameLen;

  mbOrNodeSize++; // the r-tree will be allways created with the same buffsize
                  // defined in PS.h file. One day, I'll change it to allow
                  // changes on creation time

  omniHeader.nFoci = numFoci;

  strcpy(indexName, idxName);
  idxNameLen = strlen(indexName);
  for(i = idxNameLen - 1; ((indexName[i] != '.') && (i > 0)); i--);

  if (i > 0)
    if (indexName[i - 1] != '.')
      idxNameLen = i;

  i = idxNameLen;
  indexName[i++] = '.';
  indexName[i++] = 'o';
  indexName[i++] = 'x';
  indexName[i++] = 'o';
  indexName[i] = '\0';
  objBuffer.createFile(indexName, objNodeSize, objNodeSize,
                                   TOmniTreeOBJHEADESIZE, TOmniTreeCACHENPAGES);
  objHeader.nodeSize = objNodeSize;
  objHeader.objSize = foci->objectSize();
  // creates the first node - blank
  // the size of one entry node is defined by objSize plus size of objId
  objNode.allocNode(objHeader.nodeSize, objHeader.objSize);
  objIdNodeInUse = objHeader.totalNodes;
  objHeader.totalNodes++;
  objBuffer.insertNewBuffer(objNode.getNodeArea());
  objWriteHeader();

  i = idxNameLen;
  indexName[i++] = '.';
  indexName[i++] = 'o';
  indexName[i++] = 'x';
  indexName[i++] = 'r';
  indexName[i] = '\0';

   // erase old index if exists
  if (access(indexName, 0) >= 0 &&
        // file exist, erase it
        DR_DestroyIndex(indexName) != DRE_OK) {
    printErrorMsgNAbort("createIndex error! Unable to destroy old index.");
  }

  if (typeid(TDistType) == typeid(int)) {
    // create a new empty index
    if (DR_CreateIndex(indexName, omniHeader.nFoci, DR_INT) != DRE_OK)
      printErrorMsgNAbort("createIndex error! Unable to create new index.");
  }
  else if (typeid(TDistType) == typeid(float)) {
    if (DR_CreateIndex(indexName, omniHeader.nFoci, DR_FLOAT) != DRE_OK)
      printErrorMsgNAbort("createIndex error! Unable to create new index.");
  }
  else {
    printErrorMsgNAbort("createIndex error! Distance type not allowed.");
  }

  // open index
  if ((rTreeId = DR_OpenIndex(indexName)) < 0)
    printErrorMsgNAbort("createIndex error! Unable to open new index.");

  i = idxNameLen;
  indexName[i++] = '.';
  indexName[i++] = 'o';
  indexName[i++] = 'x';
  indexName[i++] = 'h';
  indexName[i] = '\0';

  omniHeader.mbOrSize = (sizeof(TDistType) * omniHeader.nFoci);

  omniHeaderFile = open(indexName, O_CREAT | O_TRUNC | O_RDWR | O_BINARY,
            S_IREAD | S_IWRITE);

  if (omniHeaderFile < 0)
    printErrorMsgNAbort("createIndex error! Unable to create header file.");
  writeOmniHeader(foci);
  close(omniHeaderFile);
}

/*
//------------------------------------------------------------------------------
// Close the actual index and reopen overwriting all data. There must be an
// opened index.
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::dropIndex() {
  objBuffer.dropThis();
  objHeader.totalNodes = 0;
  objHeader.totalObjects = 0;
  objWriteHeader();

  mbOrBuffer.dropThis();
  mbOrHeader.totalNodes = 0;
  mbOrWriteHeader();
}
*/

//------------------------------------------------------------------------------
// sync obj file and R-Tree index to fasten obj access
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::synchronizeObjFile() {
  tObjHeader tempHeader; // header
  TOmniTreeNode tempNode;
  TBuffer tempBuffer; // buffer manager
  int objEntryId;
  long mbOrId; // id number of leaf child
  DRrect *obj, *newObj;
  int i;
  double count, countAux;
  TLList<DRrect *, double> path(true); // structure to store path to be followed
  char *pagebuf; // pointer to page buffer of node
  long childNo;

  memcpy(&tempHeader, &objHeader, TOmniTreeOBJHEADESIZE);
  tempHeader.totalNodes = 0;
  tempHeader.totalObjects = 0;
  tempNode.allocNode(tempHeader.nodeSize, tempHeader.objSize);
  tempBuffer.createFile("temp.oxo", tempHeader.nodeSize, tempHeader.nodeSize,
                                   TOmniTreeOBJHEADESIZE, TOmniTreeCACHENPAGES);
  // creates the first node - blank
  // the size of one entry node is defined by objSize plus size of objId
  tempBuffer.writeUserHeader(&tempHeader);

  // traverse index tree
  if (PS_GetThisPage(DRftab[rTreeId].pfd, DRrootPage(rTreeId), &pagebuf) != 0)
    printErrorMsgNAbort("kNNQuery error! Probem with r-Tree.");

  count = 0;
  for (i = DRnodeFirstEntry(rTreeId, pagebuf);
       i >= 0 && i < DRnodeCard(rTreeId);
       i = DRnodeNextEntry(rTreeId, i, pagebuf)) {
    newObj = DRrectNew(rTreeId);
    DRrectCopy(rTreeId, newObj, DRnodeRect(rTreeId, i, pagebuf));
    path.insertEntry(&newObj, (long) DRnodeChild(rTreeId, i, pagebuf), count);
    count--;
  }

  while(path.getEntry(&obj, &childNo, &countAux)) {
    if (PS_GetThisPage(DRftab[rTreeId].pfd, (int) childNo, &pagebuf) != 0)
      printErrorMsgNAbort("kNNQuery error! Probem with r-Tree.");

    if (DRnodeLevel(rTreeId, pagebuf)) { // if non-leaf node
      for (i = DRnodeFirstEntry(rTreeId, pagebuf);
           i >= 0 && i < DRnodeCard(rTreeId);
           i = DRnodeNextEntry(rTreeId, i, pagebuf)) {
        newObj = DRrectNew(rTreeId);
        DRrectCopy(rTreeId, newObj, DRnodeRect(rTreeId, i, pagebuf));
        path.insertEntry(&newObj, (long) DRnodeChild(rTreeId, i, pagebuf),
                         count);
        count--;
      }
      if (PS_UnfixPage(DRftab[rTreeId].pfd, (int) childNo, FALSE) != PSE_OK)
        printErrorMsgNAbort("synchronizeObjFile error! Unable to unfix "
                            "R-Tree internal page!");
    }
    else {	/* leaf node*/
      for (i = DRnodeFirstEntry(rTreeId, pagebuf);
           i >= 0 && i < DRnodeCard(rTreeId);
           i = DRnodeNextEntry(rTreeId, i, pagebuf)) {
        newObj = DRnodeRect(rTreeId, i, pagebuf);
        mbOrId = (long) DRnodeChild(rTreeId, i, pagebuf);
        objReadNode(objToNodeId(mbOrId));
        objEntryId = objNode.toEntryId(mbOrId);

        if (tempNode.insertEntry(objNode.getEntryArea(objEntryId),
                                 objNode.readObjId(objEntryId)) < 0) {
          // the node is full, a new one must be created
          tempBuffer.insertNewBuffer(tempNode.getNodeArea());
          tempHeader.totalNodes++;
          tempNode.ressetNode();
          tempBuffer.writeUserHeader(&tempHeader);
          tempNode.insertEntry(objNode.getEntryArea(objEntryId),
                               objNode.readObjId(objEntryId));
        }
        DRnodeChild(rTreeId, i, pagebuf) = tempHeader.totalObjects;
        tempHeader.totalObjects++;
      }
      /* unpin the leaf node page and mark it dirty */
      if (PS_UnfixPage(DRftab[rTreeId].pfd, (int) childNo, TRUE) != PSE_OK)
        printErrorMsgNAbort("synchronizeObjFile error! Unable to unfix "
                            "R-Tree leaf page!");
    }
    DRrectFree(rTreeId, obj);
  }

  if (PS_UnfixPage(DRftab[rTreeId].pfd, DRrootPage(rTreeId), FALSE) != PSE_OK)
    printErrorMsgNAbort("kNNQuery error! Unable to unfix R-Tree page!");

  tempBuffer.insertNewBuffer(tempNode.getNodeArea());
  tempHeader.totalNodes++;
  tempBuffer.writeUserHeader(&tempHeader);
  tempBuffer.closeThis();
  tempNode.deleteNode();

  if (objIsNodeInUseChanged) {
    objBuffer.writeBuffer(objIdNodeInUse, objNode.getNodeArea());
    objIsNodeInUseChanged = false;
  }
  objWriteHeader();
  objBuffer.closeThis();
  indexName[strlen(indexName) - 1] = 'o';
  unlink(indexName);
  rename("temp.oxo", indexName);
  objBuffer.openFile(indexName);
  objReadHeader();
}

//------------------------------------------------------------------------------
// Insert an object.
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::insert(TObjType *obj, long objId) {
  long mbOrId = objHeader.totalObjects;
  int errorNo;

  objReadNode(objHeader.totalNodes - 1); // get the last node
  if (objNode.insertEntry(obj->bufferArea(), objId) < 0) {
    // the node is full, a new one must be created
    objSaveNodeNGetNew();
    objNode.insertEntry(obj->bufferArea(), objId);
  }
  objHeader.totalObjects++; // the header is saved whenever a new node is
                              // created, just to save disk accesses
  objIsNodeInUseChanged = true;

  omniRegionConvertor.toOmniRegion(obj);
  #ifdef TOmniTreeSTATISTICS
    counters.objDistanceCalculation = (counters.objDistanceCalculation
                                  + omniHeader.nFoci) % TBuffCOUNTERSMAXVALUE;
  #endif
  errorNo = DR_InsertEntry(rTreeId, (int) mbOrId,
                         omniRegionConvertor.getRectBufferArea());
  if (errorNo < 0) {
    DRperror("R-Tree insertion");
    printErrorMsgNAbort("insert error! Probem with r-Tree.");
  }
//  else
//    printf("insert: record %d insert OK\n", objId);
}

//------------------------------------------------------------------------------
// Get the object using the internal id. It is used to access the object
// without a query.
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::getObject(long objPos, TObjType *obj, long *objId) {
  if (objPos >= objHeader.totalObjects)
    printErrorMsgNAbort("getObject error! Access violation.");

  objReadNode(objToNodeId(objPos));

  objNode.getEntry(objNode.toEntryId(objPos), obj->bufferArea(), objId);
}

//------------------------------------------------------------------------------
// Performs a sequential range query
//------------------------------------------------------------------------------
template <class TObjType> long
TOmniTree<TObjType>::rQuery(TObjType *center, TDistType radius,
                    TLList<TObjType, TDistType> *resList) {
  // find the omni region which involves the query object
  omniRegionConvertor.toOmniRegion(center);
  omniRegionConvertor.setQRadius(radius);
  omniRegionConvertor.addQRadius();
  qRadius = radius;
  queryObj = *center;
  resultList = resList;
  #ifdef TOmniTreeSTATISTICS
    counters.objDistanceCalculation = (counters.objDistanceCalculation
                                  + omniHeader.nFoci) % TBuffCOUNTERSMAXVALUE;
  #endif

  // check "testrect"
  if (DRrectInvalid(rTreeId, omniRegionConvertor.getRectBufferArea()))
      return DRerrSave(DRE_RECTINVALID);

  // search index tree
  return (rQuery(DRrootPage(rTreeId)));
}

//------------------------------------------------------------------------------
// Performs a sequential nearest neighbor
//------------------------------------------------------------------------------
// ---- C++ Builder || GNU GCC ----
#if defined(__BORLANDC__) || defined(__GNUC__)
  template <class TObjType> TLList<TObjType, typename TObjType::distanceType>
#endif
// ---- Visual C++ ----
#ifdef _MSC_VER
  template <class TObjType> TLList<TObjType, TObjType::distanceType>
#endif
TOmniTree<TObjType>::kNNQuery(TObjType *center, int k) {
  TOmnikNN<TObjType> kNN(center, k, &omniRegionConvertor);

  #ifdef TOmniTreeSTATISTICS
    counters.objDistanceCalculation = (counters.objDistanceCalculation
                                  + omniHeader.nFoci) % TBuffCOUNTERSMAXVALUE;
  #endif

  kNNQuery(DRrootPage(rTreeId), &kNN);

//  kNN.toFile();
  return (kNN.getResultList());
}

//------------------------------------------------------------------------------
// Saves the actual node and creates a new one
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::objSaveNodeNGetNew() {
  objBuffer.writeBuffer(objIdNodeInUse, objNode.getNodeArea());
  objNode.ressetNode();
  objIdNodeInUse = objHeader.totalNodes;
  objHeader.totalNodes++;
  objBuffer.insertNewBuffer(objNode.getNodeArea());
  objWriteHeader(); // writes the header only with a new node, just to save disk
                 // accesses
}

//------------------------------------------------------------------------------
// Read header
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::readOmniHeader() {
  int i;
  TObjType *foci;

  lseek(omniHeaderFile, 0, SEEK_SET);
  if (read(omniHeaderFile, &omniHeader, TOmniTreeHEADERSIZE) !=
                                                            TOmniTreeHEADERSIZE)
    printErrorMsgNAbort("readHeader error! Error reading header.");

  foci = new TObjType[omniHeader.nFoci];

  for(i = 0; i < omniHeader.nFoci; i++) {
    foci[i].setSize(objHeader.objSize);
    if (read(omniHeaderFile, foci[i].bufferArea(), objHeader.objSize) !=
                                                              objHeader.objSize)
      printErrorMsgNAbort("readHeader error! Error reading foci.");
  }

  omniRegionConvertor.initFociBase(foci, omniHeader.nFoci, rTreeId);
  delete[] foci;
}

//------------------------------------------------------------------------------
// Write header
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::writeOmniHeader(TObjType *foci) {
  int i;

  lseek(omniHeaderFile, 0, SEEK_SET);
  if (write(omniHeaderFile, &omniHeader, TOmniTreeHEADERSIZE) !=
                                                            TOmniTreeHEADERSIZE)
    printErrorMsgNAbort("readHeader error! Error reading header.");

  for(i = 0; i < omniHeader.nFoci; i++) {
    if (write(omniHeaderFile, foci[i].bufferArea(), objHeader.objSize) !=
                                                              objHeader.objSize)
      printErrorMsgNAbort("readHeader error! Error reading foci.");
  }

  omniRegionConvertor.initFociBase(foci, omniHeader.nFoci, rTreeId);
}

//------------------------------------------------------------------------------
// Recursive range query
//------------------------------------------------------------------------------
template <class TObjType> long
TOmniTree<TObjType>::rQuery(int rootpage) {
  int objEntryId;
  long mbOrId;
  TDistType distance;
  TObjType auxObj(objHeader.objSize, (char *)NULL);
  long found;
  int errorNo;  // error number
  char *pagebuf; // pointer to page buffer of node
  int i;  // entry number

  found = 0;

  // get the root node page
  if (PS_GetThisPage(DRftab[rTreeId].pfd, rootpage, &pagebuf) != PSE_OK)
      return DRerrSave(DRE_PSERROR);

  // initialize "found" and "errorNo"
  errorNo = 0;
  found = 0;

  // count disk access
  drDiskAccess++;

  //--------------------------------------------------------------------
  // if the node is NOT a leaf...
  if (DRnodeLevel(rTreeId, pagebuf) > 0) {
    // for each entry in the node...
    for (i = DRnodeFirstEntry(rTreeId, pagebuf);
         i >= 0 && i < DRnodeCard(rTreeId);
         i = DRnodeNextEntry(rTreeId, i, pagebuf)) {
      // if "testrect" matches the rect of an entry...
      if ((*testfn)(rTreeId, omniRegionConvertor.getRectBufferArea(),
                    DRnodeRect(rTreeId, i, pagebuf))) {
        // search the child of the entry
        errorNo = rQuery(DRnodeChild(rTreeId, i, pagebuf));

        // if bad things happened...
        if (errorNo < 0)
            // stop searching
            break;

        // increase counter
        found += errorNo;
      }
    }
  }
  //--------------------------------------------------------------------
  // if the node IS a leaf...
  else if (DRnodeLevel(rTreeId, pagebuf) == 0) {
    // for each entry in the node...
    for (i = DRnodeFirstEntry(rTreeId, pagebuf);
         i >= 0 && i < DRnodeCard(rTreeId);
         i = DRnodeNextEntry(rTreeId, i, pagebuf)) {
      // if "testrect" matches the rect of an entry...
      if ((*testfn)(rTreeId, omniRegionConvertor.getRectBufferArea(),
                    DRnodeRect(rTreeId, i, pagebuf))) {
        // object is inside the omni region
        // must be verified to avoid false alarms
        mbOrId = (long) DRnodeChild(rTreeId, i, pagebuf);
        objReadNode(objToNodeId(mbOrId));
        objEntryId = objNode.toEntryId(mbOrId);
        auxObj.mapBuffer(objNode.getEntryArea(objEntryId));
        if (queryObj.isObjectIncluded(&auxObj, qRadius, &distance)) {
          // increase counter
          found++;
          if (resultList)
            resultList->insertEntry(&auxObj, objNode.readObjId(objEntryId),
                                 distance);
        }

        #ifdef TOmniTreeSTATISTICS
          counters.objDistanceCalculation = (counters.objDistanceCalculation
                                                   + 1) % TBuffCOUNTERSMAXVALUE;
        #endif
      }
/*
      else  {
        mbOrId = (long) DRnodeChild(rTreeId, i, pagebuf);
        objReadNode(objToNodeId(mbOrId));
        objEntryId = objNode.toEntryId(mbOrId);
        auxObj.mapBuffer(objNode.getEntryArea(objEntryId));
        if (queryObj.isObjectIncluded(&auxObj, qRadius, &distance)) {
          distance = distance;
          queryObj.isObjectIncluded(&auxObj, qRadius, &distance);
          (*testfn)(rTreeId, omniRegionConvertor.getRectBufferArea(),
                    DRnodeRect(rTreeId, i, pagebuf));
          printf("dist: %f\t", (float) distance);
        }
      }
*/
    }
  }
  //--------------------------------------------------------------------
  // the level info in the root node is corrupted
  else
    errorNo = DRerrSave(DRE_LEVELINVALID);

  //--------------------------------------------------------------------
  // unfix the root node page
  if (PS_UnfixPage(DRftab[rTreeId].pfd, rootpage, FALSE) != PSE_OK)
    return DRerrSave(DRE_PSERROR);

  // return error number if there is any
  if (errorNo < 0)
    return errorNo;

  // return the number of entries found
  return found;
}

template <class TObjType> void
TOmniTree<TObjType>::kNNQuery(int page, TOmnikNN<TObjType> *kNN) {
  int objEntryId;
  TObjType auxObj(objHeader.objSize, (char *)NULL);
  DRrect *obj, *newObj;
  int i;
  double mbOrDist;
  TLList<DRrect *, double> path(true); // structure to store path to be followed
  char *pagebuf; // pointer to page buffer of node
  long childNo;

  // search index tree
  if (PS_GetThisPage(DRftab[rTreeId].pfd, page, &pagebuf) != 0)
    printErrorMsgNAbort("kNNQuery error! Probem with r-Tree.");
  drDiskAccess++;

  for (i = DRnodeFirstEntry(rTreeId, pagebuf);
       i >= 0 && i < DRnodeCard(rTreeId);
       i = DRnodeNextEntry(rTreeId, i, pagebuf)) {
    newObj = DRrectNew(rTreeId);
    DRrectCopy(rTreeId, newObj, DRnodeRect(rTreeId, i, pagebuf));
    mbOrDist = omniRegionConvertor.qObjmbOrDistance(newObj);
    path.insertEntry(&newObj, (long) DRnodeChild(rTreeId, i, pagebuf),
                     mbOrDist);
  }

  while(path.getEntry(&obj, &childNo, &mbOrDist)) {
    if (kNN->tryNeighbor(obj)) {
      if (DRnodeLevel(rTreeId, pagebuf)) { // if non-leaf node
        kNNQuery(childNo, kNN);
      }
      else {	/* leaf node*/
        // object is inside the farthest neighbor's omni region
        // try to insert it into the knn vector with the truly distance
        // must be verified to avoid false alarms
        objReadNode(objToNodeId(childNo));
        objEntryId = objNode.toEntryId(childNo);
        auxObj.mapBuffer(objNode.getEntryArea(objEntryId));
        kNN->confirmNInsertNN(&auxObj, objNode.readObjId(objEntryId));

        #ifdef TOmniTreeSTATISTICS
          counters.objDistanceCalculation = (counters.objDistanceCalculation
                                                   + 1) % TBuffCOUNTERSMAXVALUE;
        #endif
      }
    }
    DRrectFree(rTreeId, obj);
  }

  if (PS_UnfixPage(DRftab[rTreeId].pfd, page, FALSE) != PSE_OK)
    printErrorMsgNAbort("kNNQuery error! Unable to unfix R-Tree page!");
}

//------------------------------------------------------------------------------
// Print an error message and aborts the program
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::printErrorMsgNAbort(char *errorMsg) {
  if (objHeader.nodeSize != 0) { // test if the index still opened
    objWriteHeader();
    objBuffer.closeThis();
  }

  printf("TOmniTree::%s\n", errorMsg);
  exit(1);

}

//------------------------------------------------------------------------------
// initalize all variables with zero or NULL
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::zeroAll() {
  int typeIdLen;

  strcpy(omniHeader.omniTreeVersion, TOmniTreeVERSION);

  objHeader.nodeSize =
  objHeader.totalNodes =
  objHeader.totalObjects =
  objHeader.objSize = 0;
  objIdNodeInUse = -1;
  objIsNodeInUseChanged = false;
  typeIdLen = strlen(typeid(TObjType).name());
  memset(objHeader.objTypeName, '\0', TOmniTreeOBJTYPENAME);
  typeIdLen = ((typeIdLen < TOmniTreeOBJTYPENAME) ?
               typeIdLen : TOmniTreeOBJTYPENAME);
  memcpy(objHeader.objTypeName, typeid(TObjType).name(), typeIdLen);

  omniHeader.mbOrSize = 0;
  memset(omniHeader.mbOrTypeName, '\0', TOmniTreeOBJTYPENAME);

  typeIdLen = strlen(typeid(TDistType).name());
  typeIdLen = ((typeIdLen < TOmniTreeOBJTYPENAME) ?
               typeIdLen : TOmniTreeOBJTYPENAME);
  memcpy(omniHeader.mbOrTypeName, typeid(TDistType).name(), typeIdLen);

  qRadius = 0;
  queryObj;
  // initialize "testfn"
  testfn = DRrectOverlap;
  resultList = NULL;

  #ifdef TBuffSTATISTICS
    resetStatistics();
  #endif
}

//------------------------------------------------------------------------------
// TOmniTree -> Exporting templates
//------------------------------------------------------------------------------
#include "NumericVectObj.h"
#include "StringObj.h"

// ---- C++ Builder || Visual C++ ----
#if defined(__BORLANDC__) || defined(_MSC_VER)
  template class __export TOmniTree<TNumericVectObj<int> >;
  template class __export TOmniTree<TNumericVectObj<float> >;
  template class __export TOmniTree<TNumericVectObj<double> >;
  template class __export TOmniTree<TStringObj>;
#endif

// ---- GNU GCC ----
#ifdef __GNUC__
  template class TOmniTree<TNumericVectObj<int> >;
  template class TOmniTree<TNumericVectObj<float> >;
  template class TOmniTree<TNumericVectObj<double> >;
  template class TOmniTree<TStringObj>;
#endif
//------------------------------------------------------------------------------
