//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Omni-RTree for spatial/metric data
//  File: TOmniTree.h
//  Class: TOmniTree
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 06/24/2000
//  Version: 1.00 - Cached Version
//
//  Description:
//  Sequential scan method.
//  Most of the checkings (if the index is opened, if the
//  index is initiated...) are unnecessary because the object TBuffer does it.
//  So, I decided to save some "if"s...
//
//  Modifications:
//    Author:
//    Date:
//    Description:
//==============================================================================
//---------------------------------------------------------------------------
#ifndef __OmniTreeH
#define __OmniTreeH

#include <stdlib.h>

#include "TOmniCoordinates.h"
#include "TOmnikNN.h"
#include "TOmniTreeNode.h"
#include "TBuffer.h"
#include "TLList.h"
#include "DR.h"

// ---- C++ Builder ----
#if defined(__BORLANDC__)
  #include <io.h>
  #include <values.h>
#endif

// ---- GNU GCC ----
#ifdef __GNUC__
  #include <unistd.h>
  #include <sys/types.h>
  #include <values.h>
  #ifndef O_BINARY
    #define O_BINARY 0
    // Unix treats binary and text files by the same way. So, it doesn't have
    // the O_BINARY open mode.
    // To be used with Unix, O_BINARY is defined as 0 because it doesn't change
    // the value off '|' operator.
  #endif
#endif

// ---- Visual C++ ----
#ifdef _MSC_VER
  #include <io.h>
  #include <limits.h>
  #define MAXLONG LONG_MAX
#endif

//---- Constant definitions ----
#define TOmniTreeMAXIDXNAMELENGTH 256
#define TOmniTreeSTATISTICS
#define TOmniTreeHEADERRESERVED 100
#define TOmniTreeOBJTYPENAME 30
#define TOmniTreeCACHENPAGES 100
#define TOmniTreeCOUNTERSMAXVALUE (MAXLONG - 100)
#define max(X, Y) ((X) > (Y) ? (X) : (Y))

//------------------------------------------------------------------------------
// Class TOmniTree definitions
//------------------------------------------------------------------------------
template <class TObjType>
class TOmniTree {
  public:
    //---------------------------------------------
    // Constructors and destructor methods
    //---------------------------------------------
    TOmniTree(); // null constructor - just initializes variables
    // open an existing index. The index idxName MUST exists
    TOmniTree(char *idxName);

    // create and open a new index (ovewrites the old one)
    TOmniTree(const char *idxName, int objNodeSize, int mbOrNodeSize,
                                                    TObjType *foci, int numFoci);
    ~TOmniTree();

    typedef typename TObjType::distanceType TDistType;
    //---------------------------------------------
    // index methods
    //---------------------------------------------
    // Open an existing index. The index idxName MUST exists. This object CAN'T
    // have another index opened.
    void openIndex(char *idxName);
    void closeIndex(); // close the actual index - must have one opened
    // create and open a new index (overwrites the old one). This object CAN'T
    // have another index opened.
    void createIndex(const char *idxName, int objNodeSize, int mbOrNodeSize,
                                                    TObjType *foci, int numFoci);
    // close the actual index and reopen overwriting all data. Tthere must be an
    // opened index
    // void dropIndex();
    // sync obj file and R-Tree index to fasten obj access
    void synchronizeObjFile();

    //---------------------------------------------
    // Object methods
    //---------------------------------------------
    void insert(TObjType *obj, long objId); // insert a new object
    // get the object using its position
    void getObject(long objPos, TObjType *obj, long *objId);

    //---------------------------------------------
    // Query methods
    //---------------------------------------------
    // range query
    long rQuery(TObjType *center, TDistType radius,
                TLList<TObjType, TDistType> *resList = NULL);

    TLList<TObjType, TDistType>
         kNNQuery(TObjType *center, int k); // nearest neighbor query

    // search and object with its id and returns the object and its position
    // long objIdQuery(long objId, TObjType *obj);

    //---------------------------------------------
    // General methods
    //---------------------------------------------
    float geTOmniTreeVersion() { // returns the version
      return (atof(&omniHeader.omniTreeVersion[14]));
    }

    // returns the size of the index including headers and records in dispose
    // list (if it is not empty) - not implemented for OmniRTree
//    long getIndexSize() {
//      return(0);
//    }

    // returns the size of the indexed objects. This is important for the user
    // confirm if he is using the write file. It is not interesting to check
    // all the objects that the user gives for insertion and search, but it is
    // important to give the user a way to check it by himself. That's the
    // motivation for this method
    long getObjectSize() {
      return (objHeader.objSize);
    }

    long getTotalObjects() { // returns the total number of objects stored
      return (objHeader.totalObjects);
    }

    long getTotalNodes() { // returns the total number of nodes
      return (objHeader.totalNodes + DRtreeGetTotPages(rTreeId));
    }

    long getNFoci() { // returns the number of foci
      return (omniHeader.nFoci);
    }

    long getRTreeHeight() { // returns the total number of nodes
      return (DR_getTreeHeight(rTreeId));
    }

    //---------------------------------------------
    // Statistics Related - When activated
    //---------------------------------------------
    #ifdef TOmniTreeSTATISTICS
      // counters for performance meter
      struct performanceCounters {
        long objDiskRead;  // read disk counter
        long objDiskWrite; // write disk counter
        long objTotalDiskAccess; // total number of disk accesses
        long objDistanceCalculation; // number of distance calculations
        long mbOrTotalDiskAccess; // total number of disk accesses
        long mbOrDistanceCalculation; // number of distance calculations
      };
      performanceCounters counters;

      performanceCounters getStatistics() {
        TBuffer::performanceCounters objCounters = objBuffer.getStatistics();

        counters.objDiskRead = objCounters.relativeTotalDiskRead;
        counters.objDiskWrite = objCounters.relativeTotalDiskWrite;
        counters.objTotalDiskAccess = objCounters.relativeTotalDiskAccess;
        counters.mbOrTotalDiskAccess = DR_getDiskAccess();
        counters.mbOrDistanceCalculation = DR_getDistCalc();
        return counters;
      }

      void resetStatistics() {
        objBuffer.resetStatistics();
        counters.objDistanceCalculation = 0;
        DR_resetCounters();
      }
    #endif

  private:
    friend TOmniCoordinates<TObjType>;

    //---------------------------------------------
    // Constants and types
    //---------------------------------------------
    #define TOmniTreeVERSION "OMTUGA06/2000-1.0\0" // index version - used to
                            // check if the index was built in the same version
                            // the first 6 characters indicates that it is
                            // a index created by TOmniTree object. Than, comes
                            // the month/year-version number of the object.

    struct tObjHeader {
      int nodeSize; // size of each node - defined in creation time
      long totalObjects; // number of objects stored
      int totalNodes; // number of nodes in use
      long objSize; // size of each object
      char objTypeName[TOmniTreeOBJTYPENAME]; // object type name
      char reserved[TOmniTreeHEADERRESERVED]; // reservs space for future use
    };
    #define TOmniTreeOBJHEADESIZE sizeof(tObjHeader)

    struct TOmniHeader {
      char omniTreeVersion[20];  // version of the TOmniTree object
      int nFoci;
      char mbOrTypeName[TOmniTreeOBJTYPENAME]; // object type name
      long mbOrSize; // size of each object
      char reserved[TOmniTreeHEADERRESERVED]; // reservs space for future use
    };
    #define TOmniTreeHEADERSIZE sizeof(TOmniHeader)

    //---------------------------------------------
    // Variables
    //---------------------------------------------
    char indexName[TOmniTreeMAXIDXNAMELENGTH]; // name of the index file in use
    TOmniHeader omniHeader; // header
    int omniHeaderFile;
    int rTreeId;

    tObjHeader objHeader; // header
    TOmniTreeNode objNode;
    TBuffer objBuffer; // buffer manager
    long objIdNodeInUse;
    char objIsNodeInUseChanged;


    TOmniCoordinates<TObjType> omniRegionConvertor;

    // the following and variables are used by query methods
    TDistType qRadius;
    TObjType queryObj;
    bool_t (*testfn)(int, char *, char *);
    TLList<TObjType, TDistType> *resultList;
    //---------------------------------------------
    // Header methods
    //---------------------------------------------
    void objReadHeader() { // read header
      objBuffer.readUserHeader(&objHeader);
    }

    void objWriteHeader() { // write header
      objBuffer.writeUserHeader(&objHeader);
    }

    void readOmniHeader(); // read header
    void writeOmniHeader(TObjType *foci); // read header

    //---------------------------------------------
    // Node methods
    //---------------------------------------------
    void objSaveNodeNGetNew(); // saves the actual node and creates a new one

    void objReadNode(long node) { // read a node from index
      if (node != objIdNodeInUse) {
        if (objIsNodeInUseChanged) {
          objBuffer.writeBuffer(objIdNodeInUse, objNode.getNodeArea());
          objIsNodeInUseChanged = false;
        }
        objIdNodeInUse = node;
        objBuffer.readBuffer(objIdNodeInUse, objNode.getNodeArea());
      }
    }

    long objToNodeId(long objPosition) { // from object position to nodeId
      #ifdef TOmniTreeNodeDebugg
        if (!objHeader.nodeSize)
          printErrorMsgNAbort("objToNodeId error! There is no index opened.");
      #endif

      return (objPosition / objNode.getNodeCard());
    }


    //---------------------------------------------
    // General methods
    //---------------------------------------------
    long rQuery(int rootpage);
    void kNNQuery(int page, TOmnikNN<TObjType> *kNN);

    void printErrorMsgNAbort(char *errorMsg); // as the name says
    void zeroAll(); // initalize all variables with zero or NULL
};

//---------------------------------------------------------------------------
#endif
