//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Sequential Scan for spatial/metric data
//  File: TOmniTreeNode.h
//  Classes: TOmniTreeNode
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 06/24/2000
//  Version: 1.00 - Cached Version
//
//  Description:
//  Sequential scan node. Most of the methods are defined as inline methods at
//  the header file to fasten it.
//
//  Modifications:
//    Author:
//    Date:
//    Description:
//==============================================================================

#ifndef __TOmniTreeNodeH
#define __TOmniTreeNodeH

#include <string.h>

//---- Constant definitions ----
#define TOmniTreeNodeDebugg

//------------------------------------------------------------------------------
// Class TOmniTreeNode definitions
//------------------------------------------------------------------------------
class TOmniTreeNode {
  public:
    //---------------------------------------------
    // Constructors and destructor methods
    //---------------------------------------------
    TOmniTreeNode();
    TOmniTreeNode(int ndSize, int obSize);
    ~TOmniTreeNode();

    //---------------------------------------------
    // Node methods
    //---------------------------------------------
    void allocNode(int ndSize, int obSize) {
      #ifdef TOmniTreeNodeDebugg
        if (nodeSize)
          printErrorMsgNAbort("allocNode error! Delete node first.");
      #endif

      nodeSize = ndSize;
      objSize = obSize;
      entrySize = objSize + sizeof(long);
      nodeCard = (nodeSize - sizeof(int)) / entrySize;
      // TBuff can partition one object into two or more disk pages, but the
      // buffer size MUST be enough to sotre at leas one object.
      if (!nodeCard)
        printErrorMsgNAbort("allocNode error! Bad node size."
                            " Give me more space, please.");
      nodeBuffer = new char[nodeSize];
      usedSlots = (int *) nodeBuffer;
      firstEntryArea = nodeBuffer + sizeof(int); // jumps usedSlots
      memset(nodeBuffer, '\0', nodeSize);
    }

    void deleteNode() {
      #ifdef TOmniTreeNodeDebugg
        if (!nodeSize)
          printErrorMsgNAbort("deleteNode error! Allocate node first.");
      #endif

      delete[] nodeBuffer;
      zeroAll();
    }

    // This method is to be used ONLY with the last node.
    // To remove an object from another node, replace it with the last
    // object on the file and shrink the last node (remove last).
    // Returns the new node usage, to be checked if is empty
    int shrinkNode() {
      #ifdef TOmniTreeNodeDebugg
        if (!nodeBuffer)
          printErrorMsgNAbort("bufferArea error! Allocate node first.");
      #endif

      return (*usedSlots--);
    }

    char *getNodeArea() { // returns a pointer to a buffer area
      #ifdef TOmniTreeNodeDebugg
        if (!nodeBuffer)
          printErrorMsgNAbort("bufferArea error! Allocate node first.");
      #endif

      return (nodeBuffer);
    }

    void ressetNode() { // set the entire node to blank
      #ifdef TOmniTreeNodeDebugg
        if (!nodeBuffer)
          printErrorMsgNAbort("ressetNode error! Allocate node first.");
      #endif

      memset(nodeBuffer, '\0', nodeSize);
    }

    //---------------------------------------------
    // Entries methods
    //---------------------------------------------
    int insertEntry(void *objBuff, long objId);
    void getEntry(int entryId, void *objBuff, long *objId);
    void putEntry(int entryId, void *objBuff, long objId);

    long readObjId(int entryId) {
      long *pObjId;
      #ifdef TOmniTreeNodeDebugg
        if (!nodeBuffer)
          printErrorMsgNAbort("readObjId error! Allocate node first.");

        if ((entryId < 0) || (entryId >= *usedSlots))
          printErrorMsgNAbort("readObjId error! Access violation.");
      #endif

      pObjId = (long *) (firstEntryArea + ((entrySize * entryId) + objSize));
      return (*pObjId);
    }

    void writeObjId(int entryId, long objId) {
      long *pObjId;
      #ifdef TOmniTreeNodeDebugg
        if (!nodeBuffer)
          printErrorMsgNAbort("writeObjId error! Allocate node first.");

        if ((entryId < 0) || (entryId >= *usedSlots))
          printErrorMsgNAbort("writeObjId error! Access violation.");
      #endif

      pObjId = (long *) (firstEntryArea + ((entrySize * entryId) + objSize));
      *pObjId = objId;
    }

    char *getEntryArea(int entryId) {
      #ifdef TOmniTreeNodeDebugg
        if (!nodeBuffer)
          printErrorMsgNAbort("getEntryArea error! Allocate node first.");

        if ((entryId < 0) || (entryId >= *usedSlots))
          printErrorMsgNAbort("getEntryArea error! Access violation.");
      #endif

      return (firstEntryArea + (entrySize * entryId));
    }


    long toEntryId(long objPosition) {
      #ifdef TOmniTreeNodeDebugg
        if (!nodeCard)
          printErrorMsgNAbort("toEntryId error! Allocate node first.");
      #endif

      return (objPosition % nodeCard);
    }

    //---------------------------------------------
    // General methods
    //---------------------------------------------
    int getNodeSize() {
      #ifdef TOmniTreeNodeDebugg
        if (!nodeSize)
          printErrorMsgNAbort("getNodeSize error! Allocate node first.");
      #endif

      return (nodeCard);
    }

    int getNodeCard() {
      #ifdef TOmniTreeNodeDebugg
        if (!nodeCard)
          printErrorMsgNAbort("getNodeCard error! Allocate node first.");
      #endif

      return (nodeCard);
    }

    int getUsedSlots() {
      #ifdef TOmniTreeNodeDebugg
        if (!nodeBuffer)
          printErrorMsgNAbort("getUsedSlots error! Allocate node first.");
      #endif

      return (*usedSlots);
    }

    int getEmptySlots() {
      #ifdef TOmniTreeNodeDebugg
        if (!nodeBuffer)
          printErrorMsgNAbort("getEmptySlots error! Allocate node first.");
      #endif

      return (nodeCard - *usedSlots);
    }

    float getOccupancyRate() {
      #ifdef TOmniTreeNodeDebugg
        if (!nodeBuffer)
          printErrorMsgNAbort("getOccupancyRate error! Allocate node first.");
      #endif

      return ((float) *usedSlots / (float) nodeCard);
    }

    int getEntrySize() {
      #ifdef TOmniTreeNodeDebugg
        if (!entrySize)
          printErrorMsgNAbort("getEntrySize error! Allocate node first.");
      #endif

      return (nodeCard);
    }

  private:
    //---------------------------------------------
    // Variables
    //---------------------------------------------
    int nodeSize;
    int objSize;
    int entrySize;
    int nodeCard;
    char *nodeBuffer;
    char *firstEntryArea;
    int *usedSlots; // stored in the node buffer

    //---------------------------------------------
    // General methods
    //---------------------------------------------
    void printErrorMsgNAbort(char *errorMsg); // as the name says
    void zeroAll();
};

//---------------------------------------------------------------------------
#endif

/*
//==============================================================================
// ShortCuts
//==============================================================================

//---------------------------------------------
// Constructors and destructor methods
//---------------------------------------------
TOmniTreeNode();
TOmniTreeNode(ndSize, enSize);
~TOmniTreeNode();

//---------------------------------------------
// Node methods
//---------------------------------------------
allocNode(ndSize, enSize);
deleteNode();
shrinkNode();
getNodeArea();
ressetNode();

//---------------------------------------------
// Entries methods
//---------------------------------------------
insertEntry(objBuff, objId);
getEntry(entryId, objBuff, objId);
putEntry(entryId, objBuff, objId);
readObjId(entryId);
writeObjId(entryId, objId);
getEntryArea(entryId);
toEntryId(objPosition);

//---------------------------------------------
// General methods
//---------------------------------------------
getNodeSize();
getNodeCard();
getUsedSlots();
getEmptySlots();
getOccupancyRate()
getEntrySize();

//---------------------------------------------
// Constants and Variables
//---------------------------------------------
TOmniTreeNodeDebugg

nodeSize;
objSize;
entrySize;
nodeCard;
nodeBuffer;
firstEntryArea;
usedSlots;

//---------------------------------------------
// General methods
//---------------------------------------------
printErrorMsgNAbort(errorMsg);
zeroAll();
*/

