//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Sequential Scan for spatial/metric data
//  File: TOmniCoordinates.h
//  Class: TOmniCoordinates
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 06/24/2000
//  Version: 0.00
//
//  Description:
//  Converts an object to it's Omni coordinates.
//  Checks for overlaping between query regions and omni regions.
//
//  Modifications:
//    Author:
//    Date:
//    Description:
//==============================================================================
//---------------------------------------------------------------------------
#ifndef __TOmniCoordinatesH
#define __TOmniCoordinatesH


//------------------------------------------------------------------------------
// Class TOmniRegionConvertor definitions
//------------------------------------------------------------------------------
template <class objType>
class TOmniCoordinates {
  public:
    TOmniCoordinates() {
      zeroAll();
    };

    TOmniCoordinates(objType *foci, int nFoci) {
      zeroAll();
      initFociBase(foci, nFoci);
    }

    ~TOmniCoordinates();

    typedef typename objType::distanceType TDistType;
    void initFociBase(objType *foci, int numFoci);
    void toOmniRegion(objType *obj);
    char omniIncludes(TDistType *eRegion);
    void setQRadius(TDistType radius) { qRadius = radius;};
    char *getRegionBufferArea() {
      return ((char *) qObjRegion);
    }


  private:
    int nFoci;
    objType *fociBase;
    // the next variables are created to aid queries by reducing the cost
    // to instantiate them every query
    TDistType *qObjRegion; // query object in OmniRegion format
    TDistType qRadius;
    void zeroAll() {
      nFoci = 0;
      fociBase = NULL;
    }
};

//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
template <class objType>
TOmniCoordinates<objType>::~TOmniCoordinates() {
  if (fociBase) {
    delete[] fociBase;
    fociBase = NULL;

    delete[] qObjRegion;
    qObjRegion = NULL;
  }
}

//------------------------------------------------------------------------------
// Initialize the fociBase
// Since TOmniCoordinates makes a copy of the focus vector, TOmniTree must
// delete the parameter foci.
//------------------------------------------------------------------------------
template <class objType> void
TOmniCoordinates<objType>::initFociBase(objType *foci, int numFoci) {
  int i;

  nFoci = numFoci;
  fociBase = new objType[nFoci];

  for(i = 0; i < nFoci; i++, foci++)
    fociBase[i] = *foci;

  // the both next variables are created to aid queries by reducing the cost
  // to instantiate them every query
  qObjRegion = new TDistType[numFoci]; // query object in OmniRegion
                                                // format
}

//------------------------------------------------------------------------------
// Calculates the OmniRegion that involves the objType
//------------------------------------------------------------------------------
template <class objType> void
TOmniCoordinates<objType>::toOmniRegion(objType *obj) {
  int i;
  TDistType *region;

  region = qObjRegion;
  for(i = 0; i < nFoci; i++, region++)
    *region = fociBase[i].distance(obj);
}

//------------------------------------------------------------------------------
// Check if a OmniRegion plus a radius includes an objectRegion.
// Returns true or false. The method convertNKeepToQuery MUST be called prior to
// the first call to this method, while performing a query.
//------------------------------------------------------------------------------
template <class objType> char
TOmniCoordinates<objType>::omniIncludes(TDistType *omniRegion) {
  int i;
  char result;
  TDistType *p;

  result = true;
  for(i = 0, p = qObjRegion; (i < nFoci) && result; i++, omniRegion++, p++) {
    result = ((*omniRegion >= (*p - qRadius)) &&
              (*omniRegion <= (*p + qRadius)));
  }
  return (result);
}

//---------------------------------------------------------------------------
#endif

