//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Sequential Scan for spatial/metric data
//  File: TOmniTree.cpp
//  Class: TOmniTree
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 06/24/2000
//  Version: 1.00 - Cached Version
//
//  Description:
//  Sequential scan method.
//  Most of the checkings (if the index is opened, if the
//  index is initiated...) are unnecessary because the object TBuffer does it.
//  So, I decided to save some "if"s...
//
//  Modifications:
//    Author:
//    Date:
//    Description:
//==============================================================================

#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <typeinfo>

//---------------------------------------------------------------------------
#pragma hdrstop

#include "TOmniTree.h"

//==============================================================================
// Class TOmniTree definitions
//==============================================================================
//------------------------------------------------------------------------------
// Null constructor
//------------------------------------------------------------------------------
template <class TObjType>
TOmniTree<TObjType>::TOmniTree() {
  zeroAll(); // initalize all variables with zero or NULL
}

//------------------------------------------------------------------------------
// Open an existing index. The index idxName MUST exists
//------------------------------------------------------------------------------
template <class TObjType>
TOmniTree<TObjType>::TOmniTree(char *idxName) {
  zeroAll(); // initalize all variables with zero or NULL

  openIndex(idxName);
}

//------------------------------------------------------------------------------
// Create and open a new index (ovewrites the old one)
// Since TOmniTree makes a copy of the focus vector, the user must delete
// the parameter foci. Actually, TOmniRegionConvertor is the one who performs
// the copy.
//------------------------------------------------------------------------------
template <class TObjType>
TOmniTree<TObjType>::TOmniTree(const char *idxName, int objNodeSize,
                                 int mbOrNodeSize, TObjType *foci, int numFoci) {
  zeroAll(); // initalize all variables with zero or NULL

  createIndex(idxName, objNodeSize, mbOrNodeSize, foci, numFoci);
}

//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
template <class TObjType>
TOmniTree<TObjType>::~TOmniTree() {
  // If the index is opened, objBuff and mbOrBuff will print an error message
}

//------------------------------------------------------------------------------
// Open an existing index. The index idxName MUST exists. This object CAN'T have
// another index opened.
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::openIndex(char *idxName) {
  char objTypeNameAux[TOmniTreeOBJTYPENAME];
  int i, idxNameLen;

  strcpy(indexName, idxName);
  idxNameLen = strlen(indexName);
  for(i = idxNameLen - 1; ((indexName[i] != '.') && (i > 0)); i--);

  if (i > 0)
    if (indexName[i - 1] != '.')
      idxNameLen = i;

  i = idxNameLen;
  indexName[i++] = '.';
  indexName[i++] = 'o';
  indexName[i++] = 'x';
  indexName[i++] = 'o';
  indexName[i] = '\0';
  memset(objTypeNameAux, '\0', TOmniTreeOBJTYPENAME);
  strcpy(objTypeNameAux, objHeader.objTypeName);
  objBuffer.openFile(indexName);
  objReadHeader();
  if (strncmp(objTypeNameAux, objHeader.objTypeName, TOmniTreeOBJTYPENAME))
    printErrorMsgNAbort("openindex error! Objects are of different type.");
  objNode.allocNode(objHeader.nodeSize, objHeader.objSize);

  i = idxNameLen;
  indexName[i++] = '.';
  indexName[i++] = 'o';
  indexName[i++] = 'x';
  indexName[i++] = 'r';
  indexName[i] = '\0';
  memset(objTypeNameAux, '\0', TOmniTreeOBJTYPENAME);
  strcpy(objTypeNameAux, mbOrHeader.mbOrTypeName);
  mbOrBuffer.openFile(indexName);
  mbOrReadHeader();
  if (strncmp(objTypeNameAux, mbOrHeader.mbOrTypeName, TOmniTreeOBJTYPENAME))
    printErrorMsgNAbort("openindex error! MbOrs are of different type.");
  mbOrNode.allocNode(mbOrHeader.nodeSize, mbOrHeader.mbOrSize);

  i = idxNameLen;
  indexName[i++] = '.';
  indexName[i++] = 'o';
  indexName[i++] = 'x';
  indexName[i++] = 'h';
  indexName[i] = '\0';

  omniHeaderFile = open(indexName, O_RDWR | O_BINARY);
  if (omniHeaderFile < 0)
    printErrorMsgNAbort("openIndex error! Unable to open header file.");

  readOmniHeader();

  close(omniHeaderFile);
}

//------------------------------------------------------------------------------
// close the actual index - must have one opened
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::closeIndex() {
  if (mbOrIsNodeInUseChanged) {
    mbOrBuffer.writeBuffer(mbOrIdNodeInUse, mbOrNode.getNodeArea());
    mbOrIsNodeInUseChanged = false;
  }
  mbOrWriteHeader();
  mbOrBuffer.closeThis();
  mbOrNode.deleteNode();

  if (objIsNodeInUseChanged) {
    objBuffer.writeBuffer(objIdNodeInUse, objNode.getNodeArea());
    objIsNodeInUseChanged = false;
  }
  objWriteHeader();
  objBuffer.closeThis();
  objNode.deleteNode();


  zeroAll(); // initalize all variables with zero or NULL
}

//------------------------------------------------------------------------------
// Create and open a new index (overwrites the old one). This object CAN'T
// have another index opened.
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::createIndex(const char *idxName, int objNodeSize,
                                 int mbOrNodeSize, TObjType *foci, int numFoci) {
  int i, idxNameLen;

  omniHeader.nFoci = numFoci;

  strcpy(indexName, idxName);
  idxNameLen = strlen(indexName);
  for(i = idxNameLen - 1; ((indexName[i] != '.') && (i > 0)); i--);

  if (i > 0)
    if (indexName[i - 1] != '.')
      idxNameLen = i;

  i = idxNameLen;
  indexName[i++] = '.';
  indexName[i++] = 'o';
  indexName[i++] = 'x';
  indexName[i++] = 'o';
  indexName[i] = '\0';
  objBuffer.createFile(indexName, objNodeSize, objNodeSize,
                                   TOmniTreeOBJHEADESIZE, TOmniTreeCACHENPAGES);
  objHeader.nodeSize = objNodeSize;
  objHeader.objSize = foci->objectSize();
  // creates the first node - blank
  // the size of one entry node is defined by objSize plus size of objId
  objNode.allocNode(objHeader.nodeSize, objHeader.objSize);
  objIdNodeInUse = objHeader.totalNodes;
  objHeader.totalNodes++;
  objBuffer.insertNewBuffer(objNode.getNodeArea());
  objWriteHeader();

  i = idxNameLen;
  indexName[i++] = '.';
  indexName[i++] = 'o';
  indexName[i++] = 'x';
  indexName[i++] = 'r';
  indexName[i] = '\0';
  mbOrBuffer.createFile(indexName, mbOrNodeSize, mbOrNodeSize,
                                  TOmniTreeMBORHEADESIZE, TOmniTreeCACHENPAGES);
  mbOrHeader.nodeSize = mbOrNodeSize;
  mbOrHeader.mbOrSize = (sizeof(TDistType) * omniHeader.nFoci);
  // creates the first node - blank
  // the size of one entry node is defined by objSize plus size of objId
  mbOrNode.allocNode(mbOrHeader.nodeSize, mbOrHeader.mbOrSize);
  mbOrIdNodeInUse = mbOrHeader.totalNodes;
  mbOrHeader.totalNodes++;
  mbOrBuffer.insertNewBuffer(mbOrNode.getNodeArea());
  mbOrWriteHeader();

  i = idxNameLen;
  indexName[i++] = '.';
  indexName[i++] = 'o';
  indexName[i++] = 'x';
  indexName[i++] = 'h';
  indexName[i] = '\0';

  omniHeaderFile = open(indexName, O_CREAT | O_TRUNC | O_RDWR | O_BINARY,
            S_IREAD | S_IWRITE);

  if (omniHeaderFile < 0)
    printErrorMsgNAbort("createIndex error! Unable to create header file.");
  writeOmniHeader(foci);
  close(omniHeaderFile);
}

//------------------------------------------------------------------------------
// Close the actual index and reopen overwriting all data. There must be an
// opened index.
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::dropIndex() {
  objBuffer.dropThis();
  objHeader.totalNodes = 0;
  objHeader.totalObjects = 0;
  objWriteHeader();

  mbOrBuffer.dropThis();
  mbOrHeader.totalNodes = 0;
  mbOrWriteHeader();
}

//------------------------------------------------------------------------------
// Insert an object.
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::insert(TObjType *obj, long objId) {
  long mbOrId = objHeader.totalObjects;

  objReadNode(objHeader.totalNodes - 1); // get the last node
  if (objNode.insertEntry(obj->bufferArea(), objId) < 0) {
    // the node is full, a new one must be created
    objSaveNodeNGetNew();
    objNode.insertEntry(obj->bufferArea(), objId);
  }
  objHeader.totalObjects++; // the header is saved whenever a new node is
                              // created, just to save disk accesses
  objIsNodeInUseChanged = true;

  mbOrReadNode(mbOrHeader.totalNodes - 1); // get the last node
  omniRegionConvertor.toOmniRegion(obj);
  #ifdef TOmniTreeSTATISTICS
    counters.objDistanceCalculation = (counters.objDistanceCalculation
                                  + omniHeader.nFoci) % TBuffCOUNTERSMAXVALUE;
  #endif
  if (mbOrNode.insertEntry(omniRegionConvertor.getRegionBufferArea(), mbOrId)
      < 0) {
    // the node is full, a new one must be created
    mbOrSaveNodeNGetNew();
    mbOrNode.insertEntry(omniRegionConvertor.getRegionBufferArea(), mbOrId);
  }
  mbOrIsNodeInUseChanged = true;
}

//------------------------------------------------------------------------------
// Get the object using the internal id. It is used to access the object
// without a query.
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::getObject(long objPos, TObjType *obj, long *objId) {
  if (objPos >= objHeader.totalObjects)
    printErrorMsgNAbort("getObject error! Access violation.");

  objReadNode(objToNodeId(objPos));

  objNode.getEntry(objNode.toEntryId(objPos), obj->bufferArea(), objId);
}

//------------------------------------------------------------------------------
// Performs a sequential range query
//------------------------------------------------------------------------------
template <class TObjType> long
TOmniTree<TObjType>::rQuery(TObjType *center, TDistType radius,
                    TLList<TObjType, TDistType> *resultList) {
  int objEntryId;
  long mbOrNodeId;
  int mbOrEntryId;
  int mbOrMaxEntry;
  long mbOrId;
  TDistType distance;
  TObjType objAux(objHeader.objSize, (char *)NULL);
  long found;

  // find the omni region which involves the query object
  omniRegionConvertor.toOmniRegion(center);
  omniRegionConvertor.setQRadius(radius);
  #ifdef TOmniTreeSTATISTICS
    counters.objDistanceCalculation = (counters.objDistanceCalculation
                                  + omniHeader.nFoci) % TBuffCOUNTERSMAXVALUE;
  #endif
  found = 0;
  
  for (mbOrNodeId = 0; mbOrNodeId < mbOrHeader.totalNodes; mbOrNodeId++) {
    mbOrReadNode(mbOrNodeId);
    mbOrMaxEntry = mbOrNode.getUsedSlots();
    for (mbOrEntryId = 0; mbOrEntryId < mbOrMaxEntry; mbOrEntryId++) {
      if (omniRegionConvertor.omniIncludes((TDistType *)
                                          mbOrNode.getEntryArea(mbOrEntryId))) {
        // object is inside the omni region
        // must be verified to avoid false alarms
        mbOrId = mbOrNode.readObjId(mbOrEntryId);
        objReadNode(objToNodeId(mbOrId));
        objEntryId = objNode.toEntryId(mbOrId);
        objAux.mapBuffer(objNode.getEntryArea(objEntryId));
        if ((*center).isObjectIncluded(&objAux, radius, &distance)) {
          found++;
          if (resultList)
            resultList->insertEntry(&objAux, objNode.readObjId(objEntryId),
                                 distance);
        }

        #ifdef TOmniTreeSTATISTICS
          counters.objDistanceCalculation = (counters.objDistanceCalculation
                                                   + 1) % TBuffCOUNTERSMAXVALUE;
        #endif
      }

//      else  {
//        mbOrId = mbOrNode.readObjId(mbOrEntryId);
//        objReadNode(objToNodeId(mbOrId));
//        objEntryId = objNode.toEntryId(mbOrId);
//        objAux.mapBuffer(objNode.getEntryArea(objEntryId));
//        if ((*center).isObjectIncluded(&objAux, radius, &distance)) {
//          distance = distance;
//          (*center).isObjectIncluded(&objAux, radius, &distance);
//          omniRegionConvertor.omniIncludes((TDistType *)
//                                          mbOrNode.getEntryArea(mbOrEntryId));
//          printf("dist: %f\t", (float) distance);
//        }
//      }



      #ifdef TOmniTreeSTATISTICS
        counters.mbOrDistanceCalculation = (counters.mbOrDistanceCalculation
                                                   + 1) % TBuffCOUNTERSMAXVALUE;
      #endif
    }
  }

  return (found);
}

//------------------------------------------------------------------------------
// Performs a sequential nearest neighbor
//------------------------------------------------------------------------------
// ---- C++ Builder || GNU GCC ----
#if defined(__BORLANDC__) || defined(__GNUC__)
  template <class TObjType> TLList<TObjType, typename TObjType::distanceType>
#endif
// ---- Visual C++ ----
#ifdef _MSC_VER
  template <class TObjType> TLList<TObjType, TObjType::distanceType>
#endif
TOmniTree<TObjType>::kNNQuery(TObjType *center, int k) {
  TOmnikNN<TObjType> kNN(center, k, &omniRegionConvertor);
  int objEntryId;
  long mbOrNodeId;
  int mbOrEntryId;
  int mbOrMaxEntry;
  long mbOrId;
  TObjType objAux(objHeader.objSize, (char *)NULL);

  #ifdef TOmniTreeSTATISTICS
    counters.objDistanceCalculation = (counters.objDistanceCalculation
                                  + omniHeader.nFoci) % TBuffCOUNTERSMAXVALUE;
  #endif

  for (mbOrNodeId = 0; mbOrNodeId < mbOrHeader.totalNodes; mbOrNodeId++) {
    mbOrReadNode(mbOrNodeId);
    mbOrMaxEntry = mbOrNode.getUsedSlots();
    for (mbOrEntryId = 0; mbOrEntryId < mbOrMaxEntry; mbOrEntryId++) {
      if (kNN.tryNeighbor((TDistType *)
                                          mbOrNode.getEntryArea(mbOrEntryId))) {
        // object is inside the farthes neighbor's omni region
        // try to insert it into the knn vector with the truly distance
        // must be verified to avoid false alarms
        mbOrId = mbOrNode.readObjId(mbOrEntryId);
        objReadNode(objToNodeId(mbOrId));
        objEntryId = objNode.toEntryId(mbOrId);
        objAux.mapBuffer(objNode.getEntryArea(objEntryId));
        kNN.confirmNInsertNN(&objAux, objNode.readObjId(objEntryId));

        #ifdef TOmniTreeSTATISTICS
          counters.objDistanceCalculation = (counters.objDistanceCalculation
                                                   + 1) % TBuffCOUNTERSMAXVALUE;
        #endif
      }
      #ifdef TOmniTreeSTATISTICS
        counters.mbOrDistanceCalculation = (counters.mbOrDistanceCalculation
                                                   + 1) % TBuffCOUNTERSMAXVALUE;
      #endif
    }
  }
//  kNN.toFile();
  return (kNN.getResultList());
}

//------------------------------------------------------------------------------
// Saves the actual node and creates a new one
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::objSaveNodeNGetNew() {
  objBuffer.writeBuffer(objIdNodeInUse, objNode.getNodeArea());
  objNode.ressetNode();
  objIdNodeInUse = objHeader.totalNodes;
  objHeader.totalNodes++;
  objBuffer.insertNewBuffer(objNode.getNodeArea());
  objWriteHeader(); // writes the header only with a new node, just to save disk
                 // accesses
}

//------------------------------------------------------------------------------
// Saves the actual node and creates a new one
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::mbOrSaveNodeNGetNew() {
  mbOrBuffer.writeBuffer(mbOrIdNodeInUse, mbOrNode.getNodeArea());
  mbOrNode.ressetNode();
  mbOrIdNodeInUse = mbOrHeader.totalNodes;
  mbOrHeader.totalNodes++;
  mbOrBuffer.insertNewBuffer(mbOrNode.getNodeArea());
  mbOrWriteHeader(); // writes the header only with a new node, just to save
                 // disk accesses
}

//------------------------------------------------------------------------------
// Read header
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::readOmniHeader() {
  int i;
  TObjType *foci;

  lseek(omniHeaderFile, 0, SEEK_SET);
  if (read(omniHeaderFile, &omniHeader, TOmniTreeHEADERSIZE) !=
                                                            TOmniTreeHEADERSIZE)
    printErrorMsgNAbort("readHeader error! Error reading header.");

  foci = new TObjType[omniHeader.nFoci];

  for(i = 0; i < omniHeader.nFoci; i++) {
    foci[i].setSize(objHeader.objSize);
    if (read(omniHeaderFile, foci[i].bufferArea(), objHeader.objSize) !=
                                                              objHeader.objSize)
      printErrorMsgNAbort("readHeader error! Error reading foci.");
  }

  omniRegionConvertor.initFociBase(foci, omniHeader.nFoci);
  delete[] foci;
}

//------------------------------------------------------------------------------
// Write header
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::writeOmniHeader(TObjType *foci) {
  int i;

  lseek(omniHeaderFile, 0, SEEK_SET);
  if (write(omniHeaderFile, &omniHeader, TOmniTreeHEADERSIZE) !=
                                                            TOmniTreeHEADERSIZE)
    printErrorMsgNAbort("readHeader error! Error reading header.");

  for(i = 0; i < omniHeader.nFoci; i++) {
    if (write(omniHeaderFile, foci[i].bufferArea(), objHeader.objSize) !=
                                                              objHeader.objSize)
      printErrorMsgNAbort("readHeader error! Error reading foci.");
  }

  omniRegionConvertor.initFociBase(foci, omniHeader.nFoci);
}

//------------------------------------------------------------------------------
// Print an error message and aborts the program
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::printErrorMsgNAbort(char *errorMsg) {
  if (objHeader.nodeSize != 0) { // test if the index still opened
    objWriteHeader();
    mbOrWriteHeader();
    objBuffer.closeThis();
    mbOrBuffer.closeThis();
  }

  printf("TOmniTree::%s\n", errorMsg);
  exit(1);

}

//------------------------------------------------------------------------------
// initalize all variables with zero or NULL
//------------------------------------------------------------------------------
template <class TObjType> void
TOmniTree<TObjType>::zeroAll() {
  int typeIdLen;

  strcpy(omniHeader.omniTreeVersion, TOmniTreeVERSION);

  objHeader.nodeSize =
  objHeader.totalNodes =
  objHeader.totalObjects =
  objHeader.objSize = 0;
  objIdNodeInUse = -1;
  objIsNodeInUseChanged = false;
  typeIdLen = strlen(typeid(TObjType).name());
  memset(objHeader.objTypeName, '\0', TOmniTreeOBJTYPENAME);
  typeIdLen = ((typeIdLen < TOmniTreeOBJTYPENAME) ?
               typeIdLen : TOmniTreeOBJTYPENAME);
  memcpy(objHeader.objTypeName, typeid(TObjType).name(), typeIdLen);

  mbOrHeader.nodeSize =
  mbOrHeader.totalNodes =
  mbOrHeader.mbOrSize = 0;
  mbOrIdNodeInUse = -1;
  mbOrIsNodeInUseChanged = false;
  memset(mbOrHeader.mbOrTypeName, '\0', TOmniTreeOBJTYPENAME);

  typeIdLen = strlen(typeid(TDistType).name());
  typeIdLen = ((typeIdLen < TOmniTreeOBJTYPENAME) ?
               typeIdLen : TOmniTreeOBJTYPENAME);
  memcpy(mbOrHeader.mbOrTypeName, typeid(TDistType).name(), typeIdLen);
  #ifdef TBuffSTATISTICS
    resetStatistics();
  #endif
}

//------------------------------------------------------------------------------
// TOmniRegionConvertor -> Exporting templates
//------------------------------------------------------------------------------
#include "NumericVectObj.h"
#include "StringObj.h"

//------------------------------------------------------------------------------
// TOmniTree -> Exporting templates
//------------------------------------------------------------------------------
// ---- C++ Builder || Visual C++ ----
#if defined(__BORLANDC__) || defined(_MSC_VER)
  template class __export TOmniTree<TNumericVectObj<int> >;
  template class __export TOmniTree<TNumericVectObj<float> >;
  template class __export TOmniTree<TNumericVectObj<double> >;
  template class __export TOmniTree<TStringObj>;
#endif

// ---- GNU GCC ----
#ifdef __GNUC__
  template class TOmniTree<TNumericVectObj<int> >;
  template class TOmniTree<TNumericVectObj<float> >;
  template class TOmniTree<TNumericVectObj<double> >;
  template class TOmniTree<TStringObj>;
#endif

//------------------------------------------------------------------------------

