//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Sequential Scan for spatial/metric data
//  File: TOmniTreeNode.h
//  Classes: TOmniTreeNode
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 06/24/2000
//  Version: 1.00 - Cached Version
//
//  Description:
//  Sequential scan node. Most of the methods are defined as inline methods at
//  the header file to fasten it.
//
//  Modifications:
//    Author:
//    Date:
//    Description:
//==============================================================================

#include <stdlib.h>
#include <stdio.h>
//---------------------------------------------------------------------------
#pragma hdrstop

#include "TOmniTreeNode.h"

//------------------------------------------------------------------------------
// Null constructor
//------------------------------------------------------------------------------
TOmniTreeNode::TOmniTreeNode() {
  zeroAll(); // initalize all variables with zero or NULL
}

//------------------------------------------------------------------------------
// Main constructor
//------------------------------------------------------------------------------
TOmniTreeNode::TOmniTreeNode(int ndSize, int obSize) {
  zeroAll(); // initalize all variables with zero or NULL

  nodeSize = ndSize;
  objSize = obSize;
  entrySize = objSize + sizeof(long);
  nodeCard = (nodeSize - sizeof(int)) / entrySize;
  if (!nodeCard)
      printErrorMsgNAbort("TOmniTreeNode error! Bad node size."
                            " Give me more space, please.");
  nodeBuffer = new char[nodeSize];
  usedSlots = (int *) nodeBuffer;
  firstEntryArea = nodeBuffer + sizeof(int); // jumps usedSlots
  memset(nodeBuffer, '\0', nodeSize);
}

//------------------------------------------------------------------------------
// Desstructor
//------------------------------------------------------------------------------
TOmniTreeNode::~TOmniTreeNode() {
  if (nodeSize) {
    delete[] nodeBuffer;
  }
  zeroAll();
}

//------------------------------------------------------------------------------
// Insert new entry. Returns the entryId (node position), or -1 if the node is
// full.
//------------------------------------------------------------------------------
int
TOmniTreeNode::insertEntry(void *objBuff, long objId) {
  int result = -1;

  #ifdef TOmniTreeNodeDebugg
    if (!nodeBuffer)
      printErrorMsgNAbort("insert error! Allocate node first.");
  #endif

  if (*usedSlots < nodeCard) {
    result = *usedSlots;
    (*usedSlots)++;
    memcpy(getEntryArea(result), objBuff, objSize);
    writeObjId(result, objId);
  }

  return (result);
}

//------------------------------------------------------------------------------
// Read an entry
//------------------------------------------------------------------------------
void
TOmniTreeNode::getEntry(int entryId, void *objBuff, long *objId) {
  #ifdef TOmniTreeNodeDebugg
    if (!nodeBuffer)
      printErrorMsgNAbort("getEntry error! Allocate node first.");

    if ((entryId < 0) || (entryId >= *usedSlots))
      printErrorMsgNAbort("getEntry error! Access violation.");
  #endif

  memcpy(objBuff, getEntryArea(entryId), objSize);
  *objId = readObjId(entryId);
}

//------------------------------------------------------------------------------
// Replace an entry
//------------------------------------------------------------------------------
void
TOmniTreeNode::putEntry(int entryId, void *objBuff, long objId) {
  #ifdef TOmniTreeNodeDebugg
    if (!nodeBuffer)
      printErrorMsgNAbort("insert error! Allocate node first.");
  #endif

  if (entryId < *usedSlots) {
    memcpy(getEntryArea(entryId), objBuff, entrySize);
    writeObjId(entryId, objId);
  }
  else
    printErrorMsgNAbort("replace error! Access violation.");
}

//------------------------------------------------------------------------------
// Print an error message and aborts the program
//------------------------------------------------------------------------------
void
TOmniTreeNode::printErrorMsgNAbort(char *errorMsg) {
  printf("TOmniTreeNode::%s\n", errorMsg);
  exit(1);
}

//------------------------------------------------------------------------------
// initalize all variables with zero or NULL
//------------------------------------------------------------------------------
void
TOmniTreeNode::zeroAll(){
  nodeSize =
  objSize =
  entrySize =
  nodeCard = 0;

  nodeBuffer = NULL;
  usedSlots = NULL;
}

