//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Sequential Scan for spatial/metric data
//  File: TOmnikNN.h
//  Class: TOmnikNN
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 06/24/2000
//  Version: 0.00
//
//  Description:
//  A class to manage the k-nearest neighbors list.
//
//  Modifications:
//    Author:
//    Date:
//    Description:
//==============================================================================
//---------------------------------------------------------------------------
#ifndef __TOmnikNNH
#define __TOmnikNNH

#include "TLList.h"
#include "TOmniCoordinates.h"

//------------------------------------------------------------------------------
// Class TOmnikNN definitions
//------------------------------------------------------------------------------
template <class TObjType>
class TOmnikNN {
  public:
    TOmnikNN(TObjType *center, int kToFind,
             TOmniCoordinates<TObjType> *convertor);
    ~TOmnikNN();
    typedef typename TObjType::distanceType TDistType;
    char tryNeighbor(TDistType *omniRegion);
    char confirmNInsertNN(TObjType *obj, long objId);
    // The next method returns the result list. After returning it, the
    // kNN object will be reinitialized. The user MUST destroy the list.
    // If the user doesn't call this method, kNN will destroy it.
    TLList<TObjType, TDistType> getResultList();
    void printNN();
    void toFile();

  private:
    struct tNeighbor {
      long objId;
      TObjType neighbor;
      TDistType trulyDistance;
    };

    char adjustkNNVet();

    int k;
    int maxNN;
    tNeighbor **kNN; // pointer to pointer to fasten swap operations
    TOmniCoordinates<TObjType> *omniCoordinates;
    TObjType qObj; // keeps the query object
    TLList<TObjType, TDistType> tie; // tie list
};

//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
template <class TObjType>
TOmnikNN<TObjType>::TOmnikNN(TObjType *center, int kToFind,
                         TOmniCoordinates<TObjType> *convertor) : tie(true) {
  int i;
  tNeighbor **p;

  k = 0;
  maxNN = kToFind;
  kNN = new tNeighbor*[maxNN];
  for(i = 0, p = kNN; i < maxNN; i++, p++)
    *p = new tNeighbor;

  omniCoordinates = convertor;
  qObj = *center;
  // find the omni region which involves the query object
  // initializes radius with 0 is not a problem because tryNeighbor allays
  // returns true when k < maxNN
  omniCoordinates->toOmniRegion(center);
  omniCoordinates->setQRadius(0);
}

//------------------------------------------------------------------------------
// Destructor
//------------------------------------------------------------------------------
template <class TObjType>
TOmnikNN<TObjType>::~TOmnikNN() {
  int i;
  tNeighbor **p;

  if (maxNN) {
    for(i = 0, p = kNN; i < maxNN; i++, p++) {
     delete *p;
     *p = NULL;
    }

    delete[] kNN;
    kNN = NULL;

    tie.destroyList(); // if maxNN is not NULL, the user didn't get the list
  }
}

//------------------------------------------------------------------------------
// Try neighbor
//------------------------------------------------------------------------------
template <class TObjType> char
TOmnikNN<TObjType>::tryNeighbor(TDistType *omniRegion) {
  return ((k < maxNN) ? true : omniCoordinates->omniIncludes(omniRegion));
}

//------------------------------------------------------------------------------
// Ordered insert neighbor. Returns true if it is really nearest
// and false in the case of a false alam
//------------------------------------------------------------------------------
template <class TObjType> char
TOmnikNN<TObjType>::confirmNInsertNN(TObjType *dbObj, long objId) {
  char result;
  TDistType trulyDistance = qObj.distance(dbObj);
  int pos;

  result = false;
  if (k < maxNN) {
    pos = k;
    k++;
    result = true;
  }
  else {
    if (kNN[k - 1]->trulyDistance > trulyDistance) {
      pos = k - 1;
      result = true;
      if (k > 1) {
        if (kNN[k - 1]->trulyDistance == kNN[k - 2]->trulyDistance) {
          // there is a tie, move the last one to the tie list
          tie.insertEntry(&(kNN[pos]->neighbor), kNN[pos]->objId,
                          kNN[k - 2]->trulyDistance);
        }
        else // the radius of the farthest will change, there are no more ties.
          tie.destroyList();
      }
    }
  }

  if (result) {
    kNN[pos]->objId = objId;
    kNN[pos]->neighbor = *dbObj;
    kNN[pos]->trulyDistance = trulyDistance;
    adjustkNNVet();
      omniCoordinates->setQRadius(kNN[pos]->trulyDistance);
    result = true;
  }

  // its a tie situation
  if ((!result) && (kNN[k - 1]->trulyDistance == trulyDistance)) {
    tie.insertEntry(dbObj, objId, trulyDistance);
    result = true;
  }

  return result;
}

//------------------------------------------------------------------------------
// Print the NN vector
//------------------------------------------------------------------------------
template <class TObjType> void
TOmnikNN<TObjType>::printNN() {
  int i;
  tNeighbor **p;

  for (i = 0, p = kNN; i < k; i++, p++) {
//    printf("ObjId: %ld\t Distance: %f\n", (*p)->, (*p)->trulyDistance < (*q)->trulyDistance) {
    printf("Obj[%2d] Distance: %.5f\n", i, (*p)->trulyDistance);
  }
}

//------------------------------------------------------------------------------
// Print the NN vector
//------------------------------------------------------------------------------
template <class TObjType> void
TOmnikNN<TObjType>::toFile() {
  FILE *fdOut;
  int i;
  tNeighbor **p;

  if ((fdOut = fopen("NNResult.txt", "at")) == NULL) {
    printf("TOmnikNN::toFile error! Can't open file '%s'.\n", "NNResult.txt");
    exit(1);
  }

  for (i = 0, p = kNN; i < k; i++, p++) {
    fprintf(fdOut, "%ld\t%.5f\n", (*p)->objId, (*p)->trulyDistance);
//    printf("Obj[%2d] Distance: %f\n", i, (*p)->trulyDistance);
  }
  fclose(fdOut);
}

//------------------------------------------------------------------------------
// Assemble the result list with the ties and the nearest neighbors
//------------------------------------------------------------------------------
// ---- C++ Builder || GNU GCC ----
#if defined(__BORLANDC__) || defined(__GNUC__)
  template <class TObjType> TLList<TObjType, typename TObjType::distanceType>
#endif
// ---- Visual C++ ----
#ifdef _MSC_VER
  template <class TObjType> TLList<TObjType, TObjType::distanceType>
#endif
TOmnikNN<TObjType>::getResultList() {
  int i;
  tNeighbor **p;

  // Copies the knn vector inside tie list and destroy the vector
  // Since the tie list was created ordered, it'll be faster to insert
  // the farthest neighbor first
  if (maxNN) {
    for(i = maxNN, p = &kNN[maxNN - 1]; i ; i--, p--) {
      tie.insertEntry(&((*p)->neighbor), (*p)->objId, (*p)->trulyDistance);

      delete *p;
      *p = NULL;
    }

    delete[] kNN;
    kNN = NULL;
    k = 0;
    maxNN = 0;
  }

  return (tie);
}

//------------------------------------------------------------------------------
// Bubble adjust the vector to maintain the order
//------------------------------------------------------------------------------
template <class TObjType> char
TOmnikNN<TObjType>::adjustkNNVet() {
  int i = k - 1;
  tNeighbor **p, // last position
            **q, // anterior to p
            *aux;
  char result;

  result = false;
  p = kNN + i;
  while (i) {
    q = p - 1;
    if ((*q)->trulyDistance > (*p)->trulyDistance) {
      aux = *p;
      *p = *q;
      *q = aux;
      i--;
      p--;
      result = true;
    }
    else
      i = 0;
  }
  return (result);
}
//---------------------------------------------------------------------------

#endif

